/*
----------------------------------------------------------------------
	
	AdAstra - Real Time Kernel

	Alain Chebrou

	bsp.c		target specific "C" routines for ARM Cortex M7
				For STM32H7 family

	When		Who	What
	28/04/17	ac	Creation
	12/04/17	ac	Add naked to pendSV_Handler()
	21/08/19	ac	Add naked to bspStart_()
	15/01/20	ac	Port to H7 from F4
	13/03/20	ac	Conditional initialization of logMess 
	23/04/20	ac	Start modification for armcc compatibility
	09/05/20	ac	Constraint 'h' on bspStart_() assembly parameters
	12/06/21	ac	Shorter bspTaskInit_()
	17/01/22	ac	Add BSP_ATTR_WEAK to bspSystemInit_() and bspHardwareInit_()
	27/01/22	ac	Add bspMainStackCheck()
	08/06/22	ac	Modulo 8 stack alignment in bspTaskInit_()

----------------------------------------------------------------------
*/
/*
----------------------------------------------------------------------
	Copyright (c) 2013-2023 Alain Chebrou - AdAstra-Soft . com
	All rights reserved.

	This file is part of the AdAstra Real Time Kernel distribution.

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions are met:

	- Redistributions of source code must retain the above copyright
	  notice, this list of conditions and the following disclaimer.

	- Redistributions in binary form must reproduce the above copyright
	  notice, this list of conditions and the following disclaimer in the
	  documentation and/or other materials provided with the distribution.

	- Neither the name of the copyright holder nor the
	  names of its contributors may be used to endorse or promote products
	  derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
	DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
	DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
	(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
	ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
----------------------------------------------------------------------
*/

//	To change the system clock configuration, you need only to change the SystemClock_Config.c file

#include	"aa.h"
#include	"aakernel.h"
#include	"aalogmes.h"
#include	"aaprintf.h"

#include	"itm.h"

#if defined USE_HAL_DRIVER
	#include "stm32h7xx_hal.h"
#endif

#if (AA_TICK_STRETCH == 1)
	#include	"timbasic.h"
#endif	// AA_TICK_STRETCH

//-----------------------------------------------------------------------------

#if (__FPU_PRESENT == 1) && (__FPU_USED == 1)
	#define	WITH_FPU		1
#endif

// Define the name of the device managed by this BSP
#if (defined STM32H742xx)
	#define	BSP_DEVICE_NAME	"STM32H742"
#elif (defined STM32H743xx  ||  defined STM32H753xx)
	#define	BSP_DEVICE_NAME	"STM32H743/53"
#elif (defined STM32H750xx)
	#define	BSP_DEVICE_NAME	"STM32H750"
#else
	#define	BSP_DEVICE_NAME	"STM32H7xx"
#endif

//-----------------------------------------------------------------------------

STATIC	uint32_t	bspTickRate ;			// SysTick frequency
		uint32_t	bspTsMhzDiv ;			// Time stamp divider for conversion to microseconds

		void		bspSystemClockInit		(void) ;
static 	void		bspTsInit_				(void) ;
static	void 		bspFpuCheck_			(void) ;

// Startup code support
		void		bspSystemInit_			(void) ;
		void		bspHardwareInit_ 		(void) ;
		void		bspMain_				(void) ;

// Exceptions handlers managed by the BSP
		void		SysTick_Handler 		(void) ;
		void		PendSV_Handler 			(void) __attribute__((naked)) ;

// This BSP defines the SysTick configuration function if tick stretch is not used
#if (AA_TICK_STRETCH == 0)
static	uint32_t	bspSysTick_Config		(uint32_t ticks) ;
#endif

//--------------------------------------------------------------------------------
//	AA_ASSERT support function

void BSP_ATTR_WEAK bspAssertFailed (uint8_t * pFileName, uint32_t line)
{
	(void) pFileName ;
	(void) line ;

	// Check if debug is enabled.
	// If debug is enabled, use bkpt instruction to jump to debugger
	// If debug is not enabled, a bkpt instruction triggers a HardFault exception
	if ((CoreDebug->DHCSR & CoreDebug_DHCSR_C_DEBUGEN_Msk) != 0u)
	{
		// Debug enabled, can use BKPT instruction
		BSP_DEBUG_BKPT () ;
	}

	// No debugger: loop
	bspDisableIrqAll () ;
	while (1)
	{
	}
}

// Support for STMicroelectronics stdperiph, HAL or LL code (stm32_assert.h)
void assert_failed (uint8_t * file, uint32_t line)
{
	bspAssertFailed (file, line) ;
}

//--------------------------------------------------------------------------------
//	Default function for centralized error handling.
//	If the debugger is enabled, it jump to the debugger.
//	It returns only if AA_ERROR_FORCERETURN_FLAG is present,
//	or the error is not fatal and return is not forbidden.

//	If another behavior is necessary the user can redefine its own function

void BSP_ATTR_WEAK bspErrorNotify (uint8_t * pFileName, uint32_t line, uint32_t errorNumber)
{
	(void) pFileName ;
	(void) line ;

	// Check if debug is enabled.
	// If debug is enabled, use bkpt instruction to jump to debugger
	// If debug is not enabled, a bkpt instruction triggers a HardFault exception
	if ((CoreDebug->DHCSR & CoreDebug_DHCSR_C_DEBUGEN_Msk) != 0u)
	{
		// Debug enabled, can use BKPT instruction
		BSP_DEBUG_BKPT () ;
	}

	// If non fatal error, and return allowed, then return
	if ((errorNumber & AA_ERROR_FORCERETURN_FLAG) != 0u  ||
		(errorNumber & (AA_ERROR_FATAL_FLAG | AA_ERROR_NORETURN_FLAG)) == 0u)
	{
		return ;
	}

	// No debugger and can't return: forever loop
	bspDisableIrqAll () ;
	while (1)
	{
	}
}

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
//	Startup code support

//	Very early minimum hardware initialization (right after reset)
//	If you use ITCM, DTCM, backup RAM, D2 SRAM ... you must enable clocks of theses devices here

void	BSP_ATTR_WEAK bspSystemInit_			(void)
{
	// On platform like Kinetis which start with the watchdog enabled : disable it

	// Call the CSMSIS system initialization routine in system_stm32h7xx.c
	// This can configure external RAM before filling the BSS section
	SystemInit () ;

	// AXISRAM, ITCM, DTCM and SRAM4 are allocated (aka clocked) by default.
	// Enable the SRAMs used for code, data or BSS.
	// because the startup code will copy FLASH to it
	#if defined(RCC_AHB2ENR_D2SRAM3EN)
		RCC->AHB2ENR |= (RCC_AHB2ENR_D2SRAM1EN | RCC_AHB2ENR_D2SRAM2EN | RCC_AHB2ENR_D2SRAM3EN);
	#elif defined(RCC_AHB2ENR_D2SRAM2EN)
		RCC->AHB2ENR |= (RCC_AHB2ENR_D2SRAM1EN | RCC_AHB2ENR_D2SRAM2EN);
	#else
		RCC->AHB2ENR |= (RCC_AHB2ENR_AHBSRAM1EN | RCC_AHB2ENR_AHBSRAM2EN);
	#endif /* RCC_AHB2ENR_D2SRAM3EN */

}

//-----------------------------------------------------------------------------
//	This is the application hardware initialization routine,
//	Called early from bspMain_(), right after data and bss init, before C++ constructors.

void	BSP_ATTR_WEAK bspHardwareInit_ (void)
{
	// Enable I and D caches
	SCB_EnableICache();
	SCB_EnableDCache();

	// Configure system clocks
	bspSystemClockInit () ;		// In SystemClock_Config.c

	// Call the CSMSIS system clock routine to store the clock frequency
	// in the SystemCoreClock global RAM location.
	SystemCoreClockUpdate () ;

	// Set interruption priority grouping to be preemptive : no sub-priority
	NVIC_SetPriorityGrouping (7 - __NVIC_PRIO_BITS) ;

	// Enable fault exceptions: Usage, Bus, MPU
	SCB->SHCSR |= SCB_SHCSR_USGFAULTENA_Msk | SCB_SHCSR_BUSFAULTENA_Msk | SCB_SHCSR_MEMFAULTENA_Msk ;
}

//-----------------------------------------------------------------------------
//	Come here right after startup initialization (end of startup):
//	Thread mode, use MSP, privileged
//	Some time ago this function was nacked and noreturn. But if it came back => crash.
//	Now if it comes back, it goes back to the startup where it causes an assert or a loop.

extern uint32_t	_sstack_ ;				// Lower address of main stack

void bspMain_ (void)
{
	// Continue the PCU initializations
	bspHardwareInit_ () ;

	// Initialize the main stack RAM from the bottom to the current main stack pointer
	{
		uint32_t	* pStack = (uint32_t *) & _sstack_ ;
		uint32_t	* pEnd ;
		aaIrqStatus_t itSts ;

		// The MPS must be constant, so disable interrupts
		itSts = bspSaveAndDisableIrq () ;
		asm volatile ("MRS %0, msp" : "=r" (pEnd)) ;
		while (pStack != pEnd)
		{
			* pStack++ = 0u ;
		}
		bspRestoreIrq (itSts) ;
	}

	aaMain_ () ;	// Call kernel entry point
}

//--------------------------------------------------------------------------------
// Initialize hardware used by BSP

void	bspInit_ ()
{
	// Disable interrupts
	bspDisableIrqAll () ;

	// Enable ITM trace
	// Set aaLogMes() output device, user can override this later
	// DEBUG is defined by makefile
	#if (defined DEBUG  &&  BSP_SWO_ENABLE == 1)
	{
		swoInit (1 << BSP_SWO_STIMULUS, SystemCoreClock, BSP_SWO_BITRATE) ;	// Init ITM stimulus
		#if (AA_WITH_LOGMES == 1)
		{
			aaLogMesSetPutChar (0u, swoPut) ;			// Initialize aaLogMes to use SWO
		}
		#endif
	}
	#endif

	bspTickRate = BSP_TICK_RATE ;	// Allows to test tick rate before sysTick is configured

	// Initialize board standard I/O
	bspBoardInit_ () ;

	// Initialize time stamp package
	bspTsInit_ () ;

	// Set PendSV and systick exceptions to lowest priority
	// NVIC_SetPriority use non shifted priority value
	NVIC_SetPriority (PendSV_IRQn,  BSP_MIN_INT_PRIO) ;

	NVIC_SetPriority (SysTick_IRQn, BSP_MIN_INT_PRIO) ;
}

//--------------------------------------------------------------------------------
/*
	The __VFP_FP__ and __SOFTFP__ are macros predefined by the compiler itself depending on the command line options.
	You can check them by adding the compiler options ' -dM -E' (https://stackoverflow.com/questions/2224334/gcc-dump-preprocessor-defines)
	The __VFP_FP__ is defined as long as you specify -mfpu=fpv4-sp-d16 (On M4)
	The __SOFTFP__ is only defined if you specify -mfloat-abi=soft. In all other cases i.e. -mfloat-abi=softfp or mfloat-abi=hard, it is not defined.
	The __FPU_PRESENT is defined in the device header file.

	The difference between -mfloat-abi=softfp or mfloat-abi=hard, is the usage of the floating point registers in the later to pass floating point parameters to functions.
*/

static	void bspFpuCheck_ (void)
{
	uint32_t mvfr0;

	mvfr0 = * (volatile uint32_t *) 0xE000EF40 ;
	switch (mvfr0)
	{
		case 0x10110021: aaPrintf ("FPU-S Single-precision only\n") ; break ;
		case 0x10110221: aaPrintf ("FPU-D Single-precision and Double-precision\n") ; break ;
		default: aaPrintf ("Unknown FPU\n") ;
	}
}

void	bspBanner_ ()
{
	uint32_t 	cpuid ;
	uint32_t	var, rev;
	uint32_t	temp;

	aaPrintf ("%s Edition\n\n", BSP_DEVICE_NAME) ;

	// Display ARM core info
	cpuid = SCB->CPUID ;
	aaPrintf("CPUID %08X ", cpuid) ;

	var = (cpuid & 0x00F00000) >> 20 ;
	rev = (cpuid & 0x0000000F) ;

	if ((cpuid & 0xFF000000) == 0x41000000)
	{
		switch ((cpuid & 0x0000FFF0) >> 4)
		{
			case 0xC20: aaPrintf("Cortex M0") ;  break ;
			case 0xC60: aaPrintf("Cortex M0+") ; break ;
			case 0xC21: aaPrintf("Cortex M1") ;  break ;
			case 0xC23: aaPrintf("Cortex M3") ;  break ;
			case 0xC24: aaPrintf("Cortex M4") ;  break ;
			case 0xC27: aaPrintf("Cortex M7") ;  break ;

			default: aaPrintf("Unknown CORE");
	    }
		aaPrintf(" r%dp%d", var, rev);
	}

	temp  = DBGMCU->IDCODE ;
	aaPrintf (" - DEVID %03X  REV %04X  %u KB FLASH\n", temp & 0xFFF, temp >> 16,
			  * (uint16_t *) FLASHSIZE_BASE) ;

	#if (__FPU_PRESENT == 1u)
		bspFpuCheck_ () ;
	#endif

	temp = bspGetSysClock () ;
	aaPrintf ("System clock  %u Hz\n", temp) ;

	temp = bspGetTickRate () ;
	aaPrintf ("Tick rate     %u Hz\n", temp) ;
}

//-----------------------------------------------------------------------------
//	Configure kernel tick frequency
//	Source clock = HCLK/8 to allow low tick frequency
//	Param: ticks is a count of HCLK tick (not a frequency)

//	WARNING: HCLK/8 not available, see errata document

#if (AA_TICK_STRETCH == 0)

static	uint32_t bspSysTick_Config (uint32_t ticks)
{
//	ticks /= 8 ;
	if ((ticks - 1u) > SysTick_LOAD_RELOAD_Msk)
	{
		AA_ASSERT (0) ;
		return (1u) ;	// Reload value impossible
	}

	SysTick->LOAD  = (uint32_t)(ticks - 1UL) ;			// Set reload register
	SysTick->VAL   = 0UL ;								// Load the SysTick Counter Value
	SysTick->CTRL  =	SysTick_CTRL_TICKINT_Msk   |	// Enable interrupt
						SysTick_CTRL_CLKSOURCE_Msk |	// Divider 8 if 0
						SysTick_CTRL_ENABLE_Msk ;		// Enable SysTick IRQ and SysTick Timer
	return 0u ;	// Function successful
}

#endif

//-----------------------------------------------------------------------------
// Not used, only to avoid linker message when unused code is not removed

uint32_t SysTick_Config(uint32_t ticks)
{
	(void) ticks ;
	return 0u ;
}

//-----------------------------------------------------------------------------
//	Configure kernel tick frequency

BSP_ATTR_USED aaError_t	bspSetTickRate (uint32_t tickHz)
{
	#if (AA_TICK_STRETCH == 1)
	{
		bspTickStretchInit_ (tickHz) ;
		return AA_ENONE ;
	}
	#else
	{
		uint32_t	temp = SystemCoreClock / tickHz ;

		if (bspSysTick_Config (temp) == 0u)
		{
			// SysTick configured without error
			bspTickRate = tickHz ;
			return AA_ENONE ;
		}
		return AA_EFAIL ;
	}
	#endif
}

//-----------------------------------------------------------------------------
//	Get kernel tick frequency (Hz) (Hz)

uint32_t	bspGetTickRate	(void)
{
	return bspTickRate ;
}

//--------------------------------------------------------------------------------
//	Get system core clock frequency (Hz)

uint32_t	bspGetSysClock 		(void)
{
	return SystemCoreClock ;
}

// ----------------------------------------------------------------------------
//	Allows to reset the hardware from software
//	Use NVIC_SystemReset from core_cmXX.h

void BSP_ATTR_WEAK BSP_ATTR_NORETURN bspResetHardware ()
{
	NVIC_SystemReset () ;
	while (1)	// To avoid message about noreturn
	{
	}
}

// ----------------------------------------------------------------------------
//	Allows to get an interrupt priority value.
//	The value is computed relative to the max allowed priority (higher urgency)
//	Offset 0 returns the max allowed priority
//	Offset 1 returns the priority just below max, etc
//	The lowest priority is bounded by BSP_MIN_INT_PRIO

uint32_t	bspIrqPrioMaxMinus		(uint32_t offset)
{
	uint32_t	priority ;

	priority = BSP_MAX_INT_PRIO + offset ;
	if (priority < BSP_MAX_INT_PRIO  ||  priority > BSP_MIN_INT_PRIO)
	{
		priority = BSP_MIN_INT_PRIO ;
	}
	return priority ;
}

// ----------------------------------------------------------------------------
//	Allows to get an interrupt priority value.
//	The value is computed relative to the min allowed priority (lower urgency)
//	Offset 0 returns the min allowed priority
//	Offset 1 returns the priority just above min, etc
//	The highest priority is bounded by BSP_MAX_INT_PRIO

uint32_t	bspIrqPrioMinPlus		(uint32_t offset)
{
	uint32_t	priority ;

	priority = BSP_MIN_INT_PRIO - offset ;
	if (priority < BSP_MAX_INT_PRIO  ||  priority > BSP_MIN_INT_PRIO)
	{
		priority = BSP_MAX_INT_PRIO ;
	}
	return priority ;
}

//--------------------------------------------------------------------------------
//--------------------------------------------------------------------------------
//	In this port the time stamp package use DWT->CYCNT counter
//	For 400 MHz CPU clock, the CYCNT counter loops after about 10.7 seconds

void	bspTsInit_ (void)
{
	/* DWT cycle counter enable. The M7 requires DWT unlocking.*/

	CoreDebug->DEMCR |= CoreDebug_DEMCR_TRCENA_Msk ;

	#if defined __CM7_REV
		DWT->LAR = 0xC5ACCE55U ;
	#endif

	DWT->CTRL |= DWT_CTRL_CYCCNTENA_Msk ;

	// DWT->CYCNT frequency is SystemCoreClock Hz
	// Compute divider to convert delta values to microseconds
	bspTsMhzDiv = SystemCoreClock / 1000000u ;
}

//--------------------------------------------------------------------------------
//	Lets busy wait for delays of some microseconds (min 1 us, 0 is forbidden)
//	Force optimization level to be independent of the general setting
//	BEWARE: At low core frequency the 1us delay may be impossible

#if defined(__CC_ARM)
	#pragma push
	#pragma Otime
#else
	#pragma GCC push_options
	#pragma GCC optimize ("O3")
#endif

void bspDelayUs (uint32_t us)
{
	volatile uint32_t start = DWT->CYCCNT ;
	volatile uint32_t cycles = (bspTsMhzDiv * us) - 42u ;	// Remove some cycles for instructions

	while ((DWT->CYCCNT - start) < cycles) ;
}

#if defined(__CC_ARM)
	#pragma pop
#else
	#pragma GCC pop_options
#endif

//--------------------------------------------------------------------------------
//	Check main stack: return unused words count
//	The main stack is from _sstack_ to _estack_ (which isnext byte above the last stack word)
//	Theses values are provided by the linker

extern uint32_t  _sstack_ ;			// From linker script: lower address of main stack
extern uint32_t  _estack_ ;			// From linker script: address after end of stack

uint32_t	bspMainStackCheck (void)
{
	uint32_t	freeCount = 0 ;
	uint32_t	* pStack = (uint32_t *) & _sstack_ ;

	while (pStack != (uint32_t *) & _estack_)
	{
		if (* pStack != 0u)
		{
			break ;
		}
		freeCount++ ;
		pStack++;
	}

	return freeCount ;
}

//--------------------------------------------------------------------------------
//	Create task switch environment

void	bspTaskInit_ (aaTcb_t * pTcb, aaTaskFunction_t pEntry, uintptr_t arg)
{
	bspStackType_t	* pStack ;

	// Simulate a stack who grows from top to bottom
	pStack = & pTcb->pStack [pTcb->stackSize] ;

	// Modulo 8 stack alignment
	pStack = (bspStackType_t *) (((uintptr_t) pStack) & ~0x07) ;

	// No need of switch environment for idle task
	if (pTcb->taskIndex != 0u)
	{
		// Build the task register context in the stack
#if (1)
		* (--pStack) = 0x01000000u ;			// xPSR : thumb
		* (--pStack) = (uint32_t) pEntry ;		// PC
		* (--pStack) = (uint32_t) aaTaskEnd_ ;	// LR Return value to terminate the thread
		pStack -= 4 ;							// R12 R3 R2 R1
		* (--pStack) = arg ;					// R0
		pStack -= 8 ;							// R11 R10 R9 R8 R7 R6 R5 R4
		* (--pStack) = 0xFFFFFFFDu ;			// LR
		--pStack ;								// R4
#else
		* (--pStack) = 0x01000000u ;			// xPSR : thumb
		* (--pStack) = (uint32_t) pEntry ;		// PC
		* (--pStack) = (uint32_t) aaTaskEnd_ ;	// LR Return value to terminate the thread
		* (--pStack) = 0x0C0Cu ;				// R12
		* (--pStack) = 0x0303u ;				// R3
		* (--pStack) = 0x0202u ;				// R2
		* (--pStack) = 0x0101u ;				// R1
		* (--pStack) = arg ;					// R0

		* (--pStack) = 0x0B0Bu ;				// R11
		* (--pStack) = 0x0A0Au ;				// R10
		* (--pStack) = 0x0909u ;				// R9
		* (--pStack) = 0x0808u ;				// R8
		* (--pStack) = 0x0707u ;				// R7
		* (--pStack) = 0x0606u ;				// R6
		* (--pStack) = 0x0505u ;				// R5
		* (--pStack) = 0x0404u ;				// R4
		* (--pStack) = 0xFFFFFFFDu ;			// LR
		* (--pStack) = 0x4444u ;				// R4
#endif
	}
	pTcb->pTopOfStack = pStack ;	// Pointer to the lower stack word used
}

//-----------------------------------------------------------------------------
//	Switch to first task context, start tick timer, and enter idle task
//	Load PSP with idle task stack pointer and switch to PSP =>  thread mode,  privileged
//	pAaCurrentTcb, which is idle task, was set by aaStart_()

// WARNING: We change the stack pointer in this function
// So the push at the beginning does't match the pop at the end.
// The compiler should not insert prologue/epilogue: declare bspStart_() as naked (GCC only)

#if defined(__CC_ARM)
	__ASM void	bspStart_		(void)
#else
	void BSP_ATTR_NAKED  bspStart_	(void)
#endif
{
	__ASM volatile
	(
		"LDR	R0, =pAaCurrentTcb	\n"		// R0 is the address of pAaCurrentTcb
		"LDR	R0, [R0]			\n"		// R0 is the value of pAaCurrentTcb (@ of TCB)
		"LDR	R0, [R0]			\n"		// R0 is the value of pAaCurrentTcb->pStack
		"MSR	PSP, R0				\n"		// Set PSP
		"MOV	R0, #2				\n"		// 2: Use PSP,  1: unprivileged
		"MSR	CONTROL, R0			\n"		// Set CONTROL to R0 => Now SP is PSP, privileged thread mode
		"ISB						\n"		// Required when changing the stack pointer

											// Reset MSP to high memory
		"MOV	R0, %[VTOR]			\n"		// NVIC VTOR, locate the vector table
		"LDR	R0, [R0]			\n"		// Read VTOR offset
		"LDR	R0, [R0]			\n"		// Read first entry of VTOR: MSP value
		"MSR	MSP, R0				\n"		// Set MSP back to top of RAM

		"MOV	R0, %[TICKRATE]		\n"		// Start SysTick
		"BL		bspSetTickRate		\n"

		"MOV	R0, #0				\n"		// Enable interrupts
		"MSR	BASEPRI, R0			\n"
		"CPSIE	I					\n"

											// SCB->ICSR |= SCB_ICSR_PENDSVSET_Msk ;
		"MOV	R0, %[ICSR]			\n"		// R0 is the address of the SCB->ICSR registers
		"LDR	R1, [R0]			\n"		// R1 is the value of ICSR
		"ORR	R1, R1, %[PENDMSK]	\n"		// R1 |= SCB_ICSR_PENDSVSET_Msk
		"STR	R1, [R0]			\n"		// Write ICSR
		"ISB						\n"		// flush pipeline to ensure exception takes effect before we return from this routine

		"LDR	R0, =aaIdleTask_	\n"
		"BX		R0					\n"		// Jump to idle task (parameter not used)

		:
		:	[ICSR]     "h" (&SCB->ICSR),	// Constraint 'h' to use registers r8..r15
			[VTOR]     "h" (&SCB->VTOR),
			[TICKRATE] "h" (BSP_TICK_RATE),
			[PENDMSK]  "h" (SCB_ICSR_PENDSVSET_Msk)
		:"r0", "r1"
	) ;
}

//--------------------------------------------------------------------------------
//	PendSV exception handler

// WARNING: We change the stack pointer in this function
// So the push at the beginning does't match the pop at the end.
// The compiler should not insert prologue/epilogue: declare PendSV_Handler() as naked (GCC only)

#if defined(__CC_ARM)
	__ASM void	PendSV_Handler (void)
#else
	void	BSP_ATTR_NAKED PendSV_Handler (void)
#endif
{
 	__ASM volatile
	(
		"MRS	R0, PSP				\n"		// Get task SP (the handler is running on MSP)
		"ISB						\n"
		"STMDB	R0!, {R4-R11}		\n"		// Save remaining registers

#if defined WITH_FPU
		"TST	R14, #0x10			\n"		// If the task is using the FPU then push high FPU registers
		"IT		EQ					\n"		// This also triggers the low FPU register push
		"VSTMDBEQ r0!, {S16-S31}	\n"
#endif

		"LDR	R4, =pAaCurrentTcb	\n"		// R4 is the address of pCurrentTcb
		"STMDB	R0!, {R4, LR}		\n"		// Save LR with EXC_RETURN (8 byte aligned)
		"LDR	R2, [R4]			\n"		// R2 is a pointer to current TCB
		"STR	R0, [R2]			\n"		// Save SP in first word of current TCB

		"BL		aaTaskSwitch_		\n"		// Do context switch: set the new value of pAaCurrentTcb

		"LDR	R4, =pAaCurrentTcb	\n"		// R4 is the address of pCurrentTcb
		"LDR	R2, [R4]			\n"		// R2 is a pointer to next TCB
		"LDR	R0, [R2]			\n"		// R0 is SP of pNextTask
		"LDMIA	R0!, {R4, LR}		\n"		// Restore LR with EXC_RETURN

#if defined WITH_FPU
		"TST	R14, #0x10			\n"		// If the task is using the FPU then pop the high FPU registers
		"IT		EQ					\n"		// Low FPU register will be popped by the CPU
		"VLDMIAEQ R0!, {S16-S31}	\n"
#endif

		"LDMIA	R0!, {R4-R11}		\n"		// Restore remaining registers
		"MSR	PSP, R0				\n"		// Restore PSP of pNextTask
		"ISB						\n"
		"BX		LR					\n"		// Return from exception
		:
		:
		:
	) ;
}

//-----------------------------------------------------------------------------
//  SysTick interrupt handler, periodic tick

void	SysTick_Handler (void)
{
	aaIntEnter () ;

	#if defined USE_HAL_DRIVER
	{
		HAL_IncTick () ;
	}
	#endif
	aaTick_ () ;

	aaIntExit () ;
}

//--------------------------------------------------------------------------------
//--------------------------------------------------------------------------------
#if (AA_TICK_STRETCH == 1)

// Define the timer to use for tick stretching
// The ratio between the frequency of the timer and the tick frequency should be at least 2
// So: timer frequency >= 2 x tick frequency
// Use TIM7: input clock 200 MHz with prescaler 50000 => timer clock is 4 Khz.
// To get tick of 1 KHz, the timer divider TKS_DIV is 4
// In timbasic.h: WITH_TIM_7 should be 0, because the interrupt handler is defined in this source

// TIMPRE setting may allows the timer clock frequency to be divided by 2.
// The use of a 32 bit timers (TIM2/TIM5) allows much longer sleep periods.

//	This structure describe the tick stretching state. It is BSP specific.

typedef struct
{
	uint32_t	arrStart ;		// Counter value of next periodic tick
	uint32_t	arrEnd ;		// Counter value of last periodic tick
	uint32_t	tickNext ; 		// Count of tick if the counter reach arrEnd and overflow

} tickStretch_t ;

static	tickStretch_t		tickState ;
static	TIM_TypeDef			* pTim ;			// Pointer to TIM registers

#define	TKS_DIV				4u
#define	TKS_ARR_MAX			16000u				// Max stretch is 16000 periodic tick => 16 seconds if tick is 1 ms
#define	TKS_PRESCALER		50000u				// Timer prescaler value
#define	TKS_FREQUENCY		1000u				// Tick frequency

// Timer descriptor
static const timDesc_t		timDesc =
{
	7u,									// Descriptor identifier
	7u,									// Timer number 1..14
	RCC_APB1LRSTR_TIM7RST_Pos,			// RCC_APBxLRSTR_TIMnRST_Pos
	RCC_APB1LENR_TIM7EN_Pos,			// RCC_APBxLRSTR_TIMnEN_Pos
	TIM_CAP_TYPE_BASIC | TIM_CAP_DMA, 	// Capabilities flags
	TIM7,								// Pointer to TIM registers
	TIM7_IRQn,
	{
		{	0u,		0u,	AA_GPIO_AF_1,	0u },	// Ch1
		{	0u,		0u,	AA_GPIO_AF_1,	0u },	// Ch2
		{	0u,		0u,	AA_GPIO_AF_1,	0u },	// Ch3
		{	0u,		0u,	AA_GPIO_AF_1,	0u },	// Ch4
		{	0u,		0u,	AA_GPIO_AF_1,	0u },	// Ch1N
		{	0u,		0u,	AA_GPIO_AF_1,	0u },	// Ch2N
		{	0u,		0u,	AA_GPIO_AF_1,	0u },	// Ch3N
		{	0u,		0u,	AA_GPIO_AF_1,	0u },	// ETR
		{	0u,		0u,	AA_GPIO_AF_1,	0u },	// BKIN
	},
} ;

//--------------------------------------------------------------------------------

void	TIM7_IRQHandler (void)
{
	aaIntEnter () ;

	if ((pTim->SR & TIM_SR_UIF) != 0u)
	{
		pTim->SR &= ~TIM_SR_UIF ;	// clear UIF flag
	}

	// Configure the timer for 1 tick
	tickState.arrStart	= 0u ;
	tickState.arrEnd	= (BSP_TICK_RATE * TKS_DIV) - 1u ;
	tickState.tickNext	= 1u ;		// Tick count to wait

	aaTick_ () ;

	aaIntExit () ;
}

//--------------------------------------------------------------------------------
//	Configure and start timer with frequency BSP_TICK_RATE

void	bspTickStretchInit_ (uint32_t tickHz)
{
	(void) tickHz  ;

	pTim = timDesc.pTim ;

	// If timer already started, ignore
	if ((pTim->CR1 & TIM_CR1_CEN) != 0u)
	{
		// Already configured
		return ;
	}

	// Configure timer frequency. No preload: we want to set ARR on the fly
	timTimeBaseInit  (& timDesc, TKS_PRESCALER, TKS_FREQUENCY, 0u) ;	// Not preload mode!

	// Configure interrupt without callback (we define our own irq handler)
	timConfigureIntr (& timDesc, BSP_MIN_INT_PRIO, NULL, 0u, TIM_INTR_UPDATE) ;

	tickState.arrStart	= 0u ;
	tickState.arrEnd	= timGetReload (& timDesc) ;
	tickState.tickNext	= 1u ;		// Tick count to wait

	// Clear SLEEPDEEP bit of Cortex System Control Register (No deep sleep)
	SCB->SCR &= ~SCB_SCR_SLEEPDEEP_Msk ;

	timStart (& timDesc) ;
}

//--------------------------------------------------------------------------------

void	bspTickStretch_ (void)
{
	uint32_t	nTick ;			// Periodic tick count to wait
	uint32_t	cnt ;			// Current value of the timer counter
	uint32_t	arrNext ;		// Value of ARR next to cnt: the timer will fire when CNT become arrNext+1 and loop back to 0
	uint32_t	arr ;

	aaCriticalEnter () ;

	// Get tick count to wait
	nTick = aaTickToWait_ () ;
	if (nTick == AA_INFINITE)
	{
		nTick = TKS_ARR_MAX / TKS_DIV ;	// Maximum allowed
	}

#if (AA_WITH_LOGMES == 1)
	// Don't sleep is there is log messages to display
	if (nTick >= 2u  &&  aaLogMesCheck_ () == 0u)
#else
	if (nTick >= 2u)
#endif
	{
		cnt = pTim->CNT ;
		arrNext = ((cnt / TKS_DIV) + 1u) * TKS_DIV - 1u ;
		if ((cnt != arrNext)  &&  (pTim->SR & TIM_SR_UIF) == 0u)
		{
			// cnt is not the last clk of the tick, and timer interrupt is not pending
			arr = arrNext + ((nTick - 1u) * TKS_DIV) ;
			if (arr > TKS_ARR_MAX)
			{
				nTick -= (arr - TKS_ARR_MAX) / TKS_DIV ;
				arr = TKS_ARR_MAX ;
			}
			pTim->ARR = arr ;
			tickState.arrStart = arrNext ;
			tickState.arrEnd   = arr ;
			tickState.tickNext = nTick ;

			bspDisableIrqAll () ;
			aaCriticalExit () ;

			// Sleep
			__DSB () ;	// Armv7-M Architectural requirement
			__WFI () ;

			aaCriticalEnter() ;
			bspEnableIrqAll () ;

			// Awake: Check if a timer interrupt is pending
			cnt = pTim->CNT ;
			if ((cnt == tickState.arrEnd)  ||  ((pTim->SR & TIM_SR_UIF) != 0u))
			{
				// Interrupt pending or in last clock tick
				nTick = tickState.tickNext - 1u ;	// Simulate past ticks, the last one will be managed by IRQ handler
			}
			else
			{
				// No timer pending interrupt: Early awakening
				// We have at least 1 clock tick to reconfigure the timer
				// Compute count of elapsed ticks
				nTick = 0u ;
				if (cnt > tickState.arrStart)
				{
					nTick = ((cnt - tickState.arrStart - 1u) / TKS_DIV) + 1u ;
				}
				// Set timer for 1 tick
				arrNext = ((cnt / TKS_DIV) + 1u) * TKS_DIV - 1u ;
				pTim->ARR = arrNext ;
			}

			// Update ticks counters
			aaUpdateTick_ (nTick) ;
			#if defined USE_HAL_DRIVER
			{
				for (cnt = 0u ; cnt < nTick ; cnt++)
				{
					HAL_IncTick () ;
				}
			}
			#endif

			tickState.tickNext = 1u ;
		}
	}
	else
	{
	}

	aaCriticalExit () ;
}

#endif	// AA_TICK_STRETCH

//--------------------------------------------------------------------------------
//--------------------------------------------------------------------------------
//	Forbidden functions from HAL

void HAL_SuspendTick (void)
{
	AA_ASSERT (0) ;
}

void HAL_ResumeTick (void)
{
	AA_ASSERT (0) ;
}

//--------------------------------------------------------------------------------
