/*
----------------------------------------------------------------------

	AdAstra - Real Time Kernel

	Alain Chebrou

	itm.c		ARM SWO trace for H7

	When		Who	What
	11/07/17	ac	Creation
	28/10/18	ac	Port to STM32H743
	24/01/20	ac	Remove dependency from HAL (PB3 configuration)

	// Switch to SWD without debugger:
	http://infocenter.arm.com/help/index.jsp?topic=/com.arm.doc.ddi0316d/Chdhfbhc.html

----------------------------------------------------------------------
*/
/*
----------------------------------------------------------------------
	Copyright (c) 2013-2023 Alain Chebrou - AdAstra-Soft . com
	All rights reserved.

	This file is part of the AdAstra Real Time Kernel distribution.

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions are met:

	- Redistributions of source code must retain the above copyright
	  notice, this list of conditions and the following disclaimer.

	- Redistributions in binary form must reproduce the above copyright
	  notice, this list of conditions and the following disclaimer in the
	  documentation and/or other materials provided with the distribution.

	- Neither the name of the copyright holder nor the
	  names of its contributors may be used to endorse or promote products
	  derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
	DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
	DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
	(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
	ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
----------------------------------------------------------------------
*/

//	SWO is available only in privileged mode

#include	<stm32h7xx.h>
#include	"aa.h"
#include	"aakernel.h"
#include	"itm.h"
#include	"gpiobasic.h"

//-------------------------------------------------------------------------------------------------

#define SWO_BASE  (0x5C003000UL)
#define SWTF_BASE (0x5C004000UL)

static	uint32_t	bItmAvailable ;		// 1 if SWO enabled (by debugger), else 0

//-------------------------------------------------------------------------------------------------
/*
	Initialize the SWO trace port for debug message printing
	portMask : Stimulus bit mask to be configured
	cpuCoreFreqHz : CPU core clock frequency in Hz
	baudrate : SWO frequency in Hz
*/

// This is very cryptic, and it can not be done from manuals, at least to date!
// See https://community.st.com/s/question/0D50X00009XkhURSAZ/stm32h7-swo-printf-not-working
// https://community.st.com/s/feed/0D50X00009nNBBHSA4

void swoInit (uint32_t portMask, uint32_t cpuCoreFreqHz, uint32_t baudrate)
{
	gpioPinDesc_t		pb3Desc ;

	uint32_t SWOPrescaler = (cpuCoreFreqHz / baudrate) - 1 ; // divider value

	// Enable debug clocks
	DBGMCU->CR = 0x00700000 ;	// D3DBGCKEN D1DBGCKEN TRACECLKEN

	// SWO->LAR unlock
	*((uint32_t *)(SWO_BASE + 0xFB0)) = 0xC5ACCE55 ; // SWO_LAR

	// SWO divider setting, SWO->CODR = PRESCALER[12:0]
	*((__IO uint32_t *)(SWO_BASE + 0x010)) = SWOPrescaler ; // SWO_CODR

	// SWO set the protocol, SWO->SPPR = PPROT[1:0] = NRZ
	*((__IO uint32_t *)(SWO_BASE + 0x0F0)) = 0x00000002 ; // SWO_SPPR

	// UNLOCK FUNNEL, SWTF->LAR unlock
	*((__IO uint32_t *)(SWTF_BASE + 0xFB0)) = 0xC5ACCE55 ; // SWTF_LAR

	// Enable ITM input of SWO trace funnel. SWTF->CTRL bit 0 ENSO = Enable
	*((__IO uint32_t *)(SWTF_BASE + 0x000)) |= 1 ; // SWFT_CTRL


	// Configure ITM
	ITM->LAR	= 0xC5ACCE55 ;							// ITM Lock Access Register, C5ACCE55 enables more write access to Control Register 0xE00 :: 0xFFC
	ITM->TCR	= 0x00010001 ;							// ITM Trace Control Register
	ITM->TPR	= ITM_TPR_PRIVMASK_Msk ;				// ITM Trace Privilege Register. All stimulus ports
	ITM->TER	= portMask ;							// ITM Trace Enable Register. Enabled tracing on stimulus ports. One bit per stimulus port.


	// RCC_AHB4ENR enable GPIOB clock
	RCC->AHB4ENR |= RCC_AHB4ENR_GPIOBEN ;

	// Configure SWO output pin PB3
	pb3Desc.port 	= 'B';
	pb3Desc.pin 	= 3 ;
	pb3Desc.af		= 0 ;
	pb3Desc.flags 	= AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH | AA_GPIO_PUSH_PULL ;
	gpioConfigurePin (& pb3Desc) ;

	// ITM/SWO works only if enabled from debugger.
	// If ITM stimulus 0 is not free, don't try to send data to SWO
	if (ITM->PORT [0].u8 == 1)
	{
		bItmAvailable = 1 ;
	}
}

//-------------------------------------------------------------------------------------------------
//	Returns 1 if SWO is enabled, else 0

uint32_t	swoIsEnabled	(void)
{
	return bItmAvailable ;
}

//-------------------------------------------------------------------------------------------------
//	Sends a character over the SWO stimulus BSP_SWO_STIMULUS defined in bspcfg.h
//	value : Character to be sent

//	Try to aggregate 4 character to send 32 bits words (4 character at the same time)
//	\n character flushes the 32 bits buffer

void swoPut (uintptr_t dev, char value)
{
	static	uint8_t		pos = 0u ;
	static	uint32_t	word = 0u ;

	(void) dev ;		// Unused device

	if (bItmAvailable == 0u)
	{
		return ;	// ITM not available or not initialized
	}

	word |= (uint32_t) value << (pos << 3u) ;
	pos ++ ;
	if (pos == 4u)
	{
		swoSend32 (word, BSP_SWO_STIMULUS) ;
		pos = 0u ;
		word = 0u ;
	}
	else
	{
		if (value == '\n')
		{
			// Flush word
			while (pos != 0u)
			{
				swoSend8 ((uint8_t) word, BSP_SWO_STIMULUS) ;
				word >>= 8u ;
				pos -- ;
			}
			word = 0u ;
		}
	}
}

//-------------------------------------------------------------------------------------------------
//	Send 1 char to stimulus port BSP_SWO_STIMULUS
//	To be used with aaPrintfEx() to print to SWO

void swoPutCharFn (char cc, uintptr_t arg)
{
	(void) arg ;
	swoPut (0u, cc) ;
}

//-------------------------------------------------------------------------------------------------

void swoSend8 (uint8_t value, uint8_t portNo)
{
	aaIrqStatus_t	intState ;

	if (bItmAvailable != 0u)
	{
		while (1)
		{
			intState = bspSaveAndDisableIrq () ;
			if (ITM->PORT [portNo].u8 == 1u)
			{
				ITM->PORT[portNo].u8 = value ;
				bspRestoreIrq (intState) ;
				break ;
			}
			bspRestoreIrq (intState) ;
		}
	}
}

//-------------------------------------------------------------------------------------------------
// Transmitted in little endian (LS byte first)

void swoSend16 (uint16_t value, uint8_t portNo)
{
	aaIrqStatus_t	intState ;

	if (bItmAvailable != 0u)
	{
		while (1)
		{
			intState = bspSaveAndDisableIrq () ;
			if (ITM->PORT [portNo].u16 == 1u)
			{
				ITM->PORT[portNo].u16 = value ;
				bspRestoreIrq (intState) ;
				break ;
			}
			bspRestoreIrq (intState) ;
		}
	}
}

//-------------------------------------------------------------------------------------------------
// Transmitted in little endian (LS byte first)

void swoSend32 (uint32_t value, uint8_t portNo)
{
	aaIrqStatus_t	intState ;

	if (bItmAvailable != 0u)
	{
		while (1)
		{
			intState = bspSaveAndDisableIrq () ;
			if (ITM->PORT [portNo].u32 == 1u)
			{
				ITM->PORT[portNo].u32 = value ;
				bspRestoreIrq (intState) ;
				break ;
			}
			bspRestoreIrq (intState) ;
		}
	}
}

//-------------------------------------------------------------------------------------------------
// Transmit a NUL terminated string (without the NUL) on terminal output

void swoPutStr (char * pStr)
{
	if (bItmAvailable != 0u)
	{
		while (* pStr != 0u)
		{
			swoPut (0u, * pStr++) ;
		}
	}
}

//-------------------------------------------------------------------------------------------------
