/* 
 * The MIT License (MIT)
 *
 * Copyright (c) 2019 Ha Thach (tinyusb.org)
 *               2022 A Chebrou (AdAstra-soft.com)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *
 */

/*
----------------------------------------------------------------------

	AdAstra - Real Time Kernel

	Alain Chebrou

	aaUsb.c		TinyUSB initialization
				AA specific CDC API

	When		Who	What
	30/03/22	ac	Creation

	Which driver for STM32:

	Synopsys driver:	F105/107/F2/F4/F7/H7/L4+
	Fdev (Device only):	F103/F0/F3/L0/L4/G4

	// On H74x / H75x               : use USB2 with integrated FS PHY
	// On H72x / H73x / H7Ax / H7Bx : use USB1 with integrated FS PHY (Only USB1 on these H7)

----------------------------------------------------------------------
*/
#include "aa.h"
#include "aakernel.h"
#include "gpiobasic.h"
#include "tusb.h"
#include "aausb.h"

//--------------------------------------------------------------------------------

#define	USB_INT_PRIORITY	BSP_IRQPRIOMIN_PLUS(2)

// The priority of the USB task must be higher than the one that transmit.
// Otherwise, for a large message, the USB task may not be able to run and process
// the queue of messages supplied by the USB interrupt.
#define	USB_TASK_PRIORITY	AA_PRIO_MAX


//	Convert ms to tick count. Add 999 to round up and avoid a result of 0
#define	MS_TO_TICKS(msec)	(((uint32_t)msec * bspGetTickRate () + 999u) / 1000u)

//--------------------------------------------------------------------------------

void	cdcInit (uint32_t idx) ;

//--------------------------------------------------------------------------------
// Define the USB task stack, to avoid use of dynamic memory
#define	USB_STACK_SIZE	256u
static	bspStackType_t	usbStack [USB_STACK_SIZE] BSP_ATTR_NOINIT ;

static	void	usbTask (uintptr_t arg)
{
	(void) arg ;

	// The sole purpose of the USB task is to process messages supplied by the USB interrupt
	// and placed in the message queue.
	// So the task remains blocked waiting for a message, and tud_task() will never return.
	tud_task() ; 	// tinyusb device task
	AA_ASSERT (0) ;	// Never return
}

//--------------------------------------------------------------------------------
// Blackpill doens't use VBUS sense (B=device) explicitly disable it

#define	VBUS_SENSE		false		// To enable / Disable VBUS sense
#define	ID_PIN			false		// To configure ID pin

// No VBUS sense for G4

static void board_vbus_sense_init (bool bEnable)
{
	if (bEnable == true)
	{
		// Enable VBUS sense (B=device)
		#if (defined STM32F4 || defined STM32L4)
			// F4xx : VBUSAEN = Host mode, VBUSBEN = Device mode
			#if (defined USB_OTG_GCCFG_VBDEN)
				USB_OTG_FS->GCCFG |= USB_OTG_GCCFG_VBDEN ;
			#else
				// F401 F411 F405/15 F407/17 F427/37 F429/39
				USB_OTG_FS->GCCFG &= ~USB_OTG_GCCFG_NOVBUSSENS ;
				USB_OTG_FS->GCCFG |=  USB_OTG_GCCFG_VBUSBSEN ;

			#endif

		#elif (defined STM32H7)
			#if (defined USB2_OTG_FS)
				// For H74x / H75x
				USB2_OTG_FS->GCCFG |= USB_OTG_GCCFG_VBDEN ;
			#else
				// For H72x / H73x
				USB1_OTG_HS->GCCFG |= USB_OTG_GCCFG_VBDEN ;
			#endif
		#endif
	}
	else
	{
		// Disable VBUS sense (B=device)
		#if (defined STM32F4 || defined STM32L4)
			#if (defined USB_OTG_GCCFG_VBDEN)
				// Deactivate HW VBUS Sensing B, and enable override
				USB_OTG_FS->GCCFG   &= ~USB_OTG_GCCFG_VBDEN;
				USB_OTG_FS->GOTGCTL |=  USB_OTG_GOTGCTL_BVALOEN ;
				USB_OTG_FS->GOTGCTL |=  USB_OTG_GOTGCTL_BVALOVAL ;

			#else
				// F401 F411 F405/15 F407/17 F427/37 F429/39
				USB_OTG_FS->GCCFG |=  USB_OTG_GCCFG_NOVBUSSENS ;
				USB_OTG_FS->GCCFG &= ~USB_OTG_GCCFG_VBUSBSEN ;
				USB_OTG_FS->GCCFG &= ~USB_OTG_GCCFG_VBUSASEN ;
			#endif

		#elif (defined STM32H7)
			// Deactivate HW VBUS Sensing B, and enable override
			#if (defined USB2_OTG_FS)
				USB2_OTG_FS->GCCFG   &= ~USB_OTG_GCCFG_VBDEN;
				USB2_OTG_FS->GOTGCTL |= USB_OTG_GOTGCTL_BVALOEN;
				USB2_OTG_FS->GOTGCTL |= USB_OTG_GOTGCTL_BVALOVAL;
			#else
				USB1_OTG_HS->GCCFG   &= ~USB_OTG_GCCFG_VBDEN;
				USB1_OTG_HS->GOTGCTL |= USB_OTG_GOTGCTL_BVALOEN;
				USB1_OTG_HS->GOTGCTL |= USB_OTG_GOTGCTL_BVALOVAL;
			#endif
		#endif
	}
}

//--------------------------------------------------------------------------------
//	The USB clock source should be configured before the call of aaUsbInit()
//	Most of the time this is done in SystemClock_Config()
//	For H7: use OTG_FS

void aaUsbInit (void)
{
	gpioPinDesc_t	io ;

	if (tusb_inited () != 0)
	{
		return ;	// Already initialized
	}

	// USB_OTG_FS GPIO Configuration
	// PA9      ------> USB_OTG_FS_VBUS
	// PA10     ------> USB_OTG_FS_ID
	// PA11     ------> USB_OTG_FS_DM
	// PA12     ------> USB_OTG_FS_DP
#if (! defined STM32G4)
	// On G4 USB pins are not alternate functions, but additional functions:
	// These GPIO are configured by the USB device
	io.port  = AA_GPIO_PORTA ;
	io.pin   = 11u ;
	io.af    = AA_GPIO_AF_10 ;
	io.flags = AA_GPIO_MODE_ALT_PP | AA_GPIO_SPEED_VERYHIGH ;
	gpioConfigurePin (& io) ;

	io.pin   = 12 ;
	gpioConfigurePin (& io) ;
#endif

	// Configure VBUS pin if VBUS sense is enabled
	if (VBUS_SENSE == true)
	{
		io.port  = AA_GPIO_PORTA ;
		io.pin   = 9u ;
		io.af    = AA_GPIO_AF_0 ;
		io.flags = AA_GPIO_MODE_INPUT ;
		gpioConfigurePin (& io) ;
	}

	// Configure ID pin if required
	if (ID_PIN == true)
	{
		io.port  = AA_GPIO_PORTA ;
		io.pin   = 10u ;
		io.af    = AA_GPIO_AF_10 ;
		io.flags = AA_GPIO_MODE_ALT_PP | AA_GPIO_SPEED_HIGH ;
		gpioConfigurePin (& io) ;
	}

#ifdef STM32F412Zx
	// Configure POWER_SWITCH IO pin
	io.port  = AA_GPIO_PORTG ;
	io.pin   = 8u ;
	io.af    = AA_GPIO_AF_0 ;
	io.flags = AA_GPIO_MODE_OUTPUT_OD ;
	gpioConfigurePin (& io) ;
#endif

	// Enable VddUSB for L4.
#if defined PWR_CR2_USV
	PWR->CR2 |=  PWR_CR2_USV ;
#endif

	// USB Peripheral clock enable
	{
		__IO uint32_t tmpreg ;
#if (defined STM32G4)
		RCC->APB1ENR1 |= RCC_APB1ENR1_USBEN ;
		tmpreg = RCC->APB1ENR1 ;
#endif

#if (defined STM32F4 || defined STM32L4)
		RCC->AHB2ENR |= RCC_AHB2ENR_OTGFSEN ;
		tmpreg = RCC->AHB2ENR ;
#endif

#if (defined STM32H7)
		// When the USB regulator is enabled, the VDD33USB voltage detector shall be enabled
//		LL_PWR_EnableUSBVoltageDetector();
		PWR->CR3 |= PWR_CR3_USB33DEN ;

		#if (defined RCC_AHB1ENR_USB2OTGFSEN)
			// For H74x / H75x
			RCC->AHB1ENR |= RCC_AHB1ENR_USB2OTGFSEN ;
			tmpreg = RCC->AHB1ENR ;
		#endif

		#if (defined RCC_AHB1ENR_USB1OTGHSEN)
			// For H72x / H73x / H7Ax / H7Bx
			RCC->AHB1ENR |= RCC_AHB1ENR_USB1OTGHSEN ;
			tmpreg = RCC->AHB1ENR ;
		#endif
#endif
		(void) tmpreg ;
	}

	// Enable / Disable VBUS sense
    board_vbus_sense_init (VBUS_SENSE) ;

    // Peripheral interrupt set priority
#if (defined STM32G4)
	NVIC_SetPriority (USB_HP_IRQn, USB_INT_PRIORITY) ;
	NVIC_SetPriority (USB_LP_IRQn, USB_INT_PRIORITY) ;
	NVIC_SetPriority (USBWakeUp_IRQn, USB_INT_PRIORITY) ;
#else
	#if (defined STM32H7  &&  ! defined USB2_OTG_FS)
		NVIC_SetPriority (OTG_HS_IRQn, USB_INT_PRIORITY) ;	// H7 with only 1 USB
	#else
		NVIC_SetPriority (OTG_FS_IRQn, USB_INT_PRIORITY) ;
	#endif
#endif

	// Initialize all user CDC device environment
	for (uint32_t ii = 0 ; ii < CFG_TUD_CDC ; ii++)
	{
		cdcInit (ii) ;
	}

	// Windows requires a delay after a reset to enumerate correctly
	aaTaskDelay (MS_TO_TICKS (50)) ;

	// Init TinyUSB
	tusb_init () ;

	// Create the USB task after tusb_init()
	aaTaskCreate (
			USB_TASK_PRIORITY,	// Task priority
			"tUSB",				// Task name
			usbTask,			// Entry point,
			0,					// Entry point parameter
			usbStack,			// Stack pointer
			USB_STACK_SIZE,		// Stack size
			AA_FLAG_STACKCHECK,	// Flags
			NULL) ;				// Created task id
}

//--------------------------------------------------------------------------------
//	The interrupt handlers

#if (defined STM32G4)

	void USB_HP_IRQHandler (void)
	{
		aaIntEnter () ;
		tud_int_handler (BOARD_DEVICE_RHPORT_NUM) ;
		aaIntExit () ;
	}

	void USB_LP_IRQHandler (void)
	{
		aaIntEnter () ;
		tud_int_handler (BOARD_DEVICE_RHPORT_NUM) ;
		aaIntExit () ;
	}

	void USBWakeUp_IRQHandler (void)
	{
		aaIntEnter () ;
		tud_int_handler (BOARD_DEVICE_RHPORT_NUM) ;
		aaIntExit () ;
	}

#elif (defined STM32H7)
	#if (defined USB2_OTG_FS)
		// MCU with 2 USB, USB2 is FS
		// Despite being named USB2_OTG
		// OTG_FS is marked as RHPort0 by TinyUSB to be consistent across stm32 port
		void OTG_FS_IRQHandler(void)
		{
			aaIntEnter () ;
			tud_int_handler(0);
			aaIntExit () ;
		}

		// Despite being named USB2_OTG
		// OTG_HS is marked as RHPort1 by TinyUSB to be consistent across stm32 port
		void OTG_HS_IRQHandler(void)
		{
			aaIntEnter () ;
			tud_int_handler(1);
			aaIntExit () ;
		}
	#else
		void OTG_HS_IRQHandler(void)
		{
			aaIntEnter () ;
			tud_int_handler(BOARD_DEVICE_RHPORT_NUM);
			aaIntExit () ;
		}
	#endif

#else
	void OTG_FS_IRQHandler(void)
	{
		aaIntEnter () ;
		tud_int_handler (BOARD_DEVICE_RHPORT_NUM) ;
		aaIntExit () ;
	}

#endif

//--------------------------------------------------------------------------------
//--------------------------------------------------------------------------------
//	CDC standard IO
//--------------------------------------------------------------------------------

typedef struct cdcDevice_s
{
	uint8_t			idx ;		// CDC index
	aaDriverDesc_t	txList ;	// The list of tasks waiting for Tx
	aaDriverDesc_t	rxList ;	// The list of tasks waiting for Rx
	uint32_t		flags ;

} cdcDevice_t ;

static	cdcDevice_t		cdcDev [CFG_TUD_CDC] ;

//--------------------------------------------------------------------------------
//	From Ha Thach: all class driver API take instance ID instead of interface number.
//	Which is zero-index and depends on the order of appearance of the interfaces within configuration.
//	I should rename the ITF to INSTANCE to make it clearer.

void	cdcInit (uint32_t idx)
{
	cdcDev [idx].idx = idx ;
	aaIoDriverDescInit (& cdcDev [idx].txList) ;
	aaIoDriverDescInit (& cdcDev [idx].rxList) ;
}

//--------------------------------------------------------------------------------
// Set or clear an option flag

void			cdcSetOption		(cdctHandle_t hCdc, uint32_t flag, uint32_t bSet)
{
	cdcDevice_t 	* pCdc = (cdcDevice_t *) hCdc ;

	if (bSet == 0u)
	{
		pCdc->flags &= ~flag ;
	}
	else
	{
		pCdc->flags |= flag ;
	}
}

//--------------------------------------------------------------------------------
// To get a CDC handle.
// Idx is CDC index from 0
// Returns NULL on error

cdctHandle_t	cdcGethandle	(uint32_t idx)
{
	if (idx >= CFG_TUD_CDC)
	{
		AA_ASSERT (0) ;
		return NULL ;
	}
	return & cdcDev [idx] ;
}

//--------------------------------------------------------------------------------
//	To register a CDC as standard input/output

void	cdcInitStdDev (aaDevIo_t * pDev, cdctHandle_t hCdc)
{
	pDev->hDev				= hCdc ;
	pDev->getCharPtr		= cdcGetChar ;
	pDev->putCharPtr		= cdcPutChar ;
	pDev->checkGetCharPtr	= cdcCheckGetChar ;
	pDev->checkPutCharPtr	= cdcCheckPutChar ;
}

//--------------------------------------------------------------------------------
// Put one char to device TX FIFO, blocking if FIFO full

void 		cdcPutChar			(cdctHandle_t hCdc, char cc)
{
	cdcDevice_t 	* pCdc = (cdcDevice_t *) hCdc ;
	uint32_t		countTx ;
	uint32_t		flagLF = 0u ;

	while (1)
	{
		aaCriticalEnter () ;
		countTx = tud_cdc_n_write_available (pCdc->idx) ;
		if (countTx == 0u)
		{
			// FIFO is full. aaIoWait() will release the critical section
			(void) aaIoWait (& pCdc->txList, 0u, 0u) ;	// Not ordered, no timeout
		}
		else
		{
			// Not full, try to put the char in the FIFO
			aaCriticalExit () ;
			if (cc == '\n'  &&  flagLF == 0u  &&  ((pCdc->flags & USB_DEV_CRLF) != 0u))
			{
				// Insert \n
				if (0 != tud_cdc_n_write_char (pCdc->idx, '\r'))
				{
					// Success
					flagLF = 1u ;
				}
			}
			else
			{
				if (0 != tud_cdc_n_write (pCdc->idx, & cc, 1))
				{
					// Success
					tud_cdc_n_write_flush (pCdc->idx) ;
					break ;
				}
				else
				{
					// Fail. Try again.
				}
			}
		}
	}
}

//--------------------------------------------------------------------------------
// Get one char from dev, blocking if no char available

char 		cdcGetChar			(cdctHandle_t hCdc)
{
	cdcDevice_t 	* pCdc = (cdcDevice_t *) hCdc ;
	uint32_t		countRx ;
	char			cc ;

	while (1)
	{
		aaCriticalEnter () ;
		countRx = tud_cdc_n_available (pCdc->idx) ;
		if (countRx == 0u)
		{
			// RX FIFO is empty. aaIoWait() will release the critical section
			(void) aaIoWait (& pCdc->rxList, 0u, 0u) ;	// Not ordered, no timeout
		}
		else
		{
			// Not empty, try to get the char from the FIFO
			aaCriticalExit () ;
			if (0 != tud_cdc_n_read (pCdc->idx, & cc, 1))
			{
				break ;	// Success
			}
			else
			{
				// Fail. Try again.
			}
		}
	}
	return cc ;
}

//--------------------------------------------------------------------------------
// Get count of char available to read from dev

uint32_t	cdcCheckGetChar		(cdctHandle_t hCdc)
{
	cdcDevice_t 	* pCdc = (cdcDevice_t *) hCdc ;

	return tud_cdc_n_available (pCdc->idx) ;
}

//--------------------------------------------------------------------------------
// Get count of char free space to write to dev without blocking

uint32_t	cdcCheckPutChar		(cdctHandle_t hCdc)
{
	cdcDevice_t 	* pCdc = (cdcDevice_t *) hCdc ;

	return tud_cdc_n_write_available (pCdc->idx) ;
}

//--------------------------------------------------------------------------------
//	Send a block of data

void	cdcWrite (cdctHandle_t hCdc, uint8_t * pData, uint32_t count, uint32_t bFlush)
{
	cdcDevice_t 	* pCdc = (cdcDevice_t *) hCdc ;
	uint32_t		n ;

	while (count != 0u)
	{
		n = tud_cdc_n_write (pCdc->idx, pData, count) ;
		count -= n ;
		pData += n ;
	}

	if (bFlush != 0u)
	{
		tud_cdc_n_write_flush (pCdc->idx) ;
// For performance test only: wait until the FIFO is empty
//while (tud_cdc_n_write_available (pCdc->itf) != CFG_TUD_CDC_TX_BUFSIZE) {}
	}

}

//--------------------------------------------------------------------------------
//	Send a 0 terminated string

void	cdcPutString (cdctHandle_t hCdc, char * pStr)
{
	cdcWrite (hCdc, (uint8_t *) pStr, aaSTRLEN (pStr), 1u) ;
}

//--------------------------------------------------------------------------------
//	End of transmission, there may be space in the TX FIFO.
//	If a task is pending for TX, wake it up.

void tud_cdc_tx_complete_cb (uint8_t itf)
{
	(void) itf ;
	aaCriticalEnter () ;
	if (aaIsIoWaitingTask (& cdcDev [itf].txList) != 0u)
	{
		// A thread is waiting for TX space, awake it
		(void) aaIoResume (& cdcDev [itf].txList) ;
	}
	aaCriticalExit () ;
}

//--------------------------------------------------------------------------------
//	New data received.
//	If a task is pending for RX, wake it up.

void tud_cdc_rx_cb (uint8_t itf)
{
	(void) itf ;
	aaCriticalEnter () ;
	if (aaIsIoWaitingTask (& cdcDev [itf].rxList) != 0u)
	{
		// A thread is waiting for RX data, awake it
		(void) aaIoResume (& cdcDev [itf].rxList) ;
	}
	aaCriticalExit () ;
}

//--------------------------------------------------------------------------------
// To detect connect / disconnect
/*
	When the cable is plugged: 		mount
	When the cable is unplugged:	suspend (no cdc state)
	When putty is connected:		cdc idx dtr:1
	When putty is disconnected:		cdc idx dtr:0
*/
/*
void tud_cdc_line_state_cb (uint8_t itf, bool dtr, bool rts)
{
	aaPrintf ("cdc %u dtr:%u rts:%u\n", itf, dtr, rts) ;
}

void tud_umount_cb(void)
{
	aaPrintf ("USB umount\n") ;
}

void tud_mount_cb(void)
{
	aaPrintf ("USB mount\n") ;
}

// Invoked when usb bus is suspended
// Within 7ms, device must draw an average of current less than 2.5 mA from bus
TU_ATTR_WEAK void tud_suspend_cb(bool remote_wakeup_en)
{
	aaPrintf ("USB suspend %u\n", remote_wakeup_en) ;
}

// Invoked when usb bus is resumed
TU_ATTR_WEAK void tud_resume_cb(void)
{
	aaPrintf ("USB resume\n") ;
}
*/

//--------------------------------------------------------------------------------
//--------------------------------------------------------------------------------
//	How to know if the PC is connected:
//	https://github.com/hathach/tinyusb/discussions/597
/*
bool was_mounted = false;

void tud_umount_cb(void)
{
	was_mounted  = true;
}

void loop()
{
    if (was_mounted && tud_ready ())
	{
        // PC is connected
    }
	else
	{
		was_mounted = false ;
        // PC is NOT connected
    }
}
*/

// Vendor class without FIFO
// https://github.com/hathach/tinyusb/pull/866
