/*
----------------------------------------------------------------------

	AdAstra - Real Time Kernel

	Alain Chebrou

	dacbasic.c	DAC1 management with DMA (1 DAC with 2 channels)

	When		Who	What
	05/05/18	ac	Creation for STM32F4
	14/03/20	ac	Port to STM37H7xx
	04/06/22	ac	Add VREFBUF setting
	12/08/22	ac	dacCap_t and dacIoDesc_t made public so user can write more optimized code

----------------------------------------------------------------------
*/
/*
----------------------------------------------------------------------
	Copyright (c) 2013-2023 Alain Chebrou - AdAstra-Soft . com
	All rights reserved.

	This file is part of the AdAstra Real Time Kernel distribution.

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions are met:

	- Redistributions of source code must retain the above copyright
	  notice, this list of conditions and the following disclaimer.

	- Redistributions in binary form must reproduce the above copyright
	  notice, this list of conditions and the following disclaimer in the
	  documentation and/or other materials provided with the distribution.

	- Neither the name of the copyright holder nor the
	  names of its contributors may be used to endorse or promote products
	  derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
	DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
	DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
	(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
	ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
----------------------------------------------------------------------
*/

#include	"aa.h"
#include	"aakernel.h"

#include	"dacbasic.h"
#include	"stm32h7xx_ll_rcc.h"
#include	"stm32h7xx_ll_bus.h"

// This function is defined in stm32h7xx_ll_rcc.c but not in stm32h7xx_ll_rcc.h. Sigh...
uint32_t	RCC_GetPCLK1ClockFreq (uint32_t HCLK) ;

//----------------------------------------------------------------------

// Include the proper DAC descriptors
#if (WITH_DAC_X == 1)

	// All H7 have a DAC
	#include "dacH7xx.h"

	// Count of DAC devices in dacDescArray
	STATIC	const uint32_t	dacArrayCount = sizeof (dacDescArray) / sizeof (dacDesc_t) ;

#endif	// WITH_DAC_X

//----------------------------------------------------------------------
//	A descriptor for the information necessary to manage each DAC

#if defined (DAC1)
const	dacCap_t	capDac1 =
	{
		& RCC->APB1LENR,
		& RCC->APB1LRSTR,
		RCC_APB1LENR_DAC12EN,
		TIM6_DAC_IRQn,
		{
			LL_DMAMUX1_REQ_DAC1_CH1,
			LL_DMAMUX1_REQ_DAC1_CH2
		},
		{
			{	'A',	4u,	AA_GPIO_AF_0,	AA_GPIO_MODE_ANALOG },		// Channel 1
			{	'A',	5u,	AA_GPIO_AF_0,	AA_GPIO_MODE_ANALOG },		// channel 2
		}
	} ;
#endif	// DAC1

#if (WITH_DAC_2 == 1  &&  defined (DAC2))
	// WARNING: On APB4, use BDMA and DMAMUX2 => not supported
	dacCap_t	capDac2 =
	{
		& RCC->APB4ENR,
		& RCC->APB4RSTR,
		RCC_APB4ENR_DAC2EN,
		{ LL_DMAMUX2_REQ_DAC2_CH1, 0u },
		DAC2_IRQn,
		{
			{	'0',	4u,	AA_GPIO_AF_0,	AA_GPIO_MODE_ANALOG },		// Channel 1
			{	'0',	5u,	AA_GPIO_AF_0,	AA_GPIO_MODE_ANALOG },		// channel 2
		}
	} ;
#endif	// DAC2


STATIC	dacIoDesc_t		ioDesc1 ;	// IO Descriptor for DAC1

//----------------------------------------------------------------------
//	Use id to get a DAC descriptor from dacDescArray

dacStatus_t		dacGetDesc (uint32_t id, const dacDesc_t ** ppDacDesc)
{
	#if (WITH_DAC_X == 1)
	{
		uint32_t	ii ;

		// Find DAC descriptor from id
		for (ii = 0u ; ii < dacArrayCount ; ii++)
		{
			if (dacDescArray [ii].id == id)
			{
				* ppDacDesc = & dacDescArray [ii] ;
				return DAC_ENONE ;
			}
		}
	}
	#else
	{
		(void) id ;
		(void) ppDacDesc ;
	}
	#endif

	return DAC_EFAIL ;	// No DAC supported
}
//----------------------------------------------------------------------
//----------------------------------------------------------------------
//	Initialize DAC common parameters
//	Flags is DAC_DUAL or a combination of DAC_MONO1 and DAC_MONO2

dacStatus_t 	dacInit			(const dacDesc_t * pDacDesc, uint32_t flags, dacHandle_t * pHdl)
{
	dacIoDesc_t		* pIoDesc ;

	if (pDacDesc->pDac != DAC1)
	{
		return DAC_EARG ;
	}
	pIoDesc = & ioDesc1 ;		// At the moment only DAC1 is managed

	// Set DAC IO descriptor
	pIoDesc->pDacDesc	= pDacDesc ;
	pIoDesc->pDacCap	= & capDac1 ;
	pIoDesc->flags		= flags ;

	// Enable the peripheral clock for DAC
	* pIoDesc->pDacCap->rccENR |= pIoDesc->pDacCap->periphMask ;

	// Reset peripheral => all configuration registers to reset values
	* pIoDesc->pDacCap->rccRSTR |=  pIoDesc->pDacCap->periphMask ;
	* pIoDesc->pDacCap->rccRSTR &= ~pIoDesc->pDacCap->periphMask ;

	//	Configure output pins
	if ((flags & (DAC_MONO1 | DAC_DUAL)) != 0u)
	{
		gpioConfigurePin (& pIoDesc->pDacCap->dacOut [0]) ;
	}
	if ((flags & (DAC_MONO2 | DAC_DUAL)) != 0u)
	{
		gpioConfigurePin (& pIoDesc->pDacCap->dacOut [1]) ;
	}

	* pHdl = (dacHandle_t *) pIoDesc ;
	return DAC_ENONE ;
}

//----------------------------------------------------------------------
//	Initialize one channel parameters.
//	flags is a combination of:
//		DAC_OUTBUFFDIS / DAC_OUTBUFFEN
//		DAC_ONCHIPDIS  / DAC_ONSHIPEN
//		DAC_HOLDEN     / DAC_HOLDDIS
//		DAC_BITS8      / DAC_BITS12
//		DAC_ALIGNLEFT  / DAC_ALIGNRIGHT

//	trigSrc is one of LL_DAC_TRIG_xxx:
//		DAC_TRIGGER_NONE when no trigger
//		LL_DAC_TRIG_SOFTWARE
//		LL_DAC_TRIG_EXT_TIM1_TRGO
//		LL_DAC_TRIG_EXT_TIM2_TRGO
//		LL_DAC_TRIG_EXT_TIM4_TRGO
//		LL_DAC_TRIG_EXT_TIM5_TRGO
//		LL_DAC_TRIG_EXT_TIM6_TRGO
//		LL_DAC_TRIG_EXT_TIM7_TRGO
//		LL_DAC_TRIG_EXT_TIM8_TRGO
//		LL_DAC_TRIG_EXT_TIM15_TRGO
//		LL_DAC_TRIG_EXT_EXTI_LINE9
//		LL_DAC_TRIG_EXT_HRTIM_TRGO1
//		LL_DAC_TRIG_EXT_HRTIM_TRGO2
//		LL_DAC_TRIG_EXT_LPTIM1_OUT
//		LL_DAC_TRIG_EXT_LPTIM2_OUT
//		LL_DAC_TRIG_EXT_LPTIM3_OUT
//		LL_DAC_TRIG_EXT_EXTI_LINE9
//		LL_DAC_TRIG_EXT_TIM23_TRGO
//		LL_DAC_TRIG_EXT_TIM24_TRGO
//		All parameter not available on all devices.

dacStatus_t 	dacSetChannel		(dacHandle_t hdl, uint32_t channel, uint32_t flags, uint32_t trigSrc)
{
	dacIoDesc_t		* pIoDesc = (dacIoDesc_t *) hdl ;
	DAC_TypeDef		* pDac = pIoDesc->pDacDesc->pDac ;
	uint32_t		* pFlagsChan ;
	uint32_t		* pDataReg ;

	// Memorize configuration
	if (channel == DAC_CHAN1)
	{
		pFlagsChan = & pIoDesc->flagsChan [0] ;
	}
	else if (channel == DAC_CHAN2)
	{
		pFlagsChan =  & pIoDesc->flagsChan [1] ;
	}
	else
	{
		return DAC_ENONE ;
	}

	* pFlagsChan = pIoDesc->flags | flags ;

	// Compute DAC data register to use
	pDataReg = & pIoDesc->dataReg [channel-1u] ;
	switch (* pFlagsChan & (DAC_DUAL | DAC_BITS8))
	{
		case (DAC_BITS8):
			* pDataReg = (uint32_t)  ((channel == DAC_CHAN1) ? & pDac->DHR8R1 : & pDac->DHR8R2) ;
			break ;

		case (DAC_BITS12):
			* pDataReg = (uint32_t)  ((channel == DAC_CHAN1) ? & pDac->DHR12R1 : & pDac->DHR12R2) ;
			break ;

		case (DAC_DUAL | DAC_BITS8):
			* pDataReg = (uint32_t) & pDac->DHR8RD ;
			break ;

		default:	// (DAC_DUAL | DAC_BITS12)
			* pDataReg = (uint32_t) & pDac->DHR12RD ;
			break ;
	}
	if ((* pFlagsChan & DAC_ALIGNLEFT) != 0u)
	{
		* pDataReg += 4u ;
	}

	// Set channel mode: output buffer, connection to on chip peripherals, hold mode
	// Set trigger source
	pDac->DHR12RD = 0u ;			// See H7Ax/H7Bx errata document
	if (channel == DAC_CHAN1)
	{
		pDac->MCR = (pDac->MCR & ~DAC_MCR_MODE1) | ((flags & DAC_MODEMASK) >> 8u) ;
		pDac->CR  = (pDac->CR  & ~DAC_CR_TSEL1)  | (trigSrc & DAC_CR_TSEL1) ;
		if (trigSrc == DAC_TRIGGER_NONE)
		{
			pIoDesc->flagsChan [0] |= DAC_TRIGGER_NONE ;
		}
	}
	else
	{
		pDac->MCR = (pDac->MCR & ~DAC_MCR_MODE2) | ((flags & DAC_MODEMASK) << 8u) ;
		pDac->CR  = (pDac->CR  & ~DAC_CR_TSEL2)  | ((trigSrc & DAC_CR_TSEL1) << 16) ;
		if (trigSrc == DAC_TRIGGER_NONE)
		{
			pIoDesc->flagsChan [1] |= DAC_TRIGGER_NONE ;
		}
	}

	return DAC_ENONE ;
}

//----------------------------------------------------------------------
//	Set the DAC callback for one independent channel or the dual channel

// This callback is used by DMA interrupt, and is used to call the DAC user interrupt callback

static	void 	dacDmaCallback		(uint32_t event, uint32_t source, uintptr_t arg)
{
	dacIoDesc_t		* pIoDesc = (dacIoDesc_t *) arg ;

	// Find the channel from the DMA num / stream num
	if ((pIoDesc->pDacDesc->dma1.dmaNum == (source >> 16u)) &&
		(DMA_GETSTREAM (& pIoDesc->pDacDesc->dma1) == (source & 0x00FFu)))
	{
		// Channel 1
		if (pIoDesc->cb [0] != NULL)
		{
			(* pIoDesc->cb [0]) (pIoDesc, event, DAC_CHAN1, pIoDesc->cbArg [0]) ;
		}
	}
	else
	{
		// Channel 2
		if (pIoDesc->cb [1] != NULL)
		{
			(* pIoDesc->cb [1]) (pIoDesc, event, DAC_CHAN2, pIoDesc->cbArg [1]) ;
		}
	}
}


dacStatus_t 	dacSetCallback		(dacHandle_t hdl, uint32_t channel, dacCallback_t cb, uintptr_t arg)
{
	dacIoDesc_t		* pIoDesc = (dacIoDesc_t *) hdl ;

	if (channel == DAC_CHAN1)
	{
		// Register the user callback
		pIoDesc->cb[0]     = cb ;
		pIoDesc->cbArg [0] = arg ;

		// Configure the DMA callback
		if (pIoDesc->pDacDesc->dma1.dmaNum != DMA_NONE)
		{
			dmaSetCallback	(& pIoDesc->pDacDesc->dma1, dacDmaCallback, (uintptr_t) hdl) ;
		}
	}
	else if (channel == DAC_CHAN2)
	{
		// Register the user callback
		pIoDesc->cb[1]     = cb ;
		pIoDesc->cbArg [1] = arg ;

		// Configure the DMA callback
		if (pIoDesc->pDacDesc->dma2.dmaNum != DMA_NONE)
		{
			dmaSetCallback	(& pIoDesc->pDacDesc->dma2, dacDmaCallback, (uintptr_t) hdl) ;
		}
	}
	else
	{
		return DAC_EARG ;
	}

	return DAC_ENONE ;
}

//----------------------------------------------------------------------
//	Configure DMA
//	channel : one of DAC_CHAN1, DAC_CHAN2
//	flags   : one of DAC_DMA_xxx and optional DAC_DMA_NOIT
//	length  : count of items in DMA buffers

dacStatus_t 	dacDmaInit		(dacHandle_t hdl, uint32_t channel, uint32_t flags, uint32_t length)
{
	dacIoDesc_t			* pIoDesc = (dacIoDesc_t *) hdl ;
	const dmaDesc_t 	* pDmaDesc ;
	dma_t				* pDma ;
	DMA_Stream_TypeDef	* pStream ;
	uint32_t			streamNum ;
	DAC_TypeDef			* pDac = pIoDesc->pDacDesc->pDac ;
	uint32_t			reg ;
	uint32_t			dmaSize ;
	uint32_t			* pFlagsChan ;

	if (channel == DAC_CHAN1)
	{
		pDmaDesc   = & pIoDesc->pDacDesc->dma1 ;
		pFlagsChan = & pIoDesc->flagsChan [0] ;
	}
	else if (channel == DAC_CHAN2)
	{
		pDmaDesc   = & pIoDesc->pDacDesc->dma2 ;
		pFlagsChan = & pIoDesc->flagsChan [1] ;
	}
	else
	{
		return DAC_EARG ;
	}

	if (pDmaDesc->dmaNum == DMA_NONE)
	{
		AA_ASSERT (0) ;
		return DAC_ESTATE ;		// No DMA descriptor
	}

	* pFlagsChan |= flags & DAC_DMA_MASK ;
	pDma		= DMA_GETDMA (pDmaDesc->dmaNum) ;
	streamNum	= DMA_GETSTREAM (pDmaDesc) ;
	pStream		= dmaStreamAddressRaw (pDma, streamNum) ;

	// Enable DMA clock
	dmaEnableClock (pDma) ;

	// Reset stream
	dmaStreamResetRaw (pDma, streamNum) ;

	// Build DAC Control Register value
	// Compute Peripheral and memory data size
	switch (* pFlagsChan & (DAC_DUAL | DAC_BITS8))
	{
		case (DAC_BITS8):
			reg = LL_DMA_PDATAALIGN_BYTE | LL_DMA_MDATAALIGN_BYTE ;
			dmaSize = 1u ;
			break ;

		case (DAC_BITS12):
			reg = LL_DMA_PDATAALIGN_HALFWORD | LL_DMA_MDATAALIGN_HALFWORD ;
			dmaSize = 2u ;
			break ;

		case (DAC_DUAL | DAC_BITS8):
			reg = LL_DMA_PDATAALIGN_HALFWORD | LL_DMA_MDATAALIGN_HALFWORD ;
			dmaSize = 2u ;
			break ;

		default:	// (DAC_DUAL | DAC_BITS12)
			reg = LL_DMA_PDATAALIGN_WORD | LL_DMA_MDATAALIGN_WORD ;
			dmaSize = 4u ;
			break ;
	}

	// Memorize DMA transfer size
	pIoDesc->dmaLength [channel-1u] = length ;
	pIoDesc->dmaSize   [channel-1u] = dmaSize * length ;

	reg |=	(((flags & DAC_DMA_CIRCULAR) != 0u) ? LL_DMA_MODE_CIRCULAR : LL_DMA_MODE_NORMAL) |
			LL_DMA_DIRECTION_MEMORY_TO_PERIPH 	|
			LL_DMA_PERIPH_NOINCREMENT 			|
			LL_DMA_MEMORY_INCREMENT   			|
			LL_DMA_PRIORITY_HIGH ;

	// Set double buffer if required
	if ((flags & DAC_DMA_DOUBLE) != 0u)
	{
		reg |= DMA_SxCR_DBM ;
	}

	pStream->CR &= ~(DMA_SxCR_DIR | DMA_SxCR_CIRC | DMA_SxCR_PINC | DMA_SxCR_MINC | DMA_SxCR_PSIZE | DMA_SxCR_MSIZE | DMA_SxCR_PL | DMA_SxCR_PFCTRL | DMA_SxCR_DBM) ;
	pStream->CR |= reg ;

	// Select the DMA transfer request for the channel
	((dmaMux_t *) DMAMUX1)->CCR [streamNum + ((pDma == (dma_t *)DMA1) ? 0u : 8u)] = pIoDesc->pDacCap->chanReqNum [channel-1u] ;

	// Set DMA transfer addresses of peripheral
	pStream->PAR = pIoDesc->dataReg [channel-1u] ;

	// Configure interrupts
	if ((flags & DAC_DMA_NOIT) == 0u)
	{
		//	Configure NVIC for DMA transfer complete/error interrupts
		dmaSetIrq (pDmaDesc) ;

		// Enable DMA transfer interrupts
		pStream->CR |= DMA_SxCR_TCIE ;	// Enable DMA IT transfer complete
		pStream->CR |= DMA_SxCR_TEIE ;	// Enable DMA IT error

		if ((flags & DAC_DMA_CIRCULAR) != 0u)
		{
			pStream->CR |= DMA_SxCR_HTIE ;	// Enable DMA IT half transfer complete
		}

		// Configure NVIC to enable DAC interruptions
		NVIC_SetPriority (pIoDesc->pDacCap->irqVector, DAC_INT_PRIORITY) ;
		NVIC_EnableIRQ   (pIoDesc->pDacCap->irqVector) ;

		// Enable DAC channel underrun interrupt
		pDac->CR |= ((flags & DAC_CHAN1) != 0u) ? LL_DAC_IT_DMAUDRIE1 : LL_DAC_IT_DMAUDRIE2 ;
	}
	else
	{
		// Disable DMA interrupts
		pStream->CR &= ~DMA_SxCR_TCIE ;	// Disable DMA IT complete
		pStream->CR &= ~DMA_SxCR_TEIE ;	// Disable DMA IT error

		// Disable DAC channel underrun interrupt
		pDac->CR &= ((flags & DAC_CHAN1) != 0u) ? ~LL_DAC_IT_DMAUDRIE1 : ~LL_DAC_IT_DMAUDRIE2 ;
	}

	// Enable DAC channel DMA request
	pDac->CR |= (channel == DAC_CHAN1) ? DAC_CR_DMAEN1 : DAC_CR_DMAEN2 ;

	 return DAC_ENONE ;
}

//----------------------------------------------------------------------
//	Set DMA buffers
//	If DAC_DMA_CIRCULAR or DAC_DMA_SINGLE only buffer 1 is used. To be called before dacStart()
//	If DAC_DMA_DOUBLE and called before dacStart(): buffer 1 and buffer 2 are used
//	If DAC_DMA_DOUBLE and called after dacStart(): only buffer 1 is used

// BEWARE: DMA buffers have alignment and size constraints (multiple of cache line size, 32 bytes)
// This function use cache maintenance, see:
// https://community.st.com/s/article/FAQ-DMA-is-not-working-on-STM32H7-devices

dacStatus_t 	dacDmaBuffer	(dacHandle_t hdl, uint32_t channel, const void * buffer1, const void * buffer2)
{
	dacIoDesc_t			* pIoDesc = (dacIoDesc_t *) hdl ;
	const dmaDesc_t 	* pDmaDesc ;
	dma_t				* pDma ;
	DMA_Stream_TypeDef	* pStream ;
	DAC_TypeDef			* pDac ;
	uint32_t			flagsChan ;

	// Find the DMA configuration according to the DAC channel
	if (channel == DAC_CHAN1)
	{
		pDmaDesc  = & pIoDesc->pDacDesc->dma1 ;
		flagsChan = pIoDesc->flagsChan [0] ;
	}
	else if (channel == DAC_CHAN2)
	{
		pDmaDesc  = & pIoDesc->pDacDesc->dma2 ;
		flagsChan = pIoDesc->flagsChan [1] ;
	}
	else
	{
		return DAC_EARG ;
	}
	if (pDmaDesc->dmaNum == DMA_NONE)
	{
		AA_ASSERT (0) ;
		return DAC_ESTATE ;
	}
	pDma		= DMA_GETDMA (pDmaDesc->dmaNum) ;
	pStream		= dmaStreamAddressRaw (pDma, DMA_GETSTREAM (pDmaDesc)) ;

	// Clean D-cache for buffer1
	// Make sure the address is 32-byte aligned and add 32-bytes to length, in case it overlaps cacheline
	SCB_CleanDCache_by_Addr ((uint32_t *) (((uint32_t) buffer1) & ~(uint32_t) 0x1Fu), (int32_t) pIoDesc->dmaSize[channel - 1u] + 32) ;

	// Set DMA addresses
	pDac = pIoDesc->pDacDesc->pDac ;
	if ((pDac->CR & DMA_SxCR_EN) == 0u)
	{
		// DAC not enabled: initialize DMA buffers
		pStream->M0AR = (uint32_t) buffer1 ;

		if ((flagsChan & DAC_DMA_DOUBLE) != 0u)
		{
			// Also set second buffer address
			// Clean D-cache for buffer2
			// Make sure the address is 32-byte aligned and add 32-bytes to length, in case it overlaps cacheline
			SCB_CleanDCache_by_Addr ((uint32_t *) (((uint32_t) buffer2) & ~(uint32_t) 0x1F), (int32_t) pIoDesc->dmaSize[channel - 1u] + 32) ;

			pStream->M1AR = (uint32_t) buffer2 ;
		}
	}
	else
	{
		// DAC enabled : update memory DMA buffer not in use (double buffer only)
		if ((pStream->CR & DMA_SxCR_CT) == 0u)
		{
			// Current target mem is M0AR, so update M1AR
			pStream->M1AR = (uint32_t) buffer1 ;
		}
		else
		{
			// Current target mem is M1AR, so update M0AR
			pStream->M0AR = (uint32_t) buffer1 ;
		}
	}

	return DAC_ENONE ;
}

//----------------------------------------------------------------------
//	Start DAC generation on specified channel

dacStatus_t 	dacStart		(dacHandle_t hdl, uint32_t channel)
{
	dacIoDesc_t			* pIoDesc = (dacIoDesc_t *) hdl ;
	const dmaDesc_t 	* pDmaDesc ;
	DAC_TypeDef			* pDac ;
	uint32_t			shift ;

	// Find the DMA configuration according to the channel
	if (channel == DAC_CHAN1)
	{
		shift = 0u ;
		pDmaDesc = & pIoDesc->pDacDesc->dma1 ;
	}
	else if (channel == DAC_CHAN2)
	{
		shift = 16u ;
		pDmaDesc = & pIoDesc->pDacDesc->dma2 ;
	}
	else
	{
		return DAC_EARG ;
	}

	// Enable the DMA transfer
	if (pDmaDesc->dmaNum != DMA_NONE)
	{
		DMA_Stream_TypeDef	* pStream ;

		pStream	= dmaStreamAddress (pDmaDesc) ;
		pStream->CR &= ~DMA_SxCR_EN ;		// Disable  DMA stream if not already stopped
		pStream->NDTR = pIoDesc->dmaLength [channel-1u] ;

		// Clear DMA interrupt flags
		dmaClearFlags (pDmaDesc, DMA_FLAG_ALLIF) ;

		pStream->CR |=  DMA_SxCR_EN ;		// Enable DMA stream
	}

	pDac = pIoDesc->pDacDesc->pDac ;

	// Clear DAC channel overrun flag
	pDac->SR = (channel == DAC_CHAN1) ? DAC_SR_DMAUDR1 : DAC_SR_DMAUDR2 ;

	// Enable DAC channel
	pDac->CR |= DAC_CR_EN1 << shift ;
	bspDelayUs (7u) ;

	// Enable DAC channel trigger
	if (channel == DAC_CHAN1)
	{
		if ((pIoDesc->flagsChan [0] & DAC_TRIGGER_NONE) == 0u)
		{
			pDac->CR |= DAC_CR_TEN1 ;
		}
	}
	else
	{
		if ((pIoDesc->flagsChan [1] & DAC_TRIGGER_NONE) == 0u)
		{
			pDac->CR |= DAC_CR_TEN2 ;
		}
	}

	return DAC_ENONE ;
}

//----------------------------------------------------------------------
//	Stop DAC generation on specified channel

dacStatus_t 	dacStop		(dacHandle_t hdl, uint32_t channel)
{
	dacIoDesc_t			* pIoDesc = (dacIoDesc_t *) hdl ;
	const dmaDesc_t 	* pDmaDesc ;
	DAC_TypeDef			* pDac ;
	uint32_t			shift ;

	// Find the DMA configuration according to the channel
	if (channel == DAC_CHAN1)
	{
		shift = 0u ;
		pDmaDesc = & pIoDesc->pDacDesc->dma1 ;
	}
	else if (channel == DAC_CHAN2)
	{
		shift = 16u ;
		pDmaDesc = & pIoDesc->pDacDesc->dma2 ;
	}
	else
	{
		return DAC_EARG ;
	}

	// First switch off the DAC
	pDac = pIoDesc->pDacDesc->pDac ;
	pDac->CR &= ~DAC_CR_TEN1 << shift ;		// DAC Disable Trigger
	pDac->CR &= ~DAC_CR_EN1 << shift ;		// DAC Disable

	// Then switch off the DMA
	if(pDmaDesc->dmaNum != DMA_NONE)
	{
		DMA_Stream_TypeDef	* pStream ;

		pStream	= dmaStreamAddress (pDmaDesc) ;
		pStream->CR &= ~DMA_SxCR_EN ;	// Disable DMA stream
		while ((pStream->CR & DMA_SxCR_EN) != 0u)
		{
		}
	}

	return DAC_ENONE ;
}

//----------------------------------------------------------------------
//	Configure DAC hold mode

dacStatus_t 	dacSetHold			(dacHandle_t hdl, uint32_t channel, uint32_t sample, uint32_t hold, uint32_t refresh)
{
	dacIoDesc_t			* pIoDesc = (dacIoDesc_t *) hdl ;
	DAC_TypeDef			* pDac = pIoDesc->pDacDesc->pDac ;

	if (channel == DAC_CHAN1)
	{
		while ((pDac->SR & DAC_SR_BWST1) != 0u)
		{
		}
		pDac->SHSR1 = sample ;
		pDac->SHHR = (pDac->SHHR & ~DAC_SHHR_THOLD1)    | ((hold    << DAC_SHHR_THOLD1_Pos)    & DAC_SHHR_THOLD1) ;
		pDac->SHRR = (pDac->SHRR & ~DAC_SHRR_TREFRESH1) | ((refresh << DAC_SHRR_TREFRESH1_Pos) & DAC_SHRR_TREFRESH1) ;
	}
	else if (channel == DAC_CHAN2)
	{
		while ((pDac->SR & DAC_SR_BWST2) != 0u)
		{
		}
		pDac->SHSR2 = sample ;
		pDac->SHHR = (pDac->SHHR & ~DAC_SHHR_THOLD2)    | ((hold    << DAC_SHHR_THOLD2_Pos)    & DAC_SHHR_THOLD2) ;
		pDac->SHRR = (pDac->SHRR & ~DAC_SHRR_TREFRESH2) | ((refresh << DAC_SHRR_TREFRESH2_Pos) & DAC_SHRR_TREFRESH2) ;
	}
	else
	{
		return DAC_EARG ;
	}

	return DAC_ENONE ;
}

//----------------------------------------------------------------------
//	Set static output value

dacStatus_t 	dacOuput		(dacHandle_t hdl, uint32_t channel, uint32_t value)
{
	dacIoDesc_t		* pIoDesc = (dacIoDesc_t *) hdl ;

	if (channel == DAC_CHAN1)
	{
		* (uint32_t *) pIoDesc->dataReg [0] = value ;
	}
	else if (channel == DAC_CHAN2)
	{
		* (uint32_t *) pIoDesc->dataReg [1] = value ;
	}
	else
	{
		return DAC_EARG ;
	}
	return DAC_ENONE ;
}

//----------------------------------------------------------------------
// Triggers DAC conversion by software

dacStatus_t 	dacSwTrigger	(dacHandle_t hdl, uint32_t channel)
{
	dacIoDesc_t		* pIoDesc = (dacIoDesc_t *) hdl ;
	pIoDesc->pDacDesc->pDac->SWTRIGR = channel & 3u ;

    return DAC_ENONE ;
}

//----------------------------------------------------------------------

dacStatus_t		dacSetCalibration	(dacHandle_t hdl, uint32_t channel, uint32_t bEnable)
{
	dacIoDesc_t		* pIoDesc = (dacIoDesc_t *) hdl ;
	DAC_TypeDef		* pDac = pIoDesc->pDacDesc->pDac ;
	uint32_t 		shift ;
	uint32_t		reg ;

	if (channel == DAC_CHAN1)
	{
		shift = 0u ;
	}
	else if (channel == DAC_CHAN2)
	{
		shift = 16u ;
	}
	else
	{
		return DAC_EARG ;
	}

	// The channel should be off
	if ((pDac->CR & (DAC_CR_EN1 << shift)) != 0u)
	{
		return DAC_ESTATE;
	}

	reg = pDac->CR & ~(DAC_CR_CEN1 << shift) ;
	if (bEnable != 0u)
	{
		reg |= DAC_CR_CEN1 << shift ;
	}
	pDac->CR = reg ;
	return DAC_ENONE ;
}

//----------------------------------------------------------------------

dacStatus_t		dacSetOtrim			(dacHandle_t hdl, uint32_t channel, uint32_t otrim)
{
	dacIoDesc_t		* pIoDesc = (dacIoDesc_t *) hdl ;
	DAC_TypeDef		* pDac = pIoDesc->pDacDesc->pDac ;
	uint32_t		reg ;

	if (channel == DAC_CHAN1)
	{
		reg = pDac->CCR & ~DAC_CCR_OTRIM1 ;
		reg |= (otrim << DAC_CCR_OTRIM1_Pos) & DAC_CCR_OTRIM1 ;
		pDac->CCR = reg ;
	}
	else if (channel == DAC_CHAN2)
	{
		reg = pDac->CCR & ~DAC_CCR_OTRIM2 ;
		reg |= (otrim << DAC_CCR_OTRIM2_Pos) & DAC_CCR_OTRIM2 ;
		pDac->CCR = reg ;
	}
	else
	{
		return DAC_EARG ;
	}
	return DAC_ENONE ;
}

//----------------------------------------------------------------------

dacStatus_t		dacGetOtrim			(dacHandle_t hdl, uint32_t channel, uint32_t * pOtrim)
{
	dacIoDesc_t		* pIoDesc = (dacIoDesc_t *) hdl ;
	DAC_TypeDef		* pDac = pIoDesc->pDacDesc->pDac ;

	if (channel == DAC_CHAN1)
	{
		* pOtrim = (pDac->CCR & ~DAC_CCR_OTRIM1) >> DAC_CCR_OTRIM1_Pos ;
	}
	else if (channel == DAC_CHAN2)
	{
		* pOtrim = (pDac->CCR & ~DAC_CCR_OTRIM2) >> DAC_CCR_OTRIM2_Pos ;
	}
	else
	{
		return DAC_EARG ;
	}

	return DAC_ENONE ;
}

//----------------------------------------------------------------------
//	On most H7 and L4 the delay TTRIM is 50us min

dacStatus_t		dacGetCalFlag		(dacHandle_t hdl, uint32_t channel, uint32_t delay, uint32_t * pFlag)
{
	dacIoDesc_t		* pIoDesc = (dacIoDesc_t *) hdl ;
	DAC_TypeDef		* pDac = pIoDesc->pDacDesc->pDac ;
	uint32_t 		shift ;
	uint32_t		reg ;

	if (channel == DAC_CHAN1)
	{
		shift = 0u ;
	}
	else if (channel == DAC_CHAN2)
	{
		shift = 16u ;
	}
	else
	{
		return DAC_EARG ;
	}

	// The channel calibration should be on
	if ((pDac->CR & (DAC_CR_CEN1 << shift)) == 0u)
	{
		return DAC_ESTATE;
	}

	bspDelayUs (delay) ;
	reg = (pDac->SR >> shift) >> DAC_SR_CAL_FLAG1_Pos ;
	* pFlag = reg & 1 ;
	return DAC_ENONE ;
}

//----------------------------------------------------------------------
//	For calibration set the DAC then try the OTRIM in increment values:
//	When DAC_SR_CAL_FLAGx is set then the calibration is done.
//	Calibration is allowed only if the DAC buffer is enabled.
//	Channel is DAC_CHAN1 or DAC_CHAN2.

dacStatus_t		dacCalibration		(dacHandle_t hdl, uint32_t channel, uint32_t delay)
{
	dacIoDesc_t		* pIoDesc = (dacIoDesc_t *) hdl ;
	DAC_TypeDef		* pDac = pIoDesc->pDacDesc->pDac ;
	uint32_t 		flag ;
	uint32_t		done ;
	uint32_t		shift ;
	uint32_t		ix ;
	uint32_t		reg ;

	if (channel == DAC_CHAN1)
	{
		shift = 0u ;
	}
	else if (channel == DAC_CHAN2)
	{
		shift = 16u ;
	}
	else
	{
		return DAC_EARG ;
	}

	// The channel should be off
	if ((pDac->CR & (DAC_CR_EN1 << shift)) != 0u)
	{
		return DAC_ESTATE;
	}

	// Set calibration bit
	pDac->CR |= DAC_CR_CEN1 << shift ;

	done = DAC_EFAIL ;
	for (ix = 0u ; ix < 0x20u ; ix++)
	{
		// Set trim value
		reg = pDac->CCR & ~(DAC_CCR_OTRIM1 << shift) ;
		reg |= (ix & DAC_CCR_OTRIM1) << shift ;
		pDac->CCR = reg ;

		// Wait for CAL_FLAG
		bspDelayUs (delay) ;
		flag = 1u  &  ((pDac->SR >> shift) >> DAC_SR_CAL_FLAG1_Pos) ;

		if (flag == 1u)
		{
			// End of calibration
			done = DAC_ENONE ;
			break ;
		}
	}
	// Reset calibration bit
	pDac->CR &= ~(DAC_CR_CEN1 << shift) ;
	return done ;
}

//----------------------------------------------------------------------
//	Wave generation
//	mode is one of DAC_GEN_xx
//	MAMP is the value to set in DAC_CR MAMPx field

dacStatus_t		dacSetWave	(dacHandle_t hdl, uint32_t channel, uint32_t mode, uint32_t MAMP)
{
	dacIoDesc_t		* pIoDesc = (dacIoDesc_t *) hdl ;
	DAC_TypeDef		* pDac = pIoDesc->pDacDesc->pDac ;
	uint32_t 		shift ;
	uint32_t		reg ;

	if (channel == DAC_CHAN1)
	{
		shift = 0u ;
	}
	else if (channel == DAC_CHAN2)
	{
		shift = 16u ;
	}
	else
	{
		return DAC_EARG ;
	}

	reg = pDac->CR & ~((DAC_CR_MAMP1 | DAC_CR_WAVE1) << shift) ;
	reg |= (((mode << DAC_CR_WAVE1_Pos) & DAC_CR_WAVE1) | (((MAMP << DAC_CR_MAMP1_Pos) & DAC_CR_MAMP1))) << shift ;
	pDac->CR = reg ;

	return DAC_ENONE ;
}

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//	DAC1 interrupt handler, called from the TIM6 ISR

void	dac1BasicIrqHandler (void) ;
void	dac1BasicIrqHandler (void)
{
	dacIoDesc_t		* pIoDesc = & ioDesc1 ;

	// Check whether DAC channel1 underrun caused the DAC interrupt
	if (LL_DAC_IsActiveFlag_DMAUDR1 (DAC1) != 0u)
	{
		// Clear flag DAC channel1 underrun
		LL_DAC_ClearFlag_DMAUDR1 (DAC1) ;

	    // Call interrupt callback function
		if (pIoDesc->cb [0] != NULL)
		{
			(* pIoDesc->cb [0]) (pIoDesc, DAC_EV_UNDR, DAC_CHAN1, pIoDesc->cbArg [0]) ;
		}
	}

	// Check whether DAC channel2 underrun caused the DAC interrupt
	if (LL_DAC_IsActiveFlag_DMAUDR2 (DAC1) != 0u)
	{
		// Clear flag DAC channel2 underrun
		LL_DAC_ClearFlag_DMAUDR2 (DAC1) ;

	    // Call interrupt callback function
		if (pIoDesc->cb [1] != NULL)
		{
			(* pIoDesc->cb [1]) (pIoDesc, DAC_EV_UNDR, DAC_CHAN2, pIoDesc->cbArg [1]) ;
		}
	}
}

//----------------------------------------------------------------------
//	Configure the analog reference buffer

#if (defined VREFBUF)
void	vrefbufSet (uint32_t vrefbufMode, uint32_t scale)
{
	uint32_t	reg;

	reg = ((vrefbufMode & 1u) << 1u) | ((vrefbufMode & 2u) >> 1u) ;
	reg |= (scale << VREFBUF_CSR_VRS_Pos) & VREFBUF_CSR_VRS ;
	VREFBUF->CSR = reg ;

	if (vrefbufMode == VREFBUF_MODE_INT)
	{
		while ((VREFBUF->CSR & VREFBUF_CSR_VRR) == 0u)
		{
		}
	}
}
#endif	// VREFBUF

//----------------------------------------------------------------------
//	DAC2 interrupt handler

#if (WITH_DAC_2 == 1  &&  defined (DAC2))
void DAC2_IRQHandler (void) ;
void DAC2_IRQHandler (void)
{

}
#endif	// WITH_DAC_2 == 1  &&  defined DAC2

//----------------------------------------------------------------------
