/*
----------------------------------------------------------------------

	AdAstra - Real Time Kernel

	Alain Chebrou

	i2cC24xx.c	Master I2C library for 24Cxx EEPROM
				Allows to use 24C series of EEPROM with up to 2048 bytes

	When		Who	What
	09/04/18	ac	Creation
	04/04/20	ac	Port to STM32H7 and STM32L4+

----------------------------------------------------------------------
*/
/*
----------------------------------------------------------------------
	Copyright (c) 2013-2023 Alain Chebrou - AdAstra-Soft . com
	All rights reserved.

	This file is part of the AdAstra Real Time Kernel distribution.

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions are met:

	- Redistributions of source code must retain the above copyright
	  notice, this list of conditions and the following disclaimer.

	- Redistributions in binary form must reproduce the above copyright
	  notice, this list of conditions and the following disclaimer in the
	  documentation and/or other materials provided with the distribution.

	- Neither the name of the copyright holder nor the
	  names of its contributors may be used to endorse or promote products
	  derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
	DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
	DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
	(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
	ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
----------------------------------------------------------------------
*/

#include	"aa.h"

#include	"i2cbasic.h"
#include	"ee24Cxx.h"

//----------------------------------------------------------------------
//----------------------------------------------------------------------

#if ! defined (NULL)
	#define	NULL	((void *) 0)
#endif

const uint32_t	ee24cPageDesc [5] =
{
	//	Device 			  Page         Page
	//	address mask      count        size
		(0x00FFu << 16) | ( 16 << 8) |  8,		// EE24C01
		(0x00FFu << 16) | ( 32 << 8) |  8,		// EE24C02
		(0x00FDu << 16) | ( 32 << 8) | 16,		// EE24C04
		(0x00F9u << 16) | ( 64 << 8) | 16,		// EE24C08
		(0x00F1u << 16) | (128 << 8) | 16,		// EE24C16
} ;

//----------------------------------------------------------------------
//	Initialize an EEPROM descriptor

i2cStatus_t	ee24cInit (ee24cDesc_t * pDesc, ee24cType_t type, uint32_t i2cAddress)
{
	pDesc->i2cAddress = (uint8_t) i2cAddress ;
	pDesc->eeType = type ;
	return I2C_EOK ;
}

//----------------------------------------------------------------------
//	Write data not crossing a page boundary

static	i2cStatus_t	ee24cPageWrite (ee24cDesc_t * pDesc, uint32_t offset, uint8_t * pBuffer, uint32_t count)
{
	i2cStatus_t	status ;

	// Compute device I2C address byte
	// header [0] is I2C device address + high bits of offset
	// header [1] is low bits of offset
	pDesc->header [0] = pDesc->i2cAddress & EE24C_PAGEMASK (pDesc->eeType) ;	// Clear Px bits
	pDesc->header [0] |= (uint8_t) ((offset >> 7u) & ~EE24C_PAGEMASK (pDesc->eeType)) ;		// Add Px bits
	pDesc->header [1] = offset & 0x00FFu ;

	// Send header
	status = i2cMasterStart (pDesc->i2cHdl, pDesc->header [0], I2C_TX) ;
	if (status != I2C_EOK)
	{
		// Slave not responding ?
		return status ;
	}

	status = i2cMasterXfer (pDesc->i2cHdl, & pDesc->header [1], 1, I2C_NEXT) ;
	if (status != I2C_EOK)
	{
		return status ;
	}

	// Send bytes
	status = i2cMasterXfer (pDesc->i2cHdl, pBuffer, count, I2C_LAST) ;
	if (status != I2C_EOK)
	{
		return status ;
	}

	i2cMasterStop (pDesc->i2cHdl) ;

	// Wait until end of write
	while (1)
	{
		status = i2cMasterProbeDevice (pDesc->i2cHdl, pDesc->header [0]) ;
		if (status == I2C_EOK)
		{
			break ;
		}
	}

	return status ;
}

//----------------------------------------------------------------------
//	Write any amount of data anywhere in the EEPROM

i2cStatus_t	ee24cWrite (ee24cDesc_t * pDesc, uint32_t offset, uint8_t * pBuffer, uint32_t count)
{
	uint8_t		* pByte = pBuffer ;
	i2cStatus_t	status = I2C_EOK ;
	uint32_t	pageOffset = offset ;
	uint32_t	size = count ;
	uint32_t	temp ;

	// Write first incomplete page
	temp = pageOffset & (EE24C_PAGESIZE (pDesc->eeType) - 1) ;
	if (temp != 0)
	{
		// Offset is not at page boundary
		if (temp > size)
		{
			temp = size ;
		}
		status = ee24cPageWrite (	pDesc,
									pageOffset,
									pByte,
									EE24C_PAGESIZE (pDesc->eeType) - temp) ;
		pageOffset += temp ;
		size -= temp ;
		pByte = & pByte [temp] ;
	}

	// Write full pages
	temp = EE24C_PAGESIZE (pDesc->eeType) ;
	while (status == I2C_EOK  &&  size >= temp)
	{
		status = ee24cPageWrite (	pDesc,
									pageOffset,
									pByte,
									temp) ;
		pageOffset += temp ;
		size -= temp ;
		pByte = & pByte [temp] ;
	}


	// Write remaining in last page
	if (status == I2C_EOK  &&  size > 0)
	{
		status = ee24cPageWrite (	pDesc,
									pageOffset,
									pByte,
									size) ;
	}

	return status ;
}

//----------------------------------------------------------------------
//	Read any amount of data anywhere in the EEPROM

i2cStatus_t	ee24cRead (ee24cDesc_t * pDesc, uint32_t offset, uint8_t * pBuffer, uint32_t count)
{
	i2cStatus_t	status = I2C_EOK ;
	i2cMode_t	flags ;

	// Compute device I2C address byte
	// header [0] is the device address
	// header [1] is the byte offset
	pDesc->header [0] = pDesc->i2cAddress & EE24C_PAGEMASK (pDesc->eeType) ;	// Clear Px bits
	pDesc->header [0] |= (uint8_t) ((offset >> 7u) & ~EE24C_PAGEMASK (pDesc->eeType)) ;		// Add Px bits
	pDesc->header [1] = offset & 0x00FFu ;

	// ----------------------------- EEPROM addressing phase
	// Send EEPROM I2C address
	status = i2cMasterStart (pDesc->i2cHdl, pDesc->header [0], I2C_TX) ;
	if (status != I2C_EOK)
	{
		// Slave not responding ?
		return status ;
	}

	// Send offset in EEPROM
	status = i2cMasterXfer (pDesc->i2cHdl, & pDesc->header [1], 1, I2C_LAST) ;
	if (status != I2C_EOK)
	{
		return status ;
	}
	i2cMasterStop (pDesc->i2cHdl) ;


	// ----------------------------- Read bytes
	// Send EEPROM I2C address
	flags = I2C_RX ;
	if (count == 1)
	{
		flags = (i2cMode_t) (flags | I2C_1LAST) ;
	}
	status = i2cMasterStart (pDesc->i2cHdl, pDesc->header [0], flags) ;
	if (status != I2C_EOK)
	{
		// Slave not responding ?
		return status ;
	}

	// Read bytes
	status = i2cMasterXfer (pDesc->i2cHdl, pBuffer, count, I2C_LAST) ;
	if (status != I2C_EOK)
	{
		return status ;
	}
	i2cMasterStop (pDesc->i2cHdl) ;

	return status ;
}

//----------------------------------------------------------------------
