/*
----------------------------------------------------------------------

	AdAstra - Real Time Kernel

	Alain Chebrou

	i2cbasic.c	Master and Slave basic routines:
				7 bits address only (10 bits address not supported)
				Slave have only one address, no general call
				Clock stretching enabled (NOSTRETCH=1 not supported)
				No digital filter

				i2cTiming() function allows to compute the proper timing word for i2cConfigure().

	When		Who	What
	26/01/20	ac	Creation for STM32H7
	30/03/20	ac	Port to STM32L4+
	10/04/20	ac	Changed wait method in i2cMasterProbeDevice()
	26/10/21	ac	Port to STM32G0
	13/12/21	ac	Port to STM32G4

----------------------------------------------------------------------
*/
/*
----------------------------------------------------------------------
	Copyright (c) 2013-2023 Alain Chebrou - AdAstra-Soft . com
	All rights reserved.

	This file is part of the AdAstra Real Time Kernel distribution.

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions are met:

	- Redistributions of source code must retain the above copyright
	  notice, this list of conditions and the following disclaimer.

	- Redistributions in binary form must reproduce the above copyright
	  notice, this list of conditions and the following disclaimer in the
	  documentation and/or other materials provided with the distribution.

	- Neither the name of the copyright holder nor the
	  names of its contributors may be used to endorse or promote products
	  derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
	DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
	DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
	(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
	ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
----------------------------------------------------------------------
*/

#include	"aa.h"
#include	"aakernel.h"
#include	"aalogmes.h"

#if (defined STM32G0)
	#include	"stm32g0xx_ll_i2c.h"
	#include	"stm32g0xx_ll_bus.h"
	#include	"stm32g0xx_ll_rcc.h"
#endif

#if (defined STM32L4)
	#include	"stm32l4xx_ll_i2c.h"
	#include	"stm32l4xx_ll_bus.h"
	#include	"stm32l4xx_ll_rcc.h"
#endif

#if (defined STM32G4)
	#include	"stm32g4xx_ll_i2c.h"
	#include	"stm32g4xx_ll_bus.h"
	#include	"stm32g4xx_ll_rcc.h"
#endif

#if (defined STM32H7)
	#include	"stm32h7xx_ll_i2c.h"
	#include	"stm32h7xx_ll_bus.h"
	#include	"stm32h7xx_ll_rcc.h"
#endif

#include	"i2cbasic.h"
#include	"rccbasic.h"

//	Defining DBG_I2C to 1 allows to get a trace of all steps of an I2C exchange on aaLogMes() output
#define DBG_I2C		0

//----------------------------------------------------------------------
//----------------------------------------------------------------------

// For i2cUtil() parameter
#define	I2C_CMD_ENABLE			1		// Enable peripheral clock and reset, or Disable peripheral clock
#define	I2C_CMD_EVITEN			2		// Configure and enable event interrupt
#define	I2C_CMD_ERITEN			3		// Configure and enable error interrupt
#define	I2C_CMD_ITDISABLE		4		// Disable event and error interrupts

//----------------------------------------------------------------------
// Transaction descriptor

typedef struct i2cIo_s
{
	I2C_TypeDef		* pI2c ;			// I2C1, I2C2...
	uint32_t		flags ;				// Configuration and status flags
	uint8_t			* pBuffer ;			// Data to send or receive
	int32_t			size ;				// Remaining count of data to exchange (RX or TX)
	uint32_t		count ;				// Count of exchanged data
	i2cStatus_t		result ;			// Result of the operation
	aaTaskId_t		taskId ;			// The task to signal at end of operation
	aaSignal_t		signal ;			// The signal to use (0 if signal must not be used)

} i2cIo_t ;

// For i2cIo_t.flags
#define	I2C_FLAG_DIRMASK		0x00000001u
#define	I2C_FLAG_TX				0x00000000u
#define	I2C_FLAG_RX				0x00000001u

#define	I2C_FLAG_MODEMASK		0x00000010u
#define	I2C_FLAG_MASTER			0x00000000u
#define	I2C_FLAG_SLAVE			0x00000010u

#define	I2C_FLAG_LASTMASK		0x00001000u	// Mask of LAST xfer value
#define	I2C_FLAG_NEXT			0x00000000u	// Not last xfer
#define	I2C_FLAG_LAST			0x00001000u	// Last xfer before STOP or RESTART

#define	I2C_STATE_MASK			0x000F0000u	// Mask of mode value
#define	I2C_STATE_IDLE			0x00010000u	// Interface not initialized
#define	I2C_STATE_ADDR			0x00020000u	// In master mode, waiting for end of start condition and slave address transmission
#define	I2C_STATE_TX			0x00030000u	// Interface xfer TX
#define	I2C_STATE_TXWAIT		0x00040000u	// Interface waiting for new data or stop or restart condition
#define	I2C_STATE_RX			0x00050000u	// Interface xfer RX
#define	I2C_STATE_RXWAIT1		0x00060000u	// Interface after slave addressing for master RX. RXDR is full
#define	I2C_STATE_RXWAIT		0x00070000u	// Interface waiting for new data or stop or restart condition

// To set i2cIo_t.flags
__STATIC_INLINE void setFlagState (uint32_t * pFlags, uint32_t mode)
{
	* pFlags = (* pFlags & ~I2C_STATE_MASK) | (mode & I2C_STATE_MASK) ;
}

__STATIC_INLINE void setFlagDir (uint32_t * pFlags, uint32_t dir)
{
	* pFlags = (* pFlags & ~I2C_FLAG_DIRMASK) | (dir & I2C_FLAG_DIRMASK) ;
}

__STATIC_INLINE void setFlagMode (uint32_t * pFlags, uint32_t mode)
{
	* pFlags = (* pFlags & ~I2C_FLAG_MODEMASK) | (mode & I2C_FLAG_MODEMASK) ;
}

__STATIC_INLINE void setFlagLast (uint32_t * pFlags, uint32_t nextOrLast)
{
	* pFlags = (* pFlags & ~I2C_FLAG_LASTMASK) | (nextOrLast & I2C_FLAG_LASTMASK) ;
}


// To get i2cIo_t.flags
__STATIC_INLINE uint32_t getFlagState (uint32_t flags)
{
	return flags & I2C_STATE_MASK ;
}

__STATIC_INLINE uint32_t getFlagDir (uint32_t flags)
{
	return flags & I2C_FLAG_DIRMASK ;
}

__STATIC_INLINE uint32_t getFlagMode (uint32_t flags)
{
	return flags & I2C_FLAG_MODEMASK ;
}

__STATIC_INLINE uint32_t getFlagLast (uint32_t flags)
{
	return flags & I2C_FLAG_LASTMASK ;
}

//----------------------------------------------------------------------

#if (WITH_I2C_X == 1)

	#if (defined STM32G0)
		#include "i2cG0xx.h"		//	I2C descriptors for STM32G0xx
	#endif

	#if (defined STM32L4)
		#include "i2cL4xx.h"		//	I2C descriptors for STM32L4xx
	#endif

	#if (defined STM32G4)
		#include "i2cG4xx.h"		//	I2C descriptors for STM32G4xx
	#endif

	#if (defined STM32H7)
		#include "i2cH7xx.h"		//	I2C descriptors for STM32H7xx
	#endif

	STATIC const uint32_t		i2cArrayCount = sizeof (i2cArray) / sizeof (i2cDesc_t) ;

#endif

//----------------------------------------------------------------------
//	One I/O descriptor for each physical I2C peripheral

#if (WITH_I2C_1 == 1)
	i2cIo_t		i2cIo1 ;
#endif	// WITH_I2C_1

#if (WITH_I2C_2 == 1)
	i2cIo_t		i2cIo2 ;
#endif	// WITH_I2C_2

#if (WITH_I2C_3 == 1)
	i2cIo_t		i2cIo3 ;
#endif	// WITH_I2C_3

#if (WITH_I2C_4 == 1)
	i2cIo_t		i2cIo4 ;
#endif	// WITH_I2C_4

static	uint32_t		now ;		// For debug, used when DBG_I2C is 1. Allows to know the delay between traces in us.

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//	Get the I/O descriptor address for an I2C peripheral

static	i2cIo_t *	i2cGetIoDesc (I2C_TypeDef * pI2C) ;

static	i2cIo_t *	i2cGetIoDesc (I2C_TypeDef * pI2C)
{
	#if (WITH_I2C_X == 0)
	{
		(void) pI2C ;	// To avoid compiler warning when none I2C is supported
	}
	#endif

	#if (WITH_I2C_1 == 1)
	{
		if (pI2C == I2C1)
		{
			return & i2cIo1 ;
		}
	}
	#endif	// WITH_I2C_1

	#if (WITH_I2C_2 == 1)
	{
		if (pI2C == I2C2)
		{
			return & i2cIo2 ;
		}
	}
	#endif	// WITH_I2C_2

	#if (WITH_I2C_3 == 1)
	{
		if (pI2C == I2C3)
		{
			return & i2cIo3 ;
		}
	}
	#endif	// WITH_I2C_3

	#if (WITH_I2C_4 == 1)
	{
		if (pI2C == I2C4)
		{
			return & i2cIo4 ;
		}
	}
	#endif // WITH_I2C_4

	AA_ASSERT (0) ;
	return NULL ;
}

//----------------------------------------------------------------------
//	Some utility commands to configure I2C peripherals

static	i2cStatus_t	i2cUtil (i2cIo_t * pI2cIo, uint32_t command, uint32_t arg) ;

static	i2cStatus_t	i2cUtil (i2cIo_t * pI2cIo, uint32_t command, uint32_t arg)
{
	i2cStatus_t		result = I2C_ENONE ;

	#if (WITH_I2C_1 == 1)
	if (pI2cIo->pI2c == I2C1)
	{
		switch (command)
		{
			case I2C_CMD_ENABLE:
				if (arg == 1)
				{
					LL_APB1_GRP1_EnableClock  (LL_APB1_GRP1_PERIPH_I2C1) ;
					LL_APB1_GRP1_ForceReset   (LL_APB1_GRP1_PERIPH_I2C1) ;
					LL_APB1_GRP1_ReleaseReset (LL_APB1_GRP1_PERIPH_I2C1) ;
				}
				else
				{
					LL_APB1_GRP1_DisableClock (LL_APB1_GRP1_PERIPH_I2C1) ;
				}
				break ;

			case I2C_CMD_EVITEN:
#if (defined STM32G0)
				NVIC_SetPriority (I2C1_IRQn, arg);
				NVIC_EnableIRQ   (I2C1_IRQn) ;
#endif
#if (defined STM32L4 || defined STM32G4 || defined STM32H7)
				NVIC_SetPriority (I2C1_EV_IRQn, arg);
				NVIC_EnableIRQ   (I2C1_EV_IRQn) ;
#endif
				break ;

			case I2C_CMD_ERITEN:
#if (defined STM32G0)
				NVIC_SetPriority (I2C1_IRQn, arg);
				NVIC_EnableIRQ   (I2C1_IRQn) ;
#endif
#if (defined STM32L4 || defined STM32G4 || defined STM32H7)
				NVIC_SetPriority (I2C1_ER_IRQn, arg);
				NVIC_EnableIRQ   (I2C1_ER_IRQn) ;
#endif
				break ;

			case I2C_CMD_ITDISABLE:
#if (defined STM32G0)
				NVIC_DisableIRQ   (I2C1_IRQn) ;
#endif
#if (defined STM32L4 || defined STM32G4 || defined STM32H7)
				NVIC_DisableIRQ   (I2C1_EV_IRQn) ;
				NVIC_DisableIRQ   (I2C1_ER_IRQn) ;
#endif
				break ;

			default:
				AA_ASSERT (0) ;
				result = I2C_EFAIL ;
				break ;
		}
		return result ;
	}
	#endif	// WITH_I2C_1

	#if (WITH_I2C_2 == 1)
	if (pI2cIo->pI2c == I2C2)
	{
		switch (command)
		{
			case I2C_CMD_ENABLE:
				if (arg == 1)
				{
					LL_APB1_GRP1_EnableClock  (LL_APB1_GRP1_PERIPH_I2C2) ;
					LL_APB1_GRP1_ForceReset   (LL_APB1_GRP1_PERIPH_I2C2) ;
					LL_APB1_GRP1_ReleaseReset (LL_APB1_GRP1_PERIPH_I2C2) ;
				}
				else
				{
					LL_APB1_GRP1_DisableClock (LL_APB1_GRP1_PERIPH_I2C2) ;
				}
				break ;

			case I2C_CMD_EVITEN:
#if (defined STM32G0)
				#if (defined I2C3)
					NVIC_SetPriority (I2C2_3_IRQn, arg);
					NVIC_EnableIRQ   (I2C2_3_IRQn) ;
				#else
					NVIC_SetPriority (I2C2_IRQn, arg);
					NVIC_EnableIRQ   (I2C2_IRQn) ;
				#endif
#endif
#if (defined STM32L4 || defined STM32G4 || defined STM32H7)
				NVIC_SetPriority (I2C2_EV_IRQn, arg);
				NVIC_EnableIRQ   (I2C2_EV_IRQn) ;
#endif
				break ;

			case I2C_CMD_ERITEN:
#if (defined STM32G0)
				#if (defined I2C3)
					NVIC_SetPriority (I2C2_3_IRQn, arg);
					NVIC_EnableIRQ   (I2C2_3_IRQn) ;
				#else
					NVIC_SetPriority (I2C2_IRQn, arg);
					NVIC_EnableIRQ   (I2C2_IRQn) ;
				#endif
#endif
#if (defined STM32L4 || defined STM32G4 || defined STM32H7)
				NVIC_SetPriority (I2C2_ER_IRQn, arg);
				NVIC_EnableIRQ   (I2C2_ER_IRQn) ;
#endif
				break ;

			case I2C_CMD_ITDISABLE:
#if (defined STM32G0)
				#if (defined I2C3)
					NVIC_DisableIRQ   (I2C2_3_IRQn) ;
				#else
					NVIC_DisableIRQ   (I2C2_IRQn) ;
				#endif
#endif
#if (defined STM32L4 || defined STM32G4 || defined STM32H7)
				NVIC_DisableIRQ   (I2C2_EV_IRQn) ;
				NVIC_DisableIRQ   (I2C2_ER_IRQn) ;
#endif
				break ;

			default:
				AA_ASSERT (0) ;
				result = I2C_EFAIL ;
				break ;
		}
		return result ;
	}
	#endif	// WITH_I2C_2

	#if (WITH_I2C_3 == 1)
	if (pI2cIo->pI2c == I2C3)
	{
		switch (command)
		{
			case I2C_CMD_ENABLE:
				if (arg == 1)
				{
					LL_APB1_GRP1_EnableClock  (LL_APB1_GRP1_PERIPH_I2C3) ;
					LL_APB1_GRP1_ForceReset   (LL_APB1_GRP1_PERIPH_I2C3) ;
					LL_APB1_GRP1_ReleaseReset (LL_APB1_GRP1_PERIPH_I2C3) ;
				}
				else
				{
					LL_APB1_GRP1_DisableClock (LL_APB1_GRP1_PERIPH_I2C3) ;
				}
				break ;

			case I2C_CMD_EVITEN:
#if (defined STM32G0)
				NVIC_SetPriority (I2C2_3_IRQn, arg);
				NVIC_EnableIRQ   (I2C2_3_IRQn) ;
#endif
#if (defined STM32L4 || defined STM32G4 || defined STM32H7)
				NVIC_SetPriority (I2C3_EV_IRQn, arg);
				NVIC_EnableIRQ   (I2C3_EV_IRQn) ;
#endif
				break ;

			case I2C_CMD_ERITEN:
#if (defined STM32G0)
				NVIC_SetPriority (I2C2_3_IRQn, arg);
				NVIC_EnableIRQ   (I2C2_3_IRQn) ;
#endif
#if (defined STM32L4 || defined STM32G4 || defined STM32H7)
				NVIC_SetPriority (I2C3_ER_IRQn, arg);
				NVIC_EnableIRQ   (I2C3_ER_IRQn) ;
#endif
				break ;

			case I2C_CMD_ITDISABLE:
#if (defined STM32G0)
				NVIC_DisableIRQ   (I2C2_3_IRQn) ;
#endif
#if (defined STM32L4 || defined STM32G4 || defined STM32H7)
				NVIC_DisableIRQ   (I2C3_EV_IRQn) ;
				NVIC_DisableIRQ   (I2C3_ER_IRQn) ;
#endif
				break ;

			default:
				AA_ASSERT (0) ;
				result = I2C_EFAIL ;
				break ;
		}
		return result ;
	}
	#endif	// WITH_I2C_3

	#if (WITH_I2C_4 == 1)
	if (pI2cIo->pI2c == I2C4)
	{
		switch (command)
		{
			case I2C_CMD_ENABLE:
#if (defined STM32H7)
				if (arg == 1)
				{
					LL_APB4_GRP1_EnableClock  (LL_APB4_GRP1_PERIPH_I2C4) ;
					LL_APB4_GRP1_ForceReset   (LL_APB4_GRP1_PERIPH_I2C4) ;
					LL_APB4_GRP1_ReleaseReset (LL_APB4_GRP1_PERIPH_I2C4) ;
				}
				else
				{
					LL_APB4_GRP1_DisableClock (LL_APB4_GRP1_PERIPH_I2C4) ;
				}
#else
				if (arg == 1)
				{
					LL_APB1_GRP2_EnableClock  (LL_APB1_GRP2_PERIPH_I2C4) ;
					LL_APB1_GRP2_ForceReset   (LL_APB1_GRP2_PERIPH_I2C4) ;
					LL_APB1_GRP2_ReleaseReset (LL_APB1_GRP2_PERIPH_I2C4) ;
				}
				else
				{
					LL_APB1_GRP2_DisableClock (LL_APB1_GRP2_PERIPH_I2C4) ;
				}
#endif
				break ;

			case I2C_CMD_EVITEN:
#if (defined STM32G0)
				NVIC_SetPriority (I2C4_IRQn, arg);
				NVIC_EnableIRQ   (I2C4_IRQn) ;
#endif
#if (defined STM32L4 || defined STM32G4 || defined STM32H7)
				NVIC_SetPriority (I2C4_EV_IRQn, arg);
				NVIC_EnableIRQ   (I2C4_EV_IRQn) ;
#endif
				break ;

			case I2C_CMD_ERITEN:
#if (defined STM32G0)
				NVIC_SetPriority (I2C4_IRQn, arg);
				NVIC_EnableIRQ   (I2C4_IRQn) ;
#endif
#if (defined STM32L4 || defined STM32G4 || defined STM32H7)
				NVIC_SetPriority (I2C4_ER_IRQn, arg);
				NVIC_EnableIRQ   (I2C4_ER_IRQn) ;
#endif
				break ;

			case I2C_CMD_ITDISABLE:
#if (defined STM32G0)
				NVIC_DisableIRQ   (I2C4_IRQn) ;
#endif
#if (defined STM32L4 || defined STM32G4 || defined STM32H7)
				NVIC_DisableIRQ   (I2C4_EV_IRQn) ;
				NVIC_DisableIRQ   (I2C4_ER_IRQn) ;
#endif
				break ;

			default:
				AA_ASSERT (0) ;
				result = I2C_EFAIL ;
				break ;
		}
		return result ;
	}
	#endif	// WITH_I2C_4

	return result ;
}

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//	Computing the timing word for i2cConfigure()

#define	T_SU_DAT_100K	250u
#define	T_VD_DAT_100K	3450u

#define	T_SU_DAT_400K	100u
#define	T_VD_DAT_400K	900u

#define	T_AF			260u		// Analog Filter max

//--------------------------------------------------------------------------------
//	Returns the timing word to configure I2C, or 0 if the I2C clock is not available
//	speed:		Requested I2C baud rate
//	maxEdge:	Maximum rising or falling time of SDA (ns)

//	BEWARE: Timing is accurate only if i2c_ker_ck is an integer multiple of 1 MHz
//	All internal period values in ps for maximum precision

/*
Timing examples with ker_clk = 100 MHz:
timing   5000 F001FFFF
timing  10000 F001FFFF
timing  15000 D001EDED
timing  20000 9002F9F9
timing  25000 7004F9F9
timing  30000 6006EDED
timing  35000 5018EDED
timing  40000 401BF9F9
timing  45000 401BDDDD
timing  50000 302FF9F9
timing  55000 302FE2E2
timing  60000 302FCFCF
timing  65000 302FBFBF
timing  70000 204FEDED
timing  75000 204FDDDD
timing  80000 204FCFCF
timing  85000 204FC3C3
timing  90000 204FB8B8
timing  95000 204FAEAE
timing 100000 10BFF9F9

TIMING	0x40608AFF		// 50k, on PCLK1 100 MHz (CubeMX)
*/

uint32_t	i2cTiming (const i2cDesc_t * pI2cDesc, uint32_t speed, uint32_t maxEdge)
{
	uint32_t	presc ;			// Prescaler. I2C register value is PRESC-1
	uint32_t	tI2cClk ;		// ns
	uint32_t	temp ;
	uint32_t	timing = 0u ;

	if (speed > 400000)
	{
		return 0u ;		// The requested speed is out of range
	}

	// Get i2c_ker_clock frequency, then period
	temp = rccGetClockFrequency (pI2cDesc->pI2c) ;
	if (temp == LL_RCC_PERIPH_FREQUENCY_NO)
	{
		return 0u ;		// The I2C clock is not available
	}
	tI2cClk = 1000000 / (temp / 1000000) ;	// ticCclk is i2c_ker_ck period (ps)

	// Compute prescaler PRESC+1
	presc = ((1000000000 / speed) + 511) / 512 ;	// Up rounded, ns
	presc *= 1000 ;									// ps
	presc = (presc + tI2cClk - 1) / tI2cClk ;
	if (presc > 16)
	{
		presc = 16 ;
	}
	timing |= (presc - 1) << 28 ;
	tI2cClk = tI2cClk * presc ;		// Now ticCclk is I2C clock period (ps)

	// Compute high and low CLK period
	temp = (1000000000 / speed) / 2 ;		// Half CLK period ns
	temp = ((temp * 1000) / tI2cClk) - 1 ;
	if (temp > 255)
	{
		temp = 255 ;
	}
	timing |= temp ;		// SCLL
	timing |= temp << 8 ;	// SCLH

	// Compute SDADEL
	temp = (speed <= 100000) ? T_VD_DAT_100K : T_VD_DAT_400K ;
	temp -= maxEdge + T_AF ;
	temp = (temp * 1000) / (presc * tI2cClk) ;
	if (temp > 15)
	{
		temp = 15 ;
	}
	timing |= temp << 16 ;	// SDADEL

	// Compute SCLDEL
	temp = (speed <= 100000) ? T_SU_DAT_100K : T_SU_DAT_400K ;
	temp = (temp + maxEdge) * 1000 ;
	temp += (presc * tI2cClk) - 1 ;		// Rounded to upper
	temp /= presc * tI2cClk ;
	temp -= 1 ;
	if (temp > 15)
	{
		temp = 15 ;
	}
	timing |= temp << 20 ;	// SCLDEL

	return timing ;
}

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//	Get an I2C descriptor from its identifier

const i2cDesc_t *	i2cGetDesc	(uint32_t i2cId)
{
	#if (WITH_I2C_X == 1)
	{
		uint32_t	ii ;

		for (ii = 0u ; ii < i2cArrayCount ; ii++)
		{
			if (i2cArray [ii].id == i2cId)
			{
				return & i2cArray [ii] ;
			}
		}
	}
	#else
	{
		(void) i2cId ;		// To avoid compiler warning when none I2C is supported
	}
	#endif

	return NULL ;	// Identifier not found
}

//----------------------------------------------------------------------
// 	Configure the I2C peripheral using the descriptor, for master or slave I2C mode
//	Give the I2C handle to use as parameter for the other functions
//	This is a blocking function

i2cStatus_t		i2cConfigure		(const i2cDesc_t * pI2cDesc, uint32_t timing, aaSignal_t signal, i2cHandle_t * pHdl)
{
	i2cIo_t				* pI2cIo ;
	LL_I2C_InitTypeDef	I2C_InitStruct ;

	pI2cIo = i2cGetIoDesc (pI2cDesc->pI2c) ;
	if (pI2cIo == NULL)
	{
		return I2C_EFAIL ;
	}
	aaMEMSET (pI2cIo, 0, sizeof (i2cIo_t)) ;
	pI2cIo->pI2c = pI2cDesc->pI2c ;

	gpioConfigurePin (& pI2cDesc->scl) ;
	gpioConfigurePin (& pI2cDesc->sda) ;

	// Enable the I2C peripheral clock, and reset the peripheral
	(void) i2cUtil (pI2cIo, I2C_CMD_ENABLE, 1) ;

	I2C_InitStruct.PeripheralMode	= LL_I2C_MODE_I2C ;
	I2C_InitStruct.Timing			= timing ;
	I2C_InitStruct.AnalogFilter		= LL_I2C_ANALOGFILTER_ENABLE ;
	I2C_InitStruct.DigitalFilter	= 0u ;
	I2C_InitStruct.OwnAddress1		= 0u ;
	I2C_InitStruct.TypeAcknowledge	= LL_I2C_ACK ;
	I2C_InitStruct.OwnAddrSize		= LL_I2C_OWNADDRESS1_7BIT ;
	(void) LL_I2C_Init (pI2cDesc->pI2c, & I2C_InitStruct) ;

	// Configure Event IT:
	(void) i2cUtil (pI2cIo, I2C_CMD_EVITEN, pI2cDesc->irqPriority) ;

	// Configure Error IT:
	(void) i2cUtil (pI2cIo, I2C_CMD_ERITEN, pI2cDesc->irqPriority) ;

	// Activate  I2C
	LL_I2C_Enable (pI2cDesc->pI2c) ;

	setFlagState  (& pI2cIo->flags, I2C_STATE_IDLE) ;

	pI2cIo->signal	= signal ;
	pI2cIo->result	= I2C_EOK ;
	pI2cIo->taskId	= AA_INVALID_TASK ;

	* pHdl = (i2cHandle_t) pI2cIo ;
	return I2C_ENONE ;
}

//----------------------------------------------------------------------
//	Release the peripheral, and the handle

void 	i2cDeactivate 		(i2cHandle_t hdl)
{
	i2cIo_t		* pI2cIo = (i2cIo_t *) hdl ;

	if (pI2cIo != NULL)
	{
		// Deactivate  I2C
		LL_I2C_Disable (pI2cIo->pI2c) ;

		// Unconfigure IT:
		(void) i2cUtil (pI2cIo, I2C_CMD_ITDISABLE, 0u) ;

		// Disable the peripheral clock for I2C
		(void) i2cUtil (pI2cIo, I2C_CMD_ENABLE, 0u) ;

		pI2cIo->flags = 0u ;
	}
}

//----------------------------------------------------------------------
//	Set the I2C in reset state:
//		Release SCL and SDA lines
//		Reset internal state machines and communication bits
//		Preserve the configuration registers
//	Set the library in idle state for this handle

void	 i2cReset		(i2cHandle_t hdl)
{
	i2cIo_t		* pI2cIo = (i2cIo_t *) hdl ;
	I2C_TypeDef	* pI2c = pI2cIo->pI2c ;

	// Clear all interrupt sources
	pI2c->CR1 &= ~(I2C_CR1_ERRIE | I2C_CR1_TCIE | I2C_CR1_STOPIE | I2C_CR1_NACKIE | I2C_CR1_ADDRIE | I2C_CR1_RXIE | I2C_CR1_TXIE) ;

	LL_I2C_Disable (pI2c) ;
	bspNop () ;
	LL_I2C_Enable (pI2c) ;
	setFlagState (& pI2cIo->flags, I2C_STATE_IDLE) ;
}

//----------------------------------------------------------------------
//	Returns:
//		I2C_EBUSY	if the operation is not ended
//		I2C_EOK		if the operation is ended without error
//		Another error or state value
//	At the end of a transfer the effective count of transfered bytes is copied to * pSize

i2cStatus_t			i2cResult			(i2cHandle_t hdl, uint32_t * pSize)
{
	i2cIo_t		* pI2cIo = (i2cIo_t *) hdl ;
	i2cStatus_t	result ;

	* pSize = 0u ;

	// BEWARE: pI2cIo->result is set by an ISR, so can change between the beginning and the end of the function
	// => Get it only once
	result = pI2cIo->result ;
	if (result != I2C_EBUSY)
	{
		* pSize = pI2cIo->count ;
	}
	return result ;
}

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//	I2C Master API
//----------------------------------------------------------------------
//	Allows to check the presence of a device on the I2C bus.
//	Doesn't send or receive any data byte to/from the device
//	Returns I2C_EOK if the slave sent ACK, else I2C_EAF.
//	Blocking function

i2cStatus_t			i2cMasterProbeDevice	(i2cHandle_t hdl, uint32_t slaveAddress)
{
	i2cIo_t		* pI2cIo = (i2cIo_t *) hdl ;
	I2C_TypeDef	* pI2c = pI2cIo->pI2c ;

	if (getFlagState (pI2cIo->flags) != I2C_STATE_IDLE)
	{
		return I2C_ESTATE ;
	}

	setFlagDir		(& pI2cIo->flags, I2C_FLAG_TX) ;
	setFlagMode		(& pI2cIo->flags, I2C_FLAG_MASTER) ;
	pI2cIo->taskId	= aaTaskSelfId () ;
	pI2cIo->result	= I2C_EBUSY ;

	setFlagState	(& pI2cIo->flags, I2C_STATE_ADDR) ;

	// Enable interrupt sources
if (DBG_I2C) { now = bspTsGet () ; }
	pI2c->CR1 |= (I2C_CR1_STOPIE) ;

	// Set CR2 register
 	// NBYTES=0 and LL_I2C_MODE_AUTOEND allows to automatically generate STOP if the slave answer ACK
	// If the slave doesn't answer, NACKF is received and STOP is automatically generated
	LL_I2C_HandleTransfer (pI2c,
			slaveAddress,
			LL_I2C_ADDRSLAVE_7BIT,
			0u,						// NBYTES
			LL_I2C_MODE_AUTOEND,
			LL_I2C_GENERATE_START_WRITE) ;

	// Wait for end of START phase
// TODO: Timeout
	aaSignalWait (pI2cIo->signal, NULL, AA_SIGNAL_OR, AA_INFINITE) ;

if (DBG_I2C) { aaLogMes ("PROB   %08X %08X %08X %3d\n", pI2c->CR1, pI2c->CR2, pI2c->ISR, bspTsDelta (& now), 0) ; }

	return pI2cIo->result ;
}

//----------------------------------------------------------------------
//	Start or ReStart: Send a START condition then slave address
//	Direction is I2C_TX or I2C_RX,
//		plus I2C_1LAST of only 1 char will be received
//		plus I2C_NOBLOCK, which makes the function non blocking
//	Blocking function if I2C_NOBLOCK is not used (default)

i2cStatus_t 		i2cMasterStart			(i2cHandle_t hdl, uint32_t slaveAddress, uint32_t direction)
{
	i2cIo_t		* pI2cIo = (i2cIo_t *) hdl ;
	I2C_TypeDef	* pI2c = pI2cIo->pI2c ;

	if (getFlagState (pI2cIo->flags) != I2C_STATE_IDLE		&&
		getFlagState (pI2cIo->flags) != I2C_STATE_TXWAIT	&&
		getFlagState (pI2cIo->flags) != I2C_STATE_RXWAIT)
	{
		return I2C_ESTATE ;
	}

	setFlagDir		(& pI2cIo->flags, ((direction & I2C_TX) != 0u) ? I2C_FLAG_TX : I2C_FLAG_RX) ;
	setFlagMode		(& pI2cIo->flags, I2C_FLAG_MASTER) ;
	pI2cIo->result	= I2C_EBUSY ;
	pI2cIo->taskId	= aaTaskSelfId () ;
	pI2cIo->count	= 0u ;		// Count of exchanged bytes
	aaSignalClear (pI2cIo->taskId, pI2cIo->signal) ;

	setFlagState	(& pI2cIo->flags, I2C_STATE_ADDR) ;

	// Disable all interrupt sources
	pI2c->CR1 &= ~(I2C_CR1_ERRIE | I2C_CR1_TCIE | I2C_CR1_STOPIE | I2C_CR1_NACKIE | I2C_CR1_ADDRIE | I2C_CR1_RXIE | I2C_CR1_TXIE) ;

	if ((direction & (I2C_1LAST | I2C_RX)) == (I2C_1LAST | I2C_RX))
	{
		// Special case RX with only 1 byte
		LL_I2C_HandleTransfer (pI2c,
				slaveAddress,
				LL_I2C_ADDRSLAVE_7BIT,
				1,						// NBYTES
				LL_I2C_MODE_SOFTEND,
				LL_I2C_GENERATE_START_READ) ;
	}
	else
	{
		// General case: TX or RX with more than 1 byte
		LL_I2C_HandleTransfer (pI2c,
				slaveAddress,
				LL_I2C_ADDRSLAVE_7BIT,
				1,						// NBYTES
				LL_I2C_MODE_RELOAD,
				((direction & I2C_TX) != 0u) ? LL_I2C_GENERATE_START_WRITE : LL_I2C_GENERATE_START_READ) ;
	}

	// Enable interrupt sources
if (DBG_I2C) { now = bspTsGet () ; }
if (DBG_I2C) { aaLogMes ("strt   %08X %08X %08X %3d\n", pI2c->CR1, pI2c->CR2, pI2c->ISR, bspTsDelta (& now), 0) ; }
	pI2c->CR1 |= (I2C_CR1_ERRIE | I2C_CR1_TCIE | I2C_CR1_STOPIE | I2C_CR1_RXIE  | I2C_CR1_TXIE) ;

	if ((direction & I2C_NOBLOCK) != 0u)
	{
		return I2C_ENONE ;
	}
	aaSignalWait (pI2cIo->signal, NULL, AA_SIGNAL_OR, AA_INFINITE) ;
	return pI2cIo->result ;
}

//----------------------------------------------------------------------
//	Generate a Stop condition
//	This is a blocking function

i2cStatus_t		i2cMasterStop			(i2cHandle_t hdl)
{
	i2cIo_t			* pI2cIo = (i2cIo_t *) hdl ;
	I2C_TypeDef		* pI2c = pI2cIo->pI2c ;
	i2cStatus_t		result ;

	if (getFlagState (pI2cIo->flags) == I2C_STATE_TXWAIT  ||  getFlagState (pI2cIo->flags) == I2C_STATE_RXWAIT)
	{
if (DBG_I2C) { aaLogMes ("STOPm  %08X %08X %08X %3d\n", pI2c->CR1, pI2c->CR2, pI2c->ISR, bspTsDelta (& now), 0) ; }
		if ((pI2c->CR2 & I2C_CR2_START) != 0u)
		{
			pI2c->ICR = I2C_ICR_ADDRCF ;
		}
		// Interrupts are already inhibited
		// Generate STOP
		pI2c->CR2 = (pI2c->CR2 & ~(I2C_CR2_NBYTES | I2C_CR2_RELOAD)) | I2C_CR2_STOP ;	// Delete NBYTES and set STOP at the same time

		// Wait for stop flag, then clear it
		while ((pI2c->ISR & I2C_ISR_STOPF) == 0u)
		{
		}
		pI2c->ICR = I2C_ICR_STOPCF | I2C_ICR_NACKCF ;

		setFlagState (& pI2cIo->flags, I2C_STATE_IDLE) ;
		pI2cIo->result = I2C_EOK ;
		result = I2C_ENONE ;
	}
	else
	{
		// Stop condition not allowed in this state
		// Do not modify pI2cIo->result
if (DBG_I2C) { aaLogMes ("STOPs  %08X %08X %08X %3d\n", pI2c->CR1, pI2c->CR2, pI2c->ISR, bspTsDelta (& now), 0) ; }
		result = I2C_ESTATE ;
	}
if (DBG_I2C) { aaLogMes ("STOPe  %08X %08X %08X %3d\n", pI2c->CR1, pI2c->CR2, pI2c->ISR, bspTsDelta (& now), 0) ; }
	return result ;
}

//----------------------------------------------------------------------
//	To send or receive data, the direction was specified to i2cMasterStart().
//	Non blocking mode (I2C_NOBLOCK flag in nextOrLast): returns after initiating the transfer, use interrupt to transfer bytes
// 		At the end of the transfer: the result is memorized and a signal is emitted to the task.
//	Blocking mode (default) : returns after the end of the transfer.
//	Several transfers can be made one after the other with I2C_NEXT, the last one must specify I2C_LAST.

i2cStatus_t		i2cMasterXfer		(i2cHandle_t hdl, uint8_t * pBuffer, uint32_t numByte, uint32_t nextOrLast)
{
	i2cIo_t		* pI2cIo = (i2cIo_t *) hdl ;
	I2C_TypeDef	* pI2c = pI2cIo->pI2c ;

	// Check if i2cMasterStart() already done
	if ((getFlagMode  (pI2cIo->flags) != I2C_FLAG_MASTER)   ||
		(getFlagState (pI2cIo->flags) != I2C_STATE_TXWAIT   &&
		 getFlagState (pI2cIo->flags) != I2C_STATE_RXWAIT1  &&
		 getFlagState (pI2cIo->flags) != I2C_STATE_RXWAIT))
	{
		return I2C_ESTATE ;
	}

	if (numByte == 0u)
	{
		return I2C_EFAIL ;
	}

	pI2cIo->pBuffer = pBuffer ;
	pI2cIo->size	= (int32_t) numByte ;
	pI2cIo->result	= I2C_EBUSY ;
	pI2cIo->taskId	= aaTaskSelfId () ;
	pI2cIo->count	= 0u ;		// Count of exchanged bytes
	aaSignalClear (pI2cIo->taskId, pI2cIo->signal) ;

	setFlagLast (& pI2cIo->flags, (nextOrLast & I2C_NEXT) == I2C_NEXT ? I2C_FLAG_NEXT : I2C_FLAG_LAST) ;

if(DBG_I2C) { aaLogMes ("XFER   %08X %08X %08X %3d\n", pI2c->CR1, pI2c->CR2, pI2c->ISR, bspTsDelta (& now), 0) ; }
	if (getFlagDir (pI2cIo->flags) == I2C_FLAG_TX)
	{
		// TX direction. i2cMasterStart() already done.
		setFlagState (& pI2cIo->flags, I2C_STATE_TX) ;

		// On 1st call after i2cMasterStart() TXIS is already present
		if ((pI2c->ISR & I2C_ISR_TXIS) != 0u)
		{
			// 1rst call after i2cMasterStart()
			if ((nextOrLast & I2C_LAST) == I2C_LAST  &&  numByte == 1u)
			{
				// Special case: Only one byte to send
				pI2c->CR2 = (pI2c->CR2 & ~(I2C_CR2_NBYTES | I2C_CR2_RELOAD)) ; // Clear reload without writing NBYTES
			}
		}
		else
		{
			// Not 1st call after i2cMasterStart()
			// Initialize NBYTES in CR2
			if (numByte > 255u)
			{
				//	We'll receive TCR interrupt
				LL_I2C_SetTransferSize (pI2c, 255) ;
			}
			else
			{
				if ((nextOrLast & I2C_LAST) == I2C_LAST)
				{
					// Last buffer of the message. Clear RELOAD and NBYTES at the same time
					// We'll receive TC interrupt
					uint32_t	reg ;

					reg = pI2c->CR2 ;
					reg &= ~(I2C_CR2_NBYTES | I2C_CR2_RELOAD) ;
					reg |= (uint32_t) (pI2cIo->size << I2C_CR2_NBYTES_Pos) ;
					pI2c->CR2 = reg ;
				}
				else
				{
					// Not last buffer, we'll receive TCR interrupt
					LL_I2C_SetTransferSize (pI2c, (uint32_t) pI2cIo->size) ;
				}
			}
		}

		// Enable interrupt sources
		pI2c->CR1 |= (I2C_CR1_ERRIE | I2C_CR1_TCIE | I2C_CR1_STOPIE | I2C_CR1_NACKIE | I2C_CR1_TXIE) ;
	}
	else
	{
		// RX direction. i2cMasterStart() already done.

		// On 1st call after i2cMasterStart() RXNE is already present
		if (getFlagState (pI2cIo->flags) == I2C_STATE_RXWAIT1)
		{
			// RXDR already full, read it
			* pI2cIo->pBuffer++ = (uint8_t) pI2c->RXDR ;
			pI2cIo->size-- ;
			pI2cIo->count++ ;

			// TC or TCR is already present, enable interrupts to enable TC/TCR handling
		}
		else
		{
			// Not 1st call after i2cMasterStart()
			// Initialize NBYTES in CR2
			if (numByte > 255u)
			{
				LL_I2C_SetTransferSize (pI2c, 255) ;
			}
			else
			{
				if ((nextOrLast & I2C_LAST) == I2C_LAST)
				{
					// Special case: must NACK the last byte. Clear RELOAD and NBYTES at the same time
					uint32_t	reg ;

					reg = pI2c->CR2 ;
					reg &= ~(I2C_CR2_NBYTES | I2C_CR2_RELOAD) ;
					reg |= (uint32_t) pI2cIo->size << I2C_CR2_NBYTES_Pos ;
					pI2c->CR2 = reg ;
				}
				else
				{
					LL_I2C_SetTransferSize (pI2c, (uint32_t) pI2cIo->size) ;
				}
			}
		}

		// Enable interrupt sources
		setFlagState (& pI2cIo->flags, I2C_STATE_RX) ;
		pI2c->CR1 |= (I2C_CR1_ERRIE | I2C_CR1_TCIE | I2C_CR1_STOPIE | I2C_CR1_NACKIE | I2C_CR1_RXIE) ;
	}

	if ((nextOrLast & I2C_NOBLOCK) == I2C_NOBLOCK)
	{
		return I2C_ENONE ;
	}
	aaSignalWait (pI2cIo->signal, NULL, AA_SIGNAL_OR, AA_INFINITE) ;
	return pI2cIo->result ;
}

//----------------------------------------------------------------------
// To send a message in one operation
// Returns I2X_EOK in case of success

i2cStatus_t		i2cMasterWrite		(i2cHandle_t hdl, uint32_t slaveAddress, uint8_t * pBuffer, uint32_t numByteToWrite)
{
	i2cStatus_t		result ;

	result = i2cMasterStart (hdl, slaveAddress, I2C_TX) ;
	if (result == I2C_EOK)
	{
		result = i2cMasterXfer (hdl, pBuffer, numByteToWrite, I2C_LAST) ;
		i2cMasterStop (hdl) ;
	}
	return result  ;
}

//----------------------------------------------------------------------
// To receive a message in one operation
// Returns I2X_EOK in case of success

i2cStatus_t		i2cMasterRead		(i2cHandle_t hdl, uint32_t slaveAddress, uint8_t * pBuffer, uint32_t numByteToRead)
{
	i2cStatus_t		result ;

	result = i2cMasterStart (hdl, slaveAddress, I2C_RX) ;
	if (result == I2C_EOK)
	{
		result = i2cMasterXfer (hdl, pBuffer, numByteToRead, I2C_LAST) ;
		i2cMasterStop (hdl) ;
	}
	return result  ;
}

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//	I2C Slave API
//----------------------------------------------------------------------
//	Configure slave own address
//	The slave is set in non addressable state

i2cStatus_t		i2cSlaveInit		(i2cHandle_t hdl, uint32_t ownAddress1)
{
	i2cIo_t		* pI2cIo = (i2cIo_t *) hdl ;
	I2C_TypeDef	* pI2c = pI2cIo->pI2c ;

	if (getFlagState (pI2cIo->flags) != I2C_STATE_IDLE)
	{
		// Slave init only allowed in idle state (after i2cConfigure())
		return I2C_ESTATE ;
	}

	// Clear interrupt sources
	pI2c->CR1 &= ~(I2C_CR1_ERRIE | I2C_CR1_TCIE | I2C_CR1_STOPIE | I2C_CR1_NACKIE | I2C_CR1_ADDRIE | I2C_CR1_RXIE | I2C_CR1_TXIE) ;
	LL_I2C_DisableOwnAddress1	(pI2c) ;
	LL_I2C_SetOwnAddress1		(pI2c, ownAddress1, LL_I2C_OWNADDRESS1_7BIT) ;
	setFlagMode (& pI2cIo->flags, I2C_FLAG_SLAVE) ;

	return I2C_ENONE ;
}

//----------------------------------------------------------------------
//	Set the slave in addressable state
//	Non blocking function if flags contain I2C_NOBLOCK
//	else blocking function until the slave is addressed by the master

i2cStatus_t 	i2cSlaveStart		(i2cHandle_t hdl, uint32_t flags)
{
	i2cIo_t		* pI2cIo = (i2cIo_t *) hdl ;
	I2C_TypeDef	* pI2c = pI2cIo->pI2c ;
	uint32_t	state = getFlagState (pI2cIo->flags) ;

	if ((getFlagMode (pI2cIo->flags) != I2C_FLAG_SLAVE) ||
		(state != I2C_STATE_IDLE  &&  state != I2C_STATE_TXWAIT  &&  state != I2C_STATE_RXWAIT))
	{
		return I2C_ESTATE ;
	}

	pI2cIo->result	= I2C_EBUSY ;
	pI2cIo->taskId	= aaTaskSelfId () ;
	aaSignalClear (pI2cIo->taskId, pI2cIo->signal) ;

	// Set slave in addressable state
	setFlagState (& pI2cIo->flags, I2C_STATE_ADDR) ;
	LL_I2C_EnableOwnAddress1 (pI2c) ;

	// Enable interrupt sources
if (DBG_I2C) { aaLogMes ("  strt %08X %08X %08X %3d\n", pI2c->CR1, pI2c->CR2, pI2c->ISR, bspTsDelta (& now), 0) ; }
	pI2c->CR1 |= (I2C_CR1_ERRIE | I2C_CR1_NACKIE | I2C_CR1_ADDRIE) ;

	if ((flags & I2C_NOBLOCK) != 0u)
	{
		return I2C_ENONE ;
	}
	aaSignalWait (pI2cIo->signal, NULL, AA_SIGNAL_OR, AA_INFINITE) ;
	return pI2cIo->result ;
}

//----------------------------------------------------------------------
//	Set the slave in transfer state: wait a transfer initiated by the master
//	Non blocking mode (I2C_NOBLOCK flag in nextOrLast): returns after initiating the transfer
//		At the end of the transfer the result is memorized and a signal is emitted to the task
//	Blocking mode (default): returns after the end of the transfer
//	nextOrLast must contain one of I2C_NEXT or I2C_LAST

i2cStatus_t		i2cSlaveXfer		(i2cHandle_t hdl, uint8_t * pBuffer, uint32_t numByte, uint32_t nextOrLast)
{
	i2cIo_t		* pI2cIo = (i2cIo_t *) hdl ;
	I2C_TypeDef	* pI2c = pI2cIo->pI2c ;
	uint32_t	state = getFlagState (pI2cIo->flags) ;

	// Check if i2cSlaveStart() already done
	if ((getFlagMode (pI2cIo->flags) != I2C_FLAG_SLAVE) ||
		(state != I2C_STATE_TXWAIT  &&  state != I2C_STATE_RXWAIT))
	{
		return I2C_ESTATE ;
	}

	if (numByte == 0u)
	{
		return I2C_EFAIL ;
	}

	pI2cIo->result	= I2C_EBUSY ;
	pI2cIo->pBuffer = pBuffer ;
	pI2cIo->size	= (int32_t) numByte ;
	pI2cIo->count	= 0u ;		// Count of exchanged bytes
	pI2cIo->taskId	= aaTaskSelfId () ;
	aaSignalClear (pI2cIo->taskId, pI2cIo->signal) ;

	setFlagLast (& pI2cIo->flags, (nextOrLast == I2C_NEXT) ? I2C_FLAG_NEXT : I2C_FLAG_LAST) ;

	// If this is the first transfer after slave addressing,
	// clear ADDR flag to generate an acknowledge to the Master
	if ((pI2c->ISR & I2C_ISR_ADDR) != 0u)
	{
		pI2c->ICR = I2C_ICR_ADDRCF ;

		// Clear a possible STOP flag
		// Because when the slave was TX, the transmission may be ended by NACK, without STOP.
		// The STOP is not always present: there is no stop when a ReSTART is used.
		if ((pI2c->ISR & I2C_ISR_STOPF) != 0u)
		{
			pI2c->ICR = I2C_ICR_STOPCF ;
		}
	}

if (DBG_I2C) { aaLogMes ("  XFER %08X %08X %08X %3d\n", pI2c->CR1, pI2c->CR2, pI2c->ISR, bspTsDelta (& now), 0) ; }
	if (getFlagState (pI2cIo->flags) == I2C_STATE_TXWAIT)
	{
		// Addressed as transmitter
		setFlagState (& pI2cIo->flags, I2C_STATE_TX) ;
		// Enable interrupt sources
		pI2c->CR1 |= (I2C_CR1_ERRIE | I2C_CR1_STOPIE | I2C_CR1_ADDRIE | I2C_CR1_NACKIE | I2C_CR1_TXIE) ;

	}
	else
	{
		// Addressed as receiver
		setFlagState (& pI2cIo->flags, I2C_STATE_RX) ;
		if (numByte == 1)
		{
			// Only 1 byte to receive, NACK this byte
			pI2c->CR2 |= I2C_CR2_NACK ;
		}
		// Enable interrupt sources
		pI2c->CR1 |= (I2C_CR1_ERRIE | I2C_CR1_STOPIE | I2C_CR1_ADDRIE | I2C_CR1_RXIE) ;
	}

	if ((nextOrLast & I2C_NOBLOCK) != 0u)
	{
		return I2C_ENONE ;
	}
	aaSignalWait (pI2cIo->signal, NULL, AA_SIGNAL_OR, AA_INFINITE) ;
	return pI2cIo->result ;
}

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//	I2C Interrupt handling
//----------------------------------------------------------------------

//	End of phase: update internal state and signal the task
//	This function is used only by interrupt handlers

static	void	i2cEnd	(i2cIo_t * pI2cIo, i2cStatus_t result, uint32_t state)
{
	I2C_TypeDef		* pI2c = pI2cIo->pI2c ;

	// Disable all interrupt sources
	pI2c->CR1 &= ~(I2C_CR1_ERRIE | I2C_CR1_TCIE | I2C_CR1_STOPIE | I2C_CR1_NACKIE | I2C_CR1_ADDRIE | I2C_CR1_RXIE | I2C_CR1_TXIE) ;

	setFlagState (& pI2cIo->flags, state) ;
	pI2cIo->result = result ;

	// Signal the task
	if (pI2cIo->taskId != AA_INVALID_TASK)
	{
		aaSignalSend (pI2cIo->taskId, pI2cIo->signal) ;
	}
	pI2cIo->taskId = AA_INVALID_TASK ;
}

//----------------------------------------------------------------------

void I2Cx_EV_IRQHandler (i2cIo_t * pI2cIo) ;

void I2Cx_EV_IRQHandler (i2cIo_t * pI2cIo)
{
	I2C_TypeDef * pI2c ;
	uint32_t	isr, cr1, cr2, reg ;

	pI2c = pI2cIo->pI2c ;
	isr  = pI2c->ISR ;
	cr1	 = pI2c->CR1 ;
	cr2	 = pI2c->CR2 ;
	(void) cr1 ;
	(void) cr2 ;

	if (getFlagMode (pI2cIo->flags) == I2C_FLAG_MASTER)
	{
		//------------------------------------------------------------------
		// Master mode
		//------------------------------------------------------------------
		if (getFlagState (pI2cIo->flags) == I2C_STATE_ADDR)
		{
			// Slave addressing phase

			if ((isr & I2C_ISR_TXIS) != 0u)
			{
				// Slave addressing phase OK, ready for data transmission
if(DBG_I2C) { aaLogMes ("TXIS   %08X %08X %08X %3d\n", cr1, cr2, isr, bspTsDelta (& now), 0) ; }
				i2cEnd (pI2cIo, I2C_EOK, I2C_STATE_TXWAIT) ;
			}

			if ((isr & I2C_ISR_RXNE) != 0u)
			{
				// Slave addressing phase OK, ready for data reception
				// BEWARE: RXDR not read, but I2C_ISR_RXNE has disappeared from ISR !!!
if(DBG_I2C) { aaLogMes ("RXNE   %08X %08X %08X %3d\n", cr1, cr2, isr, bspTsDelta (& now), 0) ; }
				i2cEnd (pI2cIo, I2C_EOK, I2C_STATE_RXWAIT1) ;
			}

			if ((isr & I2C_ISR_STOPF) != 0u)
			{
				// STOPF in addressing phase is used by i2cMasterProbeDevice().
				// And when a slave doesn't ACK the addressing phase.
if(DBG_I2C) { aaLogMes ("STOP   %08X %08X %08X %3d\n", cr1, cr2, isr, bspTsDelta (& now), 0) ; }
				pI2c->ICR = I2C_ICR_STOPCF ;
				if ((isr & I2C_ISR_NACKF) != 0u)
				{
					pI2c->ICR = I2C_ICR_NACKCF ;
					i2cEnd (pI2cIo, I2C_EAF, I2C_STATE_IDLE) ;
				}
				else
				{
					i2cEnd (pI2cIo, I2C_EOK, I2C_STATE_IDLE) ;
				}
			}
		}
		//------------------------------------------------------------------
		else if (getFlagState (pI2cIo->flags) == I2C_STATE_TX)
		{
			// Master Xfer state, TX

			if (((isr & I2C_ISR_NACKF) != 0u)  &&  ((pI2c->CR1 & I2C_CR1_NACKIE) != 0u)) 	// Nack received
			{
if(DBG_I2C) { aaLogMes ("NACK   %08X %08X %08X %3d\n", cr1, cr2, isr, bspTsDelta (& now), 0) ; }
				// Disable NACK interrupt, it is automatically followed by a STOP condition
				pI2c->CR1 &= ~I2C_CR1_NACKIE ;
			}

			if ((isr & I2C_ISR_STOPF) != 0u)
			{
if(DBG_I2C) { aaLogMes ("STOP   %08X %08X %08X %3d\n", cr1, cr2, isr, bspTsDelta (& now), 0) ; }
				pI2c->ICR = I2C_ICR_STOPCF ;

				if ((isr & I2C_ISR_TXE) == 0u)
				{
					// Last byte in TXDR not sent
					pI2cIo->size++ ;
					pI2cIo->count-- ;
					pI2c->ISR |= I2C_ISR_TXE ;
				}

				if ((isr & I2C_ISR_NACKF) != 0u)
				{
					// The slave nacked the byte => end of TX
					pI2c->ICR = I2C_ICR_NACKCF ;
					i2cEnd (pI2cIo, (pI2cIo->size == 0) ? I2C_EOK : I2C_EAF, I2C_STATE_IDLE) ;
				}
				else
				{
					i2cEnd (pI2cIo, I2C_EOK, I2C_STATE_IDLE) ;
				}
			}

			if ((isr & I2C_ISR_TXIS) != 0u)	// Ready to transmit
			{
if(DBG_I2C) { aaLogMes ("TXIS   %08X %08X %08X %3d\n", cr1, cr2, isr, bspTsDelta (& now), 0) ; }
				pI2c->TXDR = * pI2cIo->pBuffer++ ;
				pI2cIo->size-- ;
				pI2cIo->count++ ;
			}

			if ((isr & I2C_ISR_TCR) != 0u)	// Transfer Complete Reload
			{
if(DBG_I2C) { aaLogMes ("TCR    %08X %08X %08X %3d   %d\n", cr1, cr2, isr, bspTsDelta (& now), (uintptr_t) pI2cIo->size) ; }
				if (pI2cIo->size > 255)
				{
					// Configure to get TCR interrupt
					LL_I2C_SetTransferSize (pI2c, 255) ;
				}
				else if (pI2cIo->size > 0)
				{
					if (getFlagLast (pI2cIo->flags) == I2C_FLAG_LAST)
					{
						// Last part of last buffer: configure to get TC interrupt
						reg = pI2c->CR2 ;
						reg &= ~(I2C_CR2_NBYTES | I2C_CR2_RELOAD) ;
						reg |= (uint32_t) pI2cIo->size << I2C_CR2_NBYTES_Pos ;
						pI2c->CR2 = reg ;	// Mandatory: only one write of CR2
					}
					else
					{
						// Next buffer: configure to get TCR interrupt
						LL_I2C_SetTransferSize (pI2c, (uint32_t) pI2cIo->size) ;
					}
				}
				else
				{
					// Size == 0
					i2cEnd (pI2cIo, I2C_EOK, I2C_STATE_TXWAIT) ;			// End of i2cMasterXfer(NEXT)
				}
			}

			// End of last xfer ?
			if ((isr & I2C_ISR_TC) != 0u)	// Transfer Complete
			{
if(DBG_I2C) { aaLogMes ("TC     %08X %08X %08X %3d  %d\n", cr1, cr2, isr, bspTsDelta (& now), pI2cIo->count) ; }
				i2cEnd (pI2cIo, I2C_EOK, I2C_STATE_TXWAIT) ;				// End of i2cMasterXfer(LAST)
			}
		}
		//------------------------------------------------------------------
		else if (getFlagState (pI2cIo->flags) == I2C_STATE_RX)
		{
			// Master Xfer state, RX
			// As the master controls the clock, he always receives the quantity he requests.

			if ((isr & I2C_ISR_RXNE) != 0u)	// Data received
			{
if(DBG_I2C) { aaLogMes ("RXNE   %08X %08X %08X %3d  %d\n", cr1, cr2, isr, bspTsDelta (& now), (uintptr_t) pI2cIo->size) ; }
				* pI2cIo->pBuffer++ = (uint8_t) pI2c->RXDR ;
				pI2cIo->size-- ;
				pI2cIo->count++ ;
			}

			if ((isr & I2C_ISR_TCR) != 0u)	// Transfer Complete Reload
			{
if(DBG_I2C) { aaLogMes ("TCR    %08X %08X %08X %3d  %d\n", cr1, cr2, isr, bspTsDelta (& now), (uintptr_t) pI2cIo->size) ; }
				if (pI2cIo->size > 255)
				{
					// Configure to get TCR interrupt
					LL_I2C_SetTransferSize (pI2c, 255) ;
				}
				else if (pI2cIo->size > 0)
				{
					if (getFlagLast (pI2cIo->flags) == I2C_FLAG_LAST)
					{
						// Last part of 'last' buffer: configure to get TC interrupt
						reg = pI2c->CR2 ;
						reg &= ~(I2C_CR2_NBYTES | I2C_CR2_RELOAD) ;
						reg |= (uint32_t) pI2cIo->size << I2C_CR2_NBYTES_Pos ;
						pI2c->CR2 = reg ;
					}
					else
					{
						// 'Next' buffer: configure to get TCR interrupt
						LL_I2C_SetTransferSize (pI2c, (uint32_t) pI2cIo->size) ;
					}
				}
				else
				{
					// Size == 0
					i2cEnd (pI2cIo, I2C_EOK, I2C_STATE_RXWAIT) ;			// End of i2cMasterXfer(NEXT)
				}
			}

			// End of last xfer ?
			if ((isr & I2C_ISR_TC) != 0u)	// Transfer Complete
			{
if(DBG_I2C) { aaLogMes ("TC     %08X %08X %08X %3d  %d\n", cr1, cr2, isr, bspTsDelta (& now), (uintptr_t) pI2cIo->size) ; }
				i2cEnd (pI2cIo, I2C_EOK, I2C_STATE_RXWAIT) ;				// End of i2cMasterXfer(LAST)
			}
		}
	}
	else
	{
		//------------------------------------------------------------------
		// Slave mode
		//------------------------------------------------------------------
		if ((isr & I2C_ISR_ADDR) != 0u)
		{
			// Slave addressed: START or ReSTART
if(DBG_I2C) { aaLogMes ("  ADDR %08X %08X %08X %3d\n", cr1, cr2, isr, bspTsDelta (& now), 0) ; }
			if ((isr & I2C_ISR_DIR) == 0u)
			{
				setFlagDir (& pI2cIo->flags, I2C_FLAG_RX) ;
				i2cEnd (pI2cIo, I2C_ESLAVERX, I2C_STATE_RXWAIT) ;	// The slave is receiver
			}
			else
			{
				setFlagDir (& pI2cIo->flags, I2C_FLAG_TX) ;
				if ((isr & I2C_ISR_TXIS) == 0u)
				{
					// Last byte in TXDR not sent, flush
					pI2cIo->size++ ;
					pI2cIo->count-- ;
					pI2c->ISR |= I2C_ISR_TXE ;
				}
				pI2c->ICR   = I2C_ICR_NACKCF ;		// Clear NACK flag (Can appears before a ReStart if the slave was TX)
				i2cEnd (pI2cIo, I2C_ESLAVETX, I2C_STATE_TXWAIT) ;	// The slave is transmitter
			}
		}
		//------------------------------------------------------------------
		else if (getFlagState (pI2cIo->flags) == I2C_STATE_TX)
		{
			// Slave Xfer state, TX

			if (((isr & I2C_ISR_NACKF) != 0u)  &&  ((pI2c->CR1 & I2C_CR1_NACKIE) != 0u))		// NACK received, end of slave TX
			{
if(DBG_I2C) { aaLogMes ("  NACK %08X %08X %08X %3d  %d\n", cr1, cr2, isr, bspTsDelta (& now), (uintptr_t) pI2cIo->size) ; }
				// Disable NACK interrupt, it is automatically followed by a STOP or ADDR (for ReSTART condition)
				pI2c->CR1 &= ~I2C_CR1_NACKIE ;
			}

			if ((isr & I2C_ISR_TXIS) != 0u)	// Ready to transmit
			{
if(DBG_I2C) { aaLogMes ("  TXIS %08X %08X %08X %3d  %d\n", cr1, cr2, isr, bspTsDelta (& now), (uintptr_t) pI2cIo->size) ; }
				if (pI2cIo->size > 0)
				{
					pI2c->TXDR = * pI2cIo->pBuffer++ ;
					pI2cIo->size-- ;
					pI2cIo->count++ ;

					if (pI2cIo->size == 0  &&  getFlagLast (pI2cIo->flags) != I2C_FLAG_LAST)
					{
						// End of xfer (not the last one)
						i2cEnd (pI2cIo, I2C_EOK, I2C_STATE_TXWAIT) ;	// The whole block has been sent
					}
				}
				else
				{
					// The master requests more bytes than the slave has.
					// This is not a problem, the slave has sent his entire message.
					pI2c->TXDR = 0u ;
					pI2cIo->size-- ;
					pI2cIo->count++ ;
				}
			}

			if ((isr & I2C_ISR_STOPF) != 0u)
			{
if(DBG_I2C) { aaLogMes ("  STOP %08X %08X %08X %3d\n", cr1, cr2, isr, bspTsDelta (& now), 0) ; }
				pI2c->ICR   = I2C_ICR_STOPCF ;		// Clear STOP flag
				pI2c->ICR   = I2C_ICR_NACKCF ;		// Clear NACK flag

				// Flush TX
				if ((isr & I2C_ISR_TXE) == 0u)
				{
					// Last byte in TXDR not sent
					pI2cIo->size++ ;
					pI2cIo->count-- ;
					pI2c->ISR |= I2C_ISR_TXE ;
				}

				i2cEnd (pI2cIo, (pI2cIo->size <= 0) ? I2C_EOK : I2C_EAF, I2C_STATE_IDLE) ;
			}
		}
		//------------------------------------------------------------------
		else if (getFlagState (pI2cIo->flags) == I2C_STATE_RX)
		{
			// Slave Xfer state, RX

			if ((isr & I2C_ISR_RXNE) != 0u)
			{
				if (pI2cIo->size != 0)
				{
					if (pI2cIo->size == 2  &&  getFlagLast (pI2cIo->flags) == I2C_FLAG_LAST)
					{
						// The next byte is the last byte to receive (end of the last xfer)
						pI2c->CR2 |= I2C_CR2_NACK ;		// NACK this byte
if(DBG_I2C) { aaLogMes ("  RXNEn%08X %08X %08X %3d  %d\n", cr1, cr2, isr, bspTsDelta (& now), (uintptr_t) pI2cIo->size) ; }
					}
					else
					{
if(DBG_I2C) { aaLogMes ("  RXNE %08X %08X %08X %3d  %d\n", cr1, cr2, isr, bspTsDelta (& now), (uintptr_t) pI2cIo->size) ; }
					}

					* pI2cIo->pBuffer++ = (uint8_t) pI2c->RXDR ;
					pI2cIo->size-- ;
					pI2cIo->count++ ;

					if (pI2cIo->size == 0  &&  getFlagLast (pI2cIo->flags) != I2C_FLAG_LAST)
					{
if(DBG_I2C) { aaLogMes ("  RXNEl%08X %08X %08X %3d  %d\n", cr1, cr2, isr, bspTsDelta (& now), (uintptr_t) pI2cIo->size) ; }
						// End of xfer (not the last one)
						i2cEnd (pI2cIo, I2C_EOK, I2C_STATE_RXWAIT) ;	// The whole message has been received
					}
				}
				else
				{
					// The master send more bytes than the slave has. Must not happen.
if(DBG_I2C) { aaLogMes ("  X2   %08X %08X %08X %3d  %d *************\n", cr1, cr2, isr, bspTsDelta (& now), (uintptr_t) pI2cIo->size) ; }
					volatile uint8_t dd = (uint8_t) pI2c->RXDR ;		// Trash the byte
					(void) dd ;
					pI2cIo->size-- ;
				}
			}

			if ((isr & I2C_ISR_STOPF) != 0u)
			{
if(DBG_I2C) { aaLogMes ("  STOPr %08X %08X %08X %3d\n", cr1, cr2, isr, bspTsDelta (& now), 0) ; }
				pI2c->ICR   = I2C_ICR_STOPCF ;		// Clear STOP flag
				pI2c->ICR   = I2C_ICR_NACKCF ;		// Clear NACK flag

				// Get last received data if any
				if ((pI2c->ISR & I2C_ISR_RXNE) != 0u)
				{
					if (pI2cIo->size > 0)
					{
						* pI2cIo->pBuffer++ = (uint8_t) pI2c->RXDR ;
						pI2cIo->size-- ;
						pI2cIo->count++ ;
					}
					else
					{
						// The master sent more bytes than the slave has. Must not happen.
						volatile uint8_t dd = (uint8_t) pI2c->RXDR ;
if(DBG_I2C) { aaLogMes ("  X3   %08X %08X %08X %3d *************\n", cr1, cr2, isr, bspTsDelta (& now), 0) ; }
						(void) dd ;
						pI2cIo->size-- ;
					}
				}

				// Return I2C_EAF if we received more bytes than the buffer length
				// e.g the master did not take into account the slave NACK
				i2cEnd (pI2cIo, (pI2cIo->size >= 0) ? I2C_EOK : I2C_EAF, I2C_STATE_IDLE) ;
			}
		}
	}
}

//----------------------------------------------------------------------

#if (WITH_I2C_1 == 1)
void I2C1_EV_IRQHandler (void) ;
void I2C1_EV_IRQHandler (void)
{
	aaIntEnter () ;
	I2Cx_EV_IRQHandler (& i2cIo1) ;
	aaIntExit () ;
}
#endif	// WITH_I2C_1

#if (WITH_I2C_2 == 1)
void I2C2_EV_IRQHandler (void) ;
void I2C2_EV_IRQHandler (void)
{
	aaIntEnter () ;
	I2Cx_EV_IRQHandler (& i2cIo2) ;
	aaIntExit () ;
}
#endif	// WITH_I2C_2

#if (WITH_I2C_3 == 1)
void I2C3_EV_IRQHandler (void) ;
void I2C3_EV_IRQHandler (void)
{
	aaIntEnter () ;
	I2Cx_EV_IRQHandler (& i2cIo3) ;
	aaIntExit () ;
}
#endif	// WITH_I2C_3

#if (WITH_I2C_4 == 1)
void I2C4_EV_IRQHandler (void) ;
void I2C4_EV_IRQHandler (void)
{
	aaIntEnter () ;
	I2Cx_EV_IRQHandler (& i2cIo4) ;
	aaIntExit () ;
}
#endif	// WITH_I2C_4

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//	Managing the errors is rudimentary, because there were no errors observed.
//	'ARLO' errors have been seen when there are bad contacts on the I2C bus.

void I2Cx_ER_IRQHandler (i2cIo_t * pI2cIo) ;

void I2Cx_ER_IRQHandler (i2cIo_t * pI2cIo)
{
	I2C_TypeDef * pI2c ;
	uint32_t	isr, cr1, cr2 ;

	pI2c = pI2cIo->pI2c ;
	isr  = pI2c->ISR ;
	cr1	 = pI2c->CR1 ;
	cr2	 = pI2c->CR2 ;
	(void) cr1 ;
	(void) cr2 ;

	if (getFlagMode (pI2cIo->flags) == I2C_FLAG_MASTER)
	{
		// Master mode
if (DBG_I2C) { aaLogMes ("ERR    %08X %08X %08X %3d  %d\n", cr1, cr2, isr, bspTsDelta (& now), (uintptr_t) pI2cIo->size) ; }
		if ((isr & I2C_ISR_ARLO) != 0)
		{
			pI2c->ICR |= I2C_ICR_ARLOCF ;
			i2cEnd (pI2cIo, I2C_EARLO, I2C_STATE_IDLE) ;
		}
		if ((isr & I2C_ISR_BERR) != 0)
		{
			pI2c->ICR |= I2C_ICR_BERRCF ;
			i2cEnd (pI2cIo, I2C_EBERR, I2C_STATE_IDLE) ;
		}
		if ((isr & I2C_ISR_OVR) != 0)
		{
			pI2c->ICR |= I2C_ICR_OVRCF ;
			i2cEnd (pI2cIo, I2C_EOVR, I2C_STATE_IDLE) ;
		}
	}
	else
	{
		// Slave mode
if(DBG_I2C) { aaLogMes ("  ERR  %08X %08X %08X %3d  %d\n", cr1, cr2, isr, bspTsDelta (& now), (uintptr_t) pI2cIo->size) ; }
//		if (getFlagState (pI2cIo->flags) == I2C_STATE_ADDR)
		{
			if ((isr & I2C_ISR_ARLO) != 0)
			{
				pI2c->ICR |= I2C_ICR_ARLOCF ;
//				i2cEnd (pI2cIo, I2C_EARLO, I2C_STATE_IDLE) ;
			}
			if ((isr & I2C_ISR_BERR) != 0)
			{
				pI2c->ICR |= I2C_ICR_BERRCF ;
//				i2cEnd (pI2cIo, I2C_EBERR, I2C_STATE_IDLE) ;
			}
			if ((isr & I2C_ISR_OVR) != 0)
			{
				pI2c->ICR |= I2C_ICR_OVRCF ;
//				i2cEnd (pI2cIo, I2C_EOVR, I2C_STATE_IDLE) ;
			}
		}
	}
}

//----------------------------------------------------------------------

#if (WITH_I2C_1 == 1)
void I2C1_ER_IRQHandler(void) ;
void I2C1_ER_IRQHandler(void)
{
	aaIntEnter () ;
	I2Cx_ER_IRQHandler (& i2cIo1) ;
	aaIntExit () ;
}
#endif	// WITH_I2C_1

#if (WITH_I2C_2 == 1)
void I2C2_ER_IRQHandler (void) ;
void I2C2_ER_IRQHandler (void)
{
	aaIntEnter () ;
	I2Cx_ER_IRQHandler (& i2cIo2) ;
	aaIntExit () ;
}
#endif	// WITH_I2C_2

#if (WITH_I2C_3 == 1)
void I2C3_ER_IRQHandler (void) ;
void I2C3_ER_IRQHandler (void)
{
	aaIntEnter () ;
	I2Cx_ER_IRQHandler (& i2cIo3) ;
	aaIntExit () ;
}
#endif	// WITH_I2C_3

#if (WITH_I2C_4 == 1)
void I2C4_ER_IRQHandler (void) ;
void I2C4_ER_IRQHandler (void)
{
	aaIntEnter () ;
	I2Cx_ER_IRQHandler (& i2cIo4) ;
	aaIntExit () ;
}
#endif	// WITH_I2C_4

//----------------------------------------------------------------------
