/*
----------------------------------------------------------------------

	AdAstra - Real Time Kernel

	Alain Chebrou

	gpiobasic.c	Some routine to manage basic GPIO and RCC features

	When		Who	What
	17/03/18	ac	Creation
	16/01/20	ac	Port to HAL drivers 1.6.0 ST changed the register bits and register names (BSSRH/BSSRL), sigh...
	04/03/21	ac	In H7 HAL V1.10.0 RCC_GetXxxClockFreq() functions are now static inside stm32l4xx_ll_rcc.c, Sigh...
					So create public rccGetXxxClockFreq()
	01/02/22	ac	rccbasic split from gpiobasic

----------------------------------------------------------------------
*/
/*
----------------------------------------------------------------------
	Copyright (c) 2013-2023 Alain Chebrou - AdAstra-Soft . com
	All rights reserved.

	This file is part of the AdAstra Real Time Kernel distribution.

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions are met:

	- Redistributions of source code must retain the above copyright
	  notice, this list of conditions and the following disclaimer.

	- Redistributions in binary form must reproduce the above copyright
	  notice, this list of conditions and the following disclaimer in the
	  documentation and/or other materials provided with the distribution.

	- Neither the name of the copyright holder nor the
	  names of its contributors may be used to endorse or promote products
	  derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
	DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
	DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
	(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
	ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
----------------------------------------------------------------------
*/

#include	"aa.h"
#include	"aakernel.h"
#include	"gpiobasic.h"

#include	"stm32h7xx_ll_bus.h"
#include	"stm32h7xx_ll_rcc.h"

//----------------------------------------------------------------------
// Since HAL V1.10.0 RCC_GetXxxClockFreq() functions are static inside stm32H7xx_ll_rcc.c

uint32_t	rccGetSystemClockFreq	(void)
{
	// SYSTEM clock frequency
	return SystemCoreClock ;
}

uint32_t	rccGetHCLKClockFreq		(void)
{
  // HCLK clock frequency
  return LL_RCC_CALC_HCLK_FREQ (SystemCoreClock, LL_RCC_GetAHBPrescaler());
}

uint32_t	rccGetPCLK1ClockFreq	(void)
{
  // PCLK1 clock frequency
  return LL_RCC_CALC_PCLK1_FREQ (rccGetHCLKClockFreq (), LL_RCC_GetAPB1Prescaler());
}

uint32_t	rccGetPCLK2ClockFreq	(void)
{
  // PCLK2 clock frequency
  return LL_RCC_CALC_PCLK2_FREQ (rccGetHCLKClockFreq (), LL_RCC_GetAPB2Prescaler());
}

uint32_t	rccGetPCLK3ClockFreq	(void)
{
  // PCLK3 clock frequency
  return LL_RCC_CALC_PCLK3_FREQ (rccGetHCLKClockFreq (), LL_RCC_GetAPB3Prescaler());
}

uint32_t	rccGetPCLK4ClockFreq	(void)
{
  // PCLK4 clock frequency
  return LL_RCC_CALC_PCLK4_FREQ (rccGetHCLKClockFreq (), LL_RCC_GetAPB4Prescaler());
}

//----------------------------------------------------------------------

#define	OFFSET_MASK		0x0000FFFFu

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//	RCC handling
//----------------------------------------------------------------------
// Peripheral clock enable
// Example:
// rccEnableClock (gpioPortFromName (pPinDesc->port), 1u << (pPinDesc->port - GPIO_PORTA)) ;
// rccEnableClock (I2C3, RCC_APB1LENR_I2C3EN) ;

void	rccEnableClock (void * deviceAddress, uint32_t enableBitMask)
{
	volatile uint32_t	reg ;

	switch ((uintptr_t) deviceAddress & ~OFFSET_MASK)
	{
		// APB buses
		case D1_APB1PERIPH_BASE:
			RCC->APB3ENR |= enableBitMask ;		// RCC_APB3ENR_LTDCEN
			reg = RCC->APB3ENR ;				// Delay after an RCC peripheral clock enabling
			break ;

		case D2_APB1PERIPH_BASE:
			if (((uint32_t) deviceAddress & OFFSET_MASK) >= CRS_BASE)
			{
				RCC->APB1HENR |= enableBitMask ;// RCC_APB1HENR_CRSEN
				reg = RCC->APB1HENR ;			// Delay after an RCC peripheral clock enabling
			}
			else
			{
				RCC->APB1LENR |= enableBitMask ;// RCC_APB1LENR_TIM2EN
				reg = RCC->APB1LENR ;			// Delay after an RCC peripheral clock enabling
			}
			break ;

		case D3_APB1PERIPH_BASE:
			RCC->APB4ENR |= enableBitMask ;		// RCC_APB4ENR_SAI4EN
			reg = RCC->APB4ENR ;				// Delay after an RCC peripheral clock enabling
			break ;

		// AHB buses
		case D1_AHB1PERIPH_BASE:
			RCC->AHB3ENR |= enableBitMask ;		// RCC_AHB3ENR_MDMAEN
			reg = RCC->AHB3ENR ;				// Delay after an RCC peripheral clock enabling
			break ;

		case D2_AHB1PERIPH_BASE:
			RCC->AHB1ENR |= enableBitMask ;		// RCC_AHB1ENR_DMA1EN
			reg = RCC->AHB1ENR ;				// Delay after an RCC peripheral clock enabling
			break ;

		case D2_AHB2PERIPH_BASE:
			RCC->AHB2ENR |= enableBitMask ;		// RCC_AHB2ENR_DCMIEN
			reg = RCC->AHB2ENR ;				// Delay after an RCC peripheral clock enabling
			break ;

		case D3_AHB1PERIPH_BASE:
			RCC->AHB4ENR |= enableBitMask ;		// RCC_AHB4ENR_GPIOAEN
			reg = RCC->AHB4ENR ;				// Delay after an RCC peripheral clock enabling
			break ;

		default:
			AA_ASSERT (0) ;
			break ;
	}
	(void) reg ;
}

//----------------------------------------------------------------------
// Peripheral clock disable

void	rccDisableClock (void * deviceAddress, uint32_t enableBitMask)
{
	switch ((uintptr_t) deviceAddress & ~OFFSET_MASK)
	{
		// APB buses
		case D1_APB1PERIPH_BASE:
			RCC->APB3ENR &= ~enableBitMask ;			// RCC_APB3ENR_LTDCEN
			break ;

		case D2_APB1PERIPH_BASE:
			if (((uint32_t) deviceAddress & OFFSET_MASK) >= CRS_BASE)
			{
				RCC->APB1HENR &= ~enableBitMask ;		// RCC_APB1HENR_CRSEN
			}
			else
			{
				RCC->APB1LENR &= ~enableBitMask ;		// RCC_APB1LENR_TIM2EN
			}
			break ;

		case D3_APB1PERIPH_BASE:
			RCC->APB4ENR &= ~enableBitMask ;			// RCC_APB4ENR_SAI4EN
			break ;

		// AHB buses
		case D1_AHB1PERIPH_BASE:
			RCC->AHB3ENR &= ~enableBitMask ;			// RCC_AHB3ENR_MDMAEN
			break ;

		case D2_AHB1PERIPH_BASE:
			RCC->AHB1ENR &= ~enableBitMask ;			// RCC_AHB1ENR_DMA1EN
			break ;

		case D2_AHB2PERIPH_BASE:
			RCC->AHB2ENR &= ~enableBitMask ;			// RCC_AHB2ENR_DCMIEN
			break ;

		case D3_AHB1PERIPH_BASE:
			RCC->AHB4ENR &= ~enableBitMask ;			// RCC_AHB4ENR_GPIOAEN
			break ;

		default:
			AA_ASSERT (0) ;
			break ;
	}
}

//----------------------------------------------------------------------
//	Most devices except TIMER and LPTIMER
//	Complicated peripherals are not present: USB, SDMMC, SAI, QSPI...

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wpedantic"		// case label is not an integer constant expression

uint32_t	rccGetClockFrequency (void * deviceAddress)
{
	uint32_t	devClock ;

	switch ((uintptr_t) deviceAddress)
	{
		case I2C1_BASE:
		case I2C2_BASE:
		case I2C3_BASE:
			devClock = LL_RCC_GetI2CClockFreq (LL_RCC_I2C123_CLKSOURCE) ;
			break ;

#if (defined I2C4)
		case I2C4_BASE:
			devClock = LL_RCC_GetI2CClockFreq (LL_RCC_I2C4_CLKSOURCE) ;
			break ;
#endif

		case SPI1_BASE:
		case SPI2_BASE:
		case SPI3_BASE:
			devClock = LL_RCC_GetSPIClockFreq (LL_RCC_SPI123_CLKSOURCE) ;
			break ;

		case SPI4_BASE:
		case SPI5_BASE:
			devClock = LL_RCC_GetSPIClockFreq (LL_RCC_SPI45_CLKSOURCE) ;
			break ;

		case SPI6_BASE:
			devClock = LL_RCC_GetSPDIFClockFreq (LL_RCC_SPI6_CLKSOURCE) ;
			break ;

		case USART1_BASE:
		case USART6_BASE:
			devClock = LL_RCC_GetUSARTClockFreq (LL_RCC_USART16_CLKSOURCE) ;
			break ;

		case USART2_BASE:
		case USART3_BASE:
		case UART4_BASE:
		case UART5_BASE:
		case UART7_BASE:
		case UART8_BASE:
			devClock = LL_RCC_GetUSARTClockFreq (LL_RCC_USART234578_CLKSOURCE) ;
			break ;

		case LPUART1_BASE:
			devClock = LL_RCC_GetLPUARTClockFreq (LL_RCC_LPUART1_CLKSOURCE) ;
			break ;

		case ADC1_BASE:
			devClock = LL_RCC_GetADCClockFreq (LL_RCC_ADC_CLKSOURCE) ;
			break ;

		case RNG_BASE:
			devClock = LL_RCC_GetRNGClockFreq (LL_RCC_RNG_CLKSOURCE) ;
			break ;
#if (0)
		case SAI1_BASE:
			devClock = LL_RCC_GetSAIClockFreq (LL_RCC_SAI1_CLKSOURCE) ;
			break ;

#if (defined LL_RCC_SAI23_CLKSOURCE)
		case SAI2_BASE:
		case SAI3_BASE:
			devClock = LL_RCC_GetSAIClockFreq (LL_RCC_SAI23_CLKSOURCE) ;
			break ;
#endif

#if (defined LL_RCC_SAI2A_CLKSOURCE)
		case SAI2_Block_A_BASE:
			devClock = LL_RCC_GetSAIClockFreq (LL_RCC_SAI2A_CLKSOURCE) ;
			break ;
#endif

#if (defined LL_RCC_SAI2ABCLKSOURCE)
		case SAI2_Block_B_BASE:
			devClock = LL_RCC_GetSAIClockFreq (LL_RCC_SAI2B_CLKSOURCE) ;
			break ;
#endif

#if (defined LL_RCC_SAI4A_CLKSOURCE)
		case SAI4_Block_A_BASE:
			devClock = LL_RCC_GetSAIClockFreq (LL_RCC_SAI4A_CLKSOURCE) ;
			break ;
#endif

#if (defined LL_RCC_SAI4B_CLKSOURCE)
		case SAI4_Block_B_BASE:
			devClock = LL_RCC_GetSAIClockFreq (LL_RCC_SAI4B_CLKSOURCE) ;
			break ;
#endif

		case SDMMC1_BASE:
			devClock = LL_RCC_GetSDMMCClockFreq (LL_RCC_SDMMC_CLKSOURCE) ;
			break ;

		case CEC_BASE:
			devClock = LL_RCC_GetCECClockFreq (LL_RCC_RNG_CLKSOURCE) ;
			break ;

		case USB_OTG_FS_BASE:
			devClock = LL_RCC_GetUSBClockFreq (LL_RCC_USB_CLKSOURCE) ;
			break ;

		case DFSDM1_Filter0_BASE:
			// Use filter 0 to obtain the frequency of the corresponding DFSDM
			devClock = LL_RCC_GetDFSDMClockFreq (LL_RCC_DFSDM1_CLKSOURCE) ;
			break ;

#if defined(DFSDM2_BASE)
		case DFSDM2_Filter0_BASE:
			// Use filter 0 to obtain the frequency of the corresponding DFSDM
			devClock = LL_RCC_GetDFSDMClockFreq (LL_RCC_DFSDM2_CLKSOURCE) ;
			break ;
#endif

#if defined(DSI)		// H747
		case DSI_BASE:
			devClock = LL_RCC_GetDSIClockFreq (LL_RCC_DSI_CLKSOURCE) ;
			break ;
#endif

		case SPDIFRX_BASE:
			devClock = LL_RCC_GetSPDIFClockFreq (LL_RCC_SPDIF_CLKSOURCE) ;
			break ;
		case SWPMI1_BASE:
			devClock = LL_RCC_GetSWPClockFreq (LL_RCC_SWP_CLKSOURCE) ;
			break ;

		case FDCAN1_BASE:
		case FDCAN2_BASE:
			devClock = LL_RCC_GetFDCANClockFreq (LL_RCC_FDCAN_CLKSOURCE) ;
			break ;
/*
		case FMC_BASE:
			devClock = LL_RCC_GetFMCClockFreq (LL_RCC_FMC_CLKSOURCE) ;
			break ;
*/
#endif	// 0

		default:
			AA_ASSERT (0) ;
			devClock = 0 ;
			break ;
	}
	return devClock ;
}

#pragma GCC diagnostic pop

//----------------------------------------------------------------------
//	Get TIMER and LPTIMER clock frequency
//	Timer clock depends on APB bus prescaler value (=> PCLK1 or PCLK2)
//	Timer clock may be PCLKx peripherals frequency x1 or x2 or x4

uint32_t	rccGetTimerClockFrequency (void * deviceAddress)
{
	uint32_t	D2PPRE ;
	uint32_t	TIMPRE ;
	uint32_t	timClock ;

	if (deviceAddress == LPTIM1)
	{
		return LL_RCC_GetLPTIMClockFreq (LL_RCC_LPTIM1_CLKSOURCE) ;
	}

	if (deviceAddress == LPTIM2)
	{
		return LL_RCC_GetLPTIMClockFreq (LL_RCC_LPTIM2_CLKSOURCE) ;
	}

	if (deviceAddress == LPTIM3  ||  deviceAddress == LPTIM4  ||  deviceAddress == LPTIM5)
	{
		return LL_RCC_GetLPTIMClockFreq (LL_RCC_LPTIM345_CLKSOURCE) ;
	}

	switch ((uintptr_t)  deviceAddress & ~OFFSET_MASK)
	{
		case D2_APB1PERIPH_BASE:
			timClock = rccGetPCLK1ClockFreq () ;		// timClk is PCLK1
			D2PPRE = (RCC->D2CFGR & RCC_D2CFGR_D2PPRE1_Msk) >> RCC_D2CFGR_D2PPRE1_Pos ;
			TIMPRE = RCC->CFGR & RCC_CFGR_TIMPRE ;
			if (D2PPRE >= 4u)		// 4 is DIV2
			{
				timClock *= 2 ;
				if (TIMPRE != 0u  &&  D2PPRE > 4u)
				{
					timClock *= 2u ;
				}
			}
			break ;

		case D2_APB2PERIPH_BASE:
			timClock = rccGetPCLK2ClockFreq () ;		// timClk is PCLK2
			D2PPRE = (RCC->D2CFGR & RCC_D2CFGR_D2PPRE2_Msk) >> RCC_D2CFGR_D2PPRE2_Pos ;
			TIMPRE = RCC->CFGR & RCC_CFGR_TIMPRE ;
			if (D2PPRE >= 4u)		// 4 is DIV2
			{
				timClock *= 2u ;
				if (TIMPRE != 0u  &&  D2PPRE > 4u)
				{
					timClock *= 2u ;
				}
			}
			break ;

		default:
			AA_ASSERT (0) ;
			timClock = 0u ;
			break ;
	}
	return timClock ;
}

//----------------------------------------------------------------------
void	rccResetPeriph (void * deviceAddress, uint32_t enableBitMask)
{
	switch ((uintptr_t) deviceAddress & ~OFFSET_MASK)
	{
		// APB buses
		case D1_APB1PERIPH_BASE:
			RCC->APB3RSTR |=  enableBitMask ;			// RCC_APB3RSTR_LTDCRST
			RCC->APB3RSTR &= ~enableBitMask ;			// RCC_APB3RSTR_LTDCRST
			break ;

		case D2_APB1PERIPH_BASE:
			if (((uint32_t) deviceAddress & OFFSET_MASK) >= CRS_BASE)
			{
				RCC->APB1HRSTR |=  enableBitMask ;		// RCC_APB1HRSTR_CRSRST
				RCC->APB1HRSTR &= ~enableBitMask ;
			}
			else
			{
				RCC->APB1LRSTR |=  enableBitMask ;		// RCC_APB1LRSTR_TIM2RST
				RCC->APB1LRSTR &= ~enableBitMask ;
			}
			break ;

		case D3_APB1PERIPH_BASE:
			RCC->APB4RSTR |=  enableBitMask ;			// RCC_APB4RSTR_SAI4RST
			RCC->APB4RSTR &= ~enableBitMask ;
			break ;

		// AHB buses
		case D1_AHB1PERIPH_BASE:
			RCC->AHB3RSTR |=  enableBitMask ;			// RCC_AHB3RSTR_MDMARST
			RCC->AHB3RSTR &= ~enableBitMask ;
			break ;

		case D2_AHB1PERIPH_BASE:
			RCC->AHB1RSTR |=  enableBitMask ;			// RCC_AHB1RSTR_DMA1RST
			RCC->AHB1RSTR &= ~enableBitMask ;
			break ;

		case D2_AHB2PERIPH_BASE:
			RCC->AHB2RSTR |=  enableBitMask ;			// RCC_AHB2RSTR_DCMIRST
			RCC->AHB2RSTR &= ~enableBitMask ;
			break ;

		case D3_AHB1PERIPH_BASE:
			RCC->AHB4RSTR |=  enableBitMask ;			// RCC_AHB4RSTR_GPIOARST
			RCC->AHB4RSTR &= ~enableBitMask ;
			break ;

		default:
			AA_ASSERT (0) ;
			break ;
	}
}

//----------------------------------------------------------------------
