/*
----------------------------------------------------------------------

	AdAstra - Real Time Kernel

	Alain Chebrou

	spiF429ZI.c		STM32F429 LQFP144 SPI descriptors for spibasic.c

	When		Who	What
	12/04/18	ac	Creation

----------------------------------------------------------------------
*/
/*
----------------------------------------------------------------------
	Copyright (c) 2013-2023 Alain Chebrou - AdAstra-Soft . com
	All rights reserved.

	This file is part of the AdAstra Real Time Kernel distribution.

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions are met:

	- Redistributions of source code must retain the above copyright
	  notice, this list of conditions and the following disclaimer.

	- Redistributions in binary form must reproduce the above copyright
	  notice, this list of conditions and the following disclaimer in the
	  documentation and/or other materials provided with the distribution.

	- Neither the name of the copyright holder nor the
	  names of its contributors may be used to endorse or promote products
	  derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
	EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
	DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
	DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
	(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
	ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
----------------------------------------------------------------------
*/

#if ! defined spiH7xx_H_
#define spiH7xx_H_
//------------------------------------------------------------------------------

/*
STM32H742/H743

		MOSI		MISO		SCK			NSS			APB

SPI_1	PA7			PA6			PA5			PA4			2
		PB5			PB4			PB3=SWO		PA15
		PD7			PG9			PG11		PG10

SPI_2	PC1			PC2			PA9			PA11
		PC3						PA12		PB1(7)
								PB10		PB9
		PB15		PB14		PB13		PB12
								PD3
		PI3			PI2			PI1			PI0

SPI_3	PB2(7)		PB4(6)		PB3(6)		PA4(6)		1
		PC12(6)		PC11(6)		PC10(6)		PA15(6)
		PD6
		PB5(7)

SPI_4	PE6			PE5			PE2			PE4			2
		PE14		PE13		PE12		PE11

SPI_5	PF9			PF8			PF7			PF6			2
		PF11		PH7			PH6			PH5
		PJ10		PJ11		PK0			PK1

SPI_6	PA7(8)		PA6(8)		PA5(8)		PA4(8)		4
		PB5(8)		PB4(8)		PB3(8)		PA15(7)
		PG14		PG12		PG13		PG8

All pins in AF5 except ()
PHx PIx PJx PKx only if pin count > 169
PB3 is also SWO
*/

/*
	Mode	CPOL	CPHA
	0		0		0
	1		0		1
	2		1		0
	3		1		1

	CPOL	Clock polarity controls the steady state value of the clock when no data is being transferred
			0: Low level when idle state

	CPHA	Controls which edge of the clock locks the data.
			0:first edge, 1:second edge
*/

// If half duplex communication is used, the MISO pin descriptor may be disabled (port field set to 0)

// NSS can be AA_GPIO_MODE_OUTPUT (with AF 0), if the NSS is managed by software (flag SPI_NSS_USER)
// NSS should be AA_GPIO_MODE_ALTERNATE if the auto NNS is selected (flag SPI_NSS_AUTO)

// SPI6 is on D3-APB4  => can't use DMA1/DMA2, use BDMA to access SRAM4 only
// => SPI6 is not supported by this DMA API

static	const spiDesc_t	spiDescArray [] =
{
	#if (WITH_SPI_1 == 1)
	{
		10,								// id
		SPI1,							// Pointer to SPI registers
		{	'B',	5,	AA_GPIO_AF_5,	AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH   | AA_GPIO_PUSH_PULL },		// MOSI
		{	'A',	6,	AA_GPIO_AF_5,	AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH   | AA_GPIO_PUSH_PULL },		// MISO
		{	'A',	5,	AA_GPIO_AF_5,	AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH   | AA_GPIO_PUSH_PULL },		// SCK
//		{	'A',	4,	AA_GPIO_AF_5,	AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_MEDIUM | AA_GPIO_PUSH_PULL },		// NSS select AUTO/HARD
		{	'A',	4,	AA_GPIO_AF_0,	AA_GPIO_MODE_OUTPUT    | AA_GPIO_SPEED_MEDIUM | AA_GPIO_PUSH_PULL },		// NSS select USER/SOFT
		{	// DMA Tx
			DMA_DMA1,							//	DMA number 1 or 2, 0 => no DMA
			3,									//	DMA stream number 0..7
			(BSP_MAX_INT_PRIO+2),				//	DMA Interrupt priority
			0,									//	Mux input (not used)
			LL_DMA_DIRECTION_MEMORY_TO_PERIPH |	//	DMA config: dir, mode, priority...
			LL_DMA_PRIORITY_HIGH
		},
		{	// DMA Rx
			DMA_DMA1,							//	DMA number 1 or 2, 0 => no DMA
			2,									//	DMA stream number 0..7
			(BSP_MAX_INT_PRIO+2),				//	DMA interrupt priority
			0,									//	Mux input (not used)
			LL_DMA_DIRECTION_PERIPH_TO_MEMORY |	//	DMA config: dir, mode, priority...
			LL_DMA_PRIORITY_HIGH
		},
		(BSP_MAX_INT_PRIO+2),					//	SPI interrupt priority
	},
	#endif	// WITH_SPI_1

	#if (WITH_SPI_2 == 1)
	{
		20,
		SPI2,
		{	'B',	15,	AA_GPIO_AF_5,	AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH | AA_GPIO_PUSH_PULL },		// MOSI
		{	'B',	14,	AA_GPIO_AF_5,	AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH | AA_GPIO_PUSH_PULL },		// MISO
		{	'B',	10,	AA_GPIO_AF_5,	AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH | AA_GPIO_PUSH_PULL },		// SCK
		{	'B',	9,	AA_GPIO_AF_5,	AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH | AA_GPIO_PUSH_PULL },		// NSS
		{	// DMA Tx
			DMA_NONE,							//	DMA number 1 or 2, 0 => no DMA
			0,									//	DMA stream number 0..7
			(BSP_MAX_INT_PRIO+2),				//	DMA Interrupt priority
			0,									//	Mux input (not used)
			LL_DMA_DIRECTION_MEMORY_TO_PERIPH |	//	DMA config: dir, mode, priority...
			LL_DMA_PRIORITY_HIGH
		},
		{	// DMA Rx
			DMA_NONE,							//	DMA number 1 or 2, 0 => no DMA
			0,									//	DMA stream number 0..7
			(BSP_MAX_INT_PRIO+2),				//	DMA interrupt priority
			0,									//	Mux input (not used)
			LL_DMA_DIRECTION_PERIPH_TO_MEMORY |	//	DMA config: dir, mode, priority...
			LL_DMA_PRIORITY_HIGH
		},
		(BSP_MAX_INT_PRIO+2),					//	SPI interrupt priority
	},
	#endif	// WITH_SPI_2

	#if (WITH_SPI_3 == 1)
	{
		30,
		SPI3,
		{	'D',	6,	AA_GPIO_AF_5,	AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH | AA_GPIO_PUSH_PULL },		// MOSI
		{	'C',	11,	AA_GPIO_AF_6,	AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH | AA_GPIO_PUSH_PULL },		// MISO
		{	'C',	10,	AA_GPIO_AF_6,	AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH | AA_GPIO_PUSH_PULL },		// SCK
		{	'A',	4,	AA_GPIO_AF_6,	AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH | AA_GPIO_PUSH_PULL },		// NSS
		{	// DMA Tx
			DMA_NONE,							//	DMA number 1 or 2, 0 => no DMA
			0,									//	DMA stream number 0..7
			(BSP_MAX_INT_PRIO+2),				//	DMA Interrupt priority
			0,									//	Mux input (not used)
			LL_DMA_DIRECTION_MEMORY_TO_PERIPH |	//	DMA config: dir, mode, priority...
			LL_DMA_PRIORITY_HIGH
		},
		{	// DMA Rx
			DMA_NONE,							//	DMA number 1 or 2, 0 => no DMA
			0,									//	DMA stream number 0..7
			(BSP_MAX_INT_PRIO+2),				//	DMA interrupt priority
			0,									//	Mux input (not used)
			LL_DMA_DIRECTION_PERIPH_TO_MEMORY |	//	DMA config: dir, mode, priority...
			LL_DMA_PRIORITY_HIGH
		},
		(BSP_MAX_INT_PRIO+2),					//	SPI interrupt priority
	},
	#endif	// WITH_SPI_3

	#if (WITH_SPI_4 == 1)
	{
		40,
		SPI4,
		{	'E',	6,	AA_GPIO_AF_5,	AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH | AA_GPIO_PUSH_PULL },		// MOSI
		{	'E',	5,	AA_GPIO_AF_5,	AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH | AA_GPIO_PUSH_PULL },		// MISO
		{	'E',	2,	AA_GPIO_AF_5,	AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH | AA_GPIO_PUSH_PULL },		// SCK
		{	'E',	4,	AA_GPIO_AF_5,	AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH | AA_GPIO_PUSH_PULL },		// NSS
		{	// DMA Tx
			DMA_NONE,							//	DMA number 1 or 2, 0 => no DMA
			0,									//	DMA stream number 0..7
			(BSP_MAX_INT_PRIO+2),				//	DMA Interrupt priority
			0,									//	Mux input (not used)
			LL_DMA_DIRECTION_MEMORY_TO_PERIPH |	//	DMA config: dir, mode, priority...
			LL_DMA_PRIORITY_HIGH
		},
		{	// DMA Rx
			DMA_NONE,							//	DMA number 1 or 2, 0 => no DMA
			0,									//	DMA stream number 0..7
			(BSP_MAX_INT_PRIO+2),				//	DMA interrupt priority
			0,									//	Mux input (not used)
			LL_DMA_DIRECTION_PERIPH_TO_MEMORY |	//	DMA config: dir, mode, priority...
			LL_DMA_PRIORITY_HIGH
		},
		(BSP_MAX_INT_PRIO+2),					//	SPI interrupt priority
	},
	#endif	// WITH_SPI_4

	#if (WITH_SPI_5 == 1)
	{
		50,
		SPI5,
		{	'F',	9,	AA_GPIO_AF_5,	AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH | AA_GPIO_PUSH_PULL },		// MOSI
		{	'F',	8,	AA_GPIO_AF_5,	AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH | AA_GPIO_PUSH_PULL },		// MISO
		{	'F',	7,	AA_GPIO_AF_5,	AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH | AA_GPIO_PUSH_PULL },		// SCK
		{	'F',	6,	AA_GPIO_AF_5,	AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH | AA_GPIO_PUSH_PULL },		// NSS
		{	// DMA Tx
			DMA_NONE,							//	DMA number 1 or 2, 0 => no DMA
			0,									//	DMA stream number 0..7
			(BSP_MAX_INT_PRIO+2),				//	DMA Interrupt priority
			0,									//	Mux input (not used)
			LL_DMA_DIRECTION_MEMORY_TO_PERIPH |	//	DMA config: dir, mode, priority...
			LL_DMA_PRIORITY_HIGH
		},
		{	// DMA Rx
			DMA_NONE,							//	DMA number 1 or 2, 0 => no DMA
			0,									//	DMA stream number 0..7
			(BSP_MAX_INT_PRIO+2),				//	DMA interrupt priority
			0,									//	Mux input (not used)
			LL_DMA_DIRECTION_PERIPH_TO_MEMORY |	//	DMA config: dir, mode, priority...
			LL_DMA_PRIORITY_HIGH
		},
		(BSP_MAX_INT_PRIO+2),					//	SPI interrupt priority
	},
	#endif	// WITH_SPI_5

	#if (WITH_SPI_6 == 1)
	{
		60,
		SPI6,
		{	'G',	14,	AA_GPIO_AF_5,	AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH | AA_GPIO_PUSH_PULL },		// MOSI
		{	'G',	12,	AA_GPIO_AF_5,	AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH | AA_GPIO_PUSH_PULL },		// MISO
		{	'G',	13,	AA_GPIO_AF_5,	AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH | AA_GPIO_PUSH_PULL },		// SCK
		{	'G',	8,	AA_GPIO_AF_5,	AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH | AA_GPIO_PUSH_PULL },		// NSS
		{	// DMA Tx
			DMA_NONE,							//	DMA number 1 or 2, 0 => unused descriptor
			0,									//	DMA stream number 0..7
			(BSP_MAX_INT_PRIO+2),				//	DMA Interrupt priority
			0,									//	Mux input (not used)
			LL_DMA_DIRECTION_MEMORY_TO_PERIPH |	//	DMA config: dir, mode, priority...
			LL_DMA_PRIORITY_HIGH
		},
		{	// DMA Rx
			DMA_NONE,							//	DMA number 1 or 2, 0 => unused descriptor
			0,									//	DMA stream number 0..7
			(BSP_MAX_INT_PRIO+2),				//	DMA interrupt priority
			0,									//	Mux input (not used)
			LL_DMA_DIRECTION_PERIPH_TO_MEMORY |	//	DMA config: dir, mode, priority...
			LL_DMA_PRIORITY_HIGH
		},
		(BSP_MAX_INT_PRIO+2),					//	SPI interrupt priority
	},
	#endif	// WITH_SPI_6
} ;

//------------------------------------------------------------------------------
#endif	// spiH7xx_H_
