/*
----------------------------------------------------------------------

	AdAstra - Real Time Kernel

	Alain Chebrou

	spibasic.c	Master only, polling or DMA exchanges, software or hardware SS management, no CRC
				polling exchanges, for SPI1 to SPI6
				DMA exchanges, use DMA Direct Mode (doesn't use DMA FIFO), for SPI1 to SPI5 only.
				For STM32H7xx

	When		Who	What
	27/02/20	ac	Creation

----------------------------------------------------------------------
*/
/*
----------------------------------------------------------------------
	Copyright (c) 2013-2023 Alain Chebrou - AdAstra-Soft . com
	All rights reserved.

	This file is part of the AdAstra Real Time Kernel distribution.

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions are met:

	- Redistributions of source code must retain the above copyright
	  notice, this list of conditions and the following disclaimer.

	- Redistributions in binary form must reproduce the above copyright
	  notice, this list of conditions and the following disclaimer in the
	  documentation and/or other materials provided with the distribution.

	- Neither the name of the copyright holder nor the
	  names of its contributors may be used to endorse or promote products
	  derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
	DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
	DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
	(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
	ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
----------------------------------------------------------------------
*/
#include	"aa.h"
#include	"aakernel.h"
#include	"gpiobasic.h"
#include	"spibasic.h"

#include	"stm32h7xx_ll_spi.h"
#include	"stm32h7xx_ll_rcc.h"
#include	"stm32h7xx_ll_bus.h"

// Used for timing measure with an oscilloscope.
#define dbgPin1(v)		bspOutput (BSP_OUTPUT1, v) ;	// Debug only
//#define dbgPin1(v)											// No debug pulse

//----------------------------------------------------------------------

#if (WITH_SPI_X == 1)
	// Include the proper SPI descriptors
	#include "spiH7xx.h"

	// Count of SPI devices in spiDescArray
	STATIC	const uint32_t	spiArrayCount = sizeof (spiDescArray) / sizeof (spiDesc_t) ;
#endif

#define	SPI_EV_EOT			0x10u		// SPI callback event: End Of Transfer

#define	SPI_DEF_TIMEOUT		10u			// Timeout in kernel tick

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//	A descriptor for the capabilities of each SPI

typedef struct spiCap_s
{
	volatile	uint32_t	* rccENR ;		// @ of register to enable SPI clock
	volatile	uint32_t	* rccRSTR ;		// @ of register to reset the SPI peripheral
				uint32_t	periphPos ;		// To use with RCC registers
				uint16_t	dmaTxReqNum ;
				uint16_t	dmaRxReqNum ;
				IRQn_Type	irqVector ;
				uint16_t	fifoDepth ;		// In bytes
				uint16_t	maxBits ;		// 32 or 16
				uint32_t	clkSrc ;		// LL_RCC_SPIxxx_CLKSOURCE

} spiCap_t ;

#if (defined SPI1)
	const spiCap_t	capSpi1 =
	{
			& RCC->APB2ENR,
			& RCC->APB2RSTR,
			LL_APB2_GRP1_PERIPH_SPI1,
			LL_DMAMUX1_REQ_SPI1_TX,
			LL_DMAMUX1_REQ_SPI1_RX,
			SPI1_IRQn,
			16u,
			32u,
			LL_RCC_SPI123_CLKSOURCE
	} ;
#endif	// SPI1

#if (defined SPI2)
	const spiCap_t	capSpi2 =
	{
			& RCC->APB1LENR,
			& RCC->APB1LRSTR,
			LL_APB1_GRP1_PERIPH_SPI2,
			LL_DMAMUX1_REQ_SPI2_TX,
			LL_DMAMUX1_REQ_SPI2_RX,
			SPI2_IRQn,
			16u,
			32u,
			LL_RCC_SPI123_CLKSOURCE
	} ;
#endif	// SPI2

#if (defined SPI3)
	const spiCap_t	capSpi3 =
	{
			& RCC->APB1LENR,
			& RCC->APB1LRSTR,
			LL_APB1_GRP1_PERIPH_SPI3,
			LL_DMAMUX1_REQ_SPI3_TX,
			LL_DMAMUX1_REQ_SPI3_RX,
			SPI3_IRQn,
			16u,
			32u,
			LL_RCC_SPI123_CLKSOURCE
	} ;
#endif	// SPI3

#if (defined SPI4)
	const spiCap_t	capSpi4 =
	{
			& RCC->APB2ENR,
			& RCC->APB2RSTR,
			LL_APB2_GRP1_PERIPH_SPI4,
			LL_DMAMUX1_REQ_SPI4_TX,
			LL_DMAMUX1_REQ_SPI4_RX,
			SPI4_IRQn,
			8u,
			16u,
			LL_RCC_SPI45_CLKSOURCE
	} ;
#endif	// SPI4

#if (defined SPI5)
	const spiCap_t	capSpi5 =
	{
			& RCC->APB2ENR,
			& RCC->APB2RSTR,
			LL_APB2_GRP1_PERIPH_SPI5,
			LL_DMAMUX1_REQ_SPI5_TX,
			LL_DMAMUX1_REQ_SPI5_RX,
			SPI5_IRQn,
			8u,
			16u,
			LL_RCC_SPI45_CLKSOURCE
	};
#endif	// SPI5

#if (defined SPI6)
	const spiCap_t	capSpi6 =
	{
			& RCC->APB4ENR,
			& RCC->APB4RSTR,
			LL_APB4_GRP1_PERIPH_SPI6,
			LL_DMAMUX2_REQ_SPI6_TX,
			LL_DMAMUX2_REQ_SPI6_RX,
			SPI6_IRQn,
			8u,
			16u,
			LL_RCC_SPI6_CLKSOURCE
	} ;
#endif	// SPI6

//----------------------------------------------------------------------
// SPI handler structure. The SPI handle is a pointer to this structure.

typedef struct
{
	const spiCap_t	* pSpiCap ;		// Capabilities of the SPI interface
	const spiDesc_t	* pSpiDesc ;	// The descriptor from spiMasterConfigure()
	SPI_TypeDef		* pSpi ;		// SPI1, SPI2...
	spiStatus_t		result ;		// Result of the operation (DMA only)
	uint32_t		timeout ;		// Transfer timeout (system tick)
	aaTaskId_t		taskId ;		// The task to signal at end of operation (DMA only)

} spiIo_t ;

//	One handler structure for each physical SPI peripheral

#if (WITH_SPI_1 == 1)
	spiIo_t		spiIo1 ;
#endif	// WITH_SPI_1

#if (WITH_SPI_2 == 1)
	spiIo_t		spiIo2 ;
#endif	// WITH_SPI_2

#if (WITH_SPI_3 == 1)
	spiIo_t		spiIo3 ;
#endif	// WITH_SPI_3

#if (WITH_SPI_4 == 1)
	spiIo_t		spiIo4 ;
#endif	// WITH_SPI_4

#if (WITH_SPI_5 == 1)
	spiIo_t		spiIo5 ;
#endif	// WITH_SPI_5

#if (WITH_SPI_6 == 1)
	spiIo_t		spiIo6 ;
#endif	// WITH_SPI_6


//----------------------------------------------------------------------
//	Returns the address of the capabilities structure associated to the SPI

static	const spiCap_t * spiGetCap (SPI_TypeDef * pSpi)
{
	switch ((uintptr_t) pSpi)
	{
		#if (defined (SPI1))
			case SPI1_BASE: return & capSpi1 ;
		#endif

		#if (defined (SPI2))
			case SPI2_BASE: return & capSpi2 ;
		#endif

		#if (defined (SPI3))
			case SPI3_BASE: return & capSpi3 ;
		#endif

		#if (defined SPI4)
			case SPI4_BASE: return & capSpi4 ;
		#endif

		#if (defined SPI5)
			case SPI5_BASE: return & capSpi5 ;
		#endif

		#if (defined SPI6)
			case SPI6_BASE: return & capSpi6 ;
		#endif

		default:
			AA_ASSERT (0) ;
			break ;
	}
	return NULL ;
}

//----------------------------------------------------------------------
//	Get the I/O descriptor address for an SPI peripheral

static	spiIo_t *	spiGetIoDesc (SPI_TypeDef * pSpi) ;

static	spiIo_t *	spiGetIoDesc (SPI_TypeDef * pSpi)
{
	#if (WITH_SPI_X == 0)
	{
		(void) pSpi ;	// To avoid compiler warning when none SPI is supported
	}
	#endif

	#if (WITH_SPI_1 == 1)
	{
		if (pSpi == SPI1)
		{
			return & spiIo1 ;
		}
	}
	#endif	// WITH_SPI_1

	#if (WITH_SPI_2 == 1)
	{
		if (pSpi == SPI2)
		{
			return & spiIo2 ;
		}
	}
	#endif	// WITH_SPI_2

	#if (WITH_SPI_3 == 1)
	{
		if (pSpi == SPI3)
		{
			return & spiIo3 ;
		}
	}
	#endif	// WITH_SPI_3

	#if (WITH_SPI_4 == 1)
	{
		if (pSpi == SPI4)
		{
			return & spiIo4 ;
		}
	}
	#endif	// WITH_SPI_4

	#if (WITH_SPI_5 == 1)
	{
		if (pSpi == SPI5)
		{
			return & spiIo5 ;
		}
	}
	#endif	// WITH_SPI_5

	#if (WITH_SPI_6 == 1)
	{
		if (pSpi == SPI6)
		{
			return & spiIo6 ;
		}
	}
	#endif	// WITH_SPI_6

	AA_ASSERT (0) ;
	return NULL ;
}

//----------------------------------------------------------------------
//	Return SPI descriptor address from an index
//	The return value is an opaque handle for the application

spiStatus_t		spiGetDesc	(uint32_t id, const spiDesc_t ** ppSpiDesc)
{
	#if (WITH_SPI_X == 1)
		uint32_t	ii ;

		for (ii = 0u ; ii < spiArrayCount ; ii++)
		{
			if (id == spiDescArray [ii].id)
			{
				* ppSpiDesc = & spiDescArray [ii] ;
				return SPI_ENONE ;
			}
		}
	#else
		(void) id ;
		(void) ppSpiDesc ;
	#endif

	return SPI_EFAIL ;
}

//----------------------------------------------------------------------
//	Set / Reset Slave Select pin on master side
//	Use NSS pin of SPI descriptor
//	SPI slave select is active low: so bSelect 1 => NSS 0

void	spiSlaveSelect (spiHandle_t hdl, uint32_t bSelect)
{
	spiIo_t		* pSpiIo = (spiIo_t *) hdl ;

	if ((pSpiIo->pSpiDesc->pinNss.port != 0u)  &&  (pSpiIo->pSpiDesc->pinNss.port != '0'))
	{
		gpioPinSet (& pSpiIo->pSpiDesc->pinNss, (bSelect & 1u) ^ 1u) ;
	}
}

//----------------------------------------------------------------------
//	Find baud rate control value from bits/s rate
//	The return value is the SPI_CR1 BR[2:0] field value

uint32_t	spiSpeedToBr (const spiDesc_t * pSpiDesc, uint32_t speed)
{
	const spiCap_t	* pSpiCap ;
	uint32_t		br ;
	uint32_t		clock ;

	pSpiCap = spiGetCap (pSpiDesc->pSpi) ;
	if (pSpiCap == NULL)
	{
		return 0u ;
	}

	// Max allowed clock is Mclk / 2u
	clock = LL_RCC_GetSPIClockFreq (pSpiCap->clkSrc) / 2u ;

	for (br = 0u ; br < 7u ; br++)
	{
		if (speed >= clock)
		{
			break ;
		}
		clock /= 2u ;
	}

	return br ;
}

//----------------------------------------------------------------------
//	Find bit/s rate from BR control register value

uint32_t	spiBrToSpeed (const spiDesc_t * pSpiDesc, uint32_t br)
{
	const spiCap_t	* pSpiCap ;
	uint32_t		clock ;

	pSpiCap = spiGetCap (pSpiDesc->pSpi) ;
	if (pSpiCap == NULL)
	{
		return 0u ;
	}

	clock = LL_RCC_GetSPIClockFreq (pSpiCap->clkSrc) ;
	return clock >> ((br & 0x7u) + 1u) ;
}

//----------------------------------------------------------------------
// Set some options for master SPI
// To use after spiMasterConfigure()

spiStatus_t		spiMasterSetOption		(spiHandle_t hdl, uint32_t option, uintptr_t value)
{
	spiIo_t			* pSpiIo = (spiIo_t *) hdl ;
	SPI_TypeDef		* pSpi = pSpiIo->pSpi ;

	switch (option)
	{
		case SPI_OPT_SWAPIO:
			if (value == 0u)
			{
				pSpi->CFG2 &= ~SPI_CFG2_IOSWP ;
			}
			else
			{
				pSpi->CFG2 |= SPI_CFG2_IOSWP ;
			}
			break ;

		case SPI_OPT_GAP:
			pSpi->CFG2 = (pSpi->CFG2 & ~SPI_CFG2_MIDI_Msk) | (((value & 0x0Fu) << SPI_CFG2_MIDI_Pos) & SPI_CFG2_MIDI_Msk) ;
			break ;

		case SPI_OPT_TIMEOUT:
			pSpiIo->timeout = (value > 0xFFFFu) ? 0xFFFFu : value ;
			break ;

		default:
			return SPI_EARG ;
	}
	return SPI_ENONE ;
}

//----------------------------------------------------------------------
// Callback to manage master SPI End Of Transfer (EOT interrupt), and DMA errors

static	void spiCallback (uint32_t event, uint32_t dir, uintptr_t arg) ;

static	void spiCallback (uint32_t event, uint32_t dir, uintptr_t arg)
{
	spiIo_t		* pSpiIo = (spiIo_t *) arg ;

dbgPin1 (1) ;
	if (event == DMA_EV_ERROR)
	{
		if (pSpiIo->taskId != AA_INVALID_TASK)
		{
			pSpiIo->result = SPI_EDMA ;
			aaSignalSend (pSpiIo->taskId, AA_SIGNAL_IO) ;
			pSpiIo->taskId = AA_INVALID_TASK ;	// Prevent multiple wake up
		}
	}
	else if (event == SPI_EV_EOT)
	{
		if (pSpiIo->taskId != AA_INVALID_TASK)
		{
			pSpiIo->result = SPI_ENONE ;
dbgPin1 (0u) ;
			aaSignalSend (pSpiIo->taskId, AA_SIGNAL_IO) ;
			pSpiIo->taskId = AA_INVALID_TASK ;	// Prevent multiple wake up
dbgPin1 (1u) ;
		}
	}
	else
	{
	}

	(void) dir ;
}

//----------------------------------------------------------------------
//	Configure one DMA stream for SPI

static	void	spiDmaConfigure (const dmaDesc_t * pDmaDesc, spiIo_t * pSpiIo) ;

static	void	spiDmaConfigure (const dmaDesc_t * pDmaDesc, spiIo_t * pSpiIo)
{
	dma_t			* pDma ;
	dmaStream_t		* pStream ;
	SPI_TypeDef 	* pSpi = pSpiIo->pSpi ;
	uint32_t		streamNum ;
	uint32_t		reg ;

	if (pDmaDesc->dmaNum == 0u)
	{
		return ;	// No DMA to configure
	}

	pDma		= DMA_GETDMA (pDmaDesc->dmaNum) ;
	streamNum	= DMA_GETSTREAM (pDmaDesc) ;
	pStream		= dmaStreamAddressRaw (pDma, streamNum) ;

	// Enable DMA clock
	dmaEnableClock (pDma) ;

	// Reset stream
	dmaStreamResetRaw (pDma, streamNum) ;

	// Configure the DMA/Stream functional parameters
	reg = pSpi->CFG1 & SPI_CFG1_DSIZE ;
	if (reg < 8u)
	{
		reg = LL_DMA_PDATAALIGN_BYTE | LL_DMA_MDATAALIGN_BYTE ;
	}
	else if (reg < 16u)
	{
		reg = LL_DMA_PDATAALIGN_HALFWORD | LL_DMA_MDATAALIGN_HALFWORD ;
	}
	else
	{
		reg = LL_DMA_PDATAALIGN_WORD | LL_DMA_MDATAALIGN_WORD ;
	}
	reg |=	LL_DMA_MODE_NORMAL        |
			LL_DMA_PERIPH_NOINCREMENT |
			LL_DMA_MEMORY_INCREMENT ;

	pStream->CR &= ~(DMA_SxCR_DIR | DMA_SxCR_CIRC | DMA_SxCR_PINC | DMA_SxCR_MINC | DMA_SxCR_PSIZE | DMA_SxCR_MSIZE | DMA_SxCR_PL | DMA_SxCR_PFCTRL) ;
	pStream->CR |= pDmaDesc->config | reg ;

	// Select the DMA transfer request
	// Set DMA transfer addresses of the peripheral
	if ((pDmaDesc->config & LL_DMA_DIRECTION_PERIPH_TO_MEMORY) != 0u)
	{
		// Periph to Memory (RX)
		((dmaMux_t *) DMAMUX1)->CCR [streamNum + ((pDma == (dma_t *) DMA1) ? 0u : 8u)] = pSpiIo->pSpiCap->dmaRxReqNum ;

		pStream->PAR = (uint32_t) & pSpi->RXDR ;
	}
	else
	{
		// Memory to Periph (TX)
		((dmaMux_t *) DMAMUX1)->CCR [streamNum + ((pDma == (dma_t *) DMA1) ? 0u : 8u)] = pSpiIo->pSpiCap->dmaTxReqNum ;

		pStream->PAR = (uint32_t) & pSpi->TXDR ;
	}

	// Configure NVIC for DMA error interrupt
	dmaSetIrq (pDmaDesc) ;
	dmaSetCallback (pDmaDesc, spiCallback, (uintptr_t) pSpiIo) ;

	// Enable DMA error interrupt
	pStream->CR |= DMA_SxCR_TEIE ;
}

//----------------------------------------------------------------------
//	Initialize SPI device: SPI master mode, polling or DMA, no CRC
//	Speed is a wish, because this SPI only allows a very limited number of speeds
//	Default exchange timeout is SPI_DEF_TIMEOUT ticks
//	No bit swap, no gap between words

spiStatus_t		spiMasterConfigure	(const spiDesc_t * pSpiDesc, uint32_t speed, uint32_t flags, spiHandle_t * pHdl)
{
	LL_SPI_InitTypeDef	SPI_InitStruct ;
	spiIo_t				* pSpiIo ;
	const spiCap_t		* pSpiCap ;
	SPI_TypeDef			* pSpi = pSpiDesc->pSpi ;

	pSpiIo = spiGetIoDesc (pSpi) ;
	if (pSpiIo == NULL)
	{
		return SPI_EARG ;
	}

	pSpiCap = spiGetCap (pSpi) ;
	if (pSpiCap == NULL)
	{
		return SPI_EARG ;
	}

	if ((flags & SPI_BITS_MASK) > pSpiCap->maxBits)
	{
		return SPI_EARG ;	// This SPI doesn't support this data size
	}
	pSpiIo->pSpiCap		= pSpiCap ;
	pSpiIo->pSpi		= pSpi ;
	pSpiIo->pSpiDesc	= pSpiDesc ;
	pSpiIo->taskId		= AA_INVALID_TASK ;
	pSpiIo->timeout		= SPI_DEF_TIMEOUT ;

	// Enable the peripheral clock for SPI
	* pSpiCap->rccENR |= pSpiCap->periphPos ;

	// Reset peripheral => all configuration registers to reset values
	* pSpiCap->rccRSTR |=  pSpiCap->periphPos ;
	* pSpiCap->rccRSTR &= ~pSpiCap->periphPos ;

	// Configure pins
	gpioConfigurePin (& pSpiDesc->pinMosi) ;
	gpioConfigurePin (& pSpiDesc->pinMiso) ;

	// Set SCK value before pin init: avoid glitch if there is external pull up/down
	if ((flags & SPI_MODE_2) == 0)
	{
		gpioConfigureAndSetPin (& pSpiDesc->pinSck, 0) ;	// SPI_MODE_0 or SPI_MODE_1 : Low level when idle state
	}
	else
	{
		gpioConfigureAndSetPin (& pSpiDesc->pinSck, 1) ;	// SPI_MODE_2 or SPI_MODE_3 : High level when idle state
	}

	// Set NSS value before pin init: avoid glitch if there is external pull up
	gpioConfigureAndSetPin (& pSpiDesc->pinNss, 1) ;

	SPI_InitStruct.Mode				= LL_SPI_MODE_MASTER ;
	SPI_InitStruct.NSS				= ((flags & SPI_NSS_MASK) == SPI_NSS_USER) ? LL_SPI_NSS_SOFT : LL_SPI_NSS_HARD_OUTPUT ;
	SPI_InitStruct.BaudRate			= spiSpeedToBr (pSpiDesc, speed) << SPI_CFG1_MBR_Pos ;
	SPI_InitStruct.CRCCalculation	= LL_SPI_CRCCALCULATION_DISABLE ;
	SPI_InitStruct.CRCPoly			= 10u ;

	// Set SPI polarity / phase
	switch (flags & SPI_MODE_MASK)
	{
		default:
		case SPI_MODE_0:		// Default
			SPI_InitStruct.ClockPolarity	= LL_SPI_POLARITY_LOW ;
			SPI_InitStruct.ClockPhase		= LL_SPI_PHASE_1EDGE ;
			break ;

		case SPI_MODE_1:
			SPI_InitStruct.ClockPolarity	= LL_SPI_POLARITY_LOW ;
			SPI_InitStruct.ClockPhase		= LL_SPI_PHASE_2EDGE ;
			break ;

		case SPI_MODE_2:
			SPI_InitStruct.ClockPolarity	= LL_SPI_POLARITY_HIGH ;
			SPI_InitStruct.ClockPhase		= LL_SPI_PHASE_1EDGE ;
			break ;

		case SPI_MODE_3:
			SPI_InitStruct.ClockPolarity	= LL_SPI_POLARITY_HIGH ;
			SPI_InitStruct.ClockPhase		= LL_SPI_PHASE_2EDGE ;
			break ;
	}

	// Set transfer mode
	switch (flags & SPI_DIR_MASK)
	{
		default:
		case SPI_DIR_FULL_DUPLEX:		// Default
			SPI_InitStruct.TransferDirection	= LL_SPI_FULL_DUPLEX ;
			break ;
/*
		case SPI_DIR_RX_ONLY:
			SPI_InitStruct.TransferDirection	= LL_SPI_SIMPLEX_RX ;
			break ;

		case SPI_DIR_HALF_RX:
			SPI_InitStruct.TransferDirection	= LL_SPI_HALF_DUPLEX_RX ;
			break ;
*/
		case SPI_DIR_HALF_TX:
			SPI_InitStruct.TransferDirection	= LL_SPI_HALF_DUPLEX_TX ;
			break ;
	}

	// Bit order default: MSB
	SPI_InitStruct.BitOrder = ((flags & SPI_LSB_FIRST) != 0u) ? LL_SPI_LSB_FIRST : LL_SPI_MSB_FIRST ;

	// Data width default: 8
	SPI_InitStruct.DataWidth = flags & SPI_BITS_MASK ;
	if (SPI_InitStruct.DataWidth < 4u)
	{
		SPI_InitStruct.DataWidth = 8u ;		// Default data width if the value is 0
	}
	if (SPI_InitStruct.DataWidth > 32u)
	{
		SPI_InitStruct.DataWidth = 32u ;
	}
	SPI_InitStruct.DataWidth-- ;

	LL_SPI_Init (pSpi, & SPI_InitStruct) ;

	// SPI keeps control of all associated GPIOs (even if master is disabled)
	pSpi->CFG2 |= SPI_CFG2_AFCNTR ;

	LL_SPI_SetStandard (pSpi, LL_SPI_PROTOCOL_MOTOROLA) ;	// Default mode is Motorola

	// Configure SPI End Of Transfer interrupt
	NVIC_SetPriority	(pSpiIo->pSpiCap->irqVector, pSpiDesc->irqPriority) ;
	NVIC_EnableIRQ		(pSpiIo->pSpiCap->irqVector) ;

	// Configure DMA
	spiDmaConfigure (& pSpiDesc->dmaTx, pSpiIo) ;
	if (LL_SPI_GetTransferDirection (pSpi) == LL_SPI_FULL_DUPLEX)
	{
		// Initialize RX DMA only if full duplex mode
		spiDmaConfigure (& pSpiDesc->dmaRx, pSpiIo) ;
	}

	* pHdl = (spiHandle_t) pSpiIo ;
	return SPI_ENONE ;
}

//----------------------------------------------------------------------
//	Data exchange with polling
//	Support only full duplex, or half duplex TX
//	pDataOut or pDataIn can be NULL if they are not necessary
//	Count is an item count (not a byte count)

spiStatus_t spiMasterExchange (spiHandle_t hdl, void * pDataOut, void * pDataIn, uint32_t count)
{
	uint32_t		ii ;
	uint32_t		txSize ;
	uint32_t		rxSize ;
	uint32_t		ffDepth ;
	spiIo_t			* pSpiIo = (spiIo_t *) hdl ;
	SPI_TypeDef		* pSpi = pSpiIo->pSpi ;

	// Check if SPI is configured
dbgPin1 (1) ;
	if ((pSpi->CFG2 & SPI_CFG2_AFCNTR) == 0u)
	{
		return SPI_ENOTINIT ;
	}

	txSize = count ;
	rxSize = count ;

	// Enable and start SPI
	pSpi->CR1 |= SPI_CR1_SPE ;
	pSpi->CR1 |= SPI_CR1_CSTART ;

	ii = (pSpi->CFG1 & SPI_CFG1_DSIZE_Msk) + 1u ;	// Data size

dbgPin1 (0u) ;
	if (ii < 9u)
	{
		// Send/receive data (3 < data size <= 8)
		uint8_t		data ;
		uint8_t		* pOut = (uint8_t *) pDataOut ;
		uint8_t		* pIn  = (uint8_t *) pDataIn ;

		if ((pSpi->CFG2 & SPI_CFG2_COMM) == 0u)
		{
			// Full duplex
			ffDepth = pSpiIo->pSpiCap->fifoDepth ;
			while (txSize != 0u  ||  rxSize != 0u)
			{
				// Write data to send
				if ((txSize > 0u)  &&  ((rxSize - txSize) < ffDepth))
				{
					data = 0u ;
					if (pOut != NULL)
					{
						data = * pOut++ ;
					}
					* ((__IO uint8_t *) & pSpi->TXDR) = data ;
					txSize -- ;
				}

				// Get received data
				while ((pSpi->SR & SPI_SR_RXP) != 0u)
				{
					data = * ((__IO uint8_t *) & pSpi->RXDR) ;
					if (pIn != NULL)
					{
						* pIn++ = data ;
					}
					rxSize -- ;
				}
			}
		}
		else
		{
			// TX only
			while (txSize != 0u)
			{
				// Write data to send
				if ((pSpi->SR & SPI_SR_TXP) != 0u)
				{
					* ((__IO uint8_t *) & pSpi->TXDR) = * pOut++ ;
					txSize -- ;
				}
			}
		}
	}
	else if (ii < 17u)
	{
		// Send/receive data ( 8 < data size <= 16)
		uint16_t	data ;
		uint16_t	* pOut = (uint16_t *) pDataOut ;
		uint16_t	* pIn  = (uint16_t *) pDataIn ;

		if ((pSpi->CFG2 & SPI_CFG2_COMM) == 0u)
		{
			// Full duplex
			ffDepth = pSpiIo->pSpiCap->fifoDepth / 2u ;
			while (txSize != 0u  ||  rxSize != 0u)
			{
				// Write data to send
				if ((txSize > 0u)  &&  ((rxSize - txSize) < ffDepth))
				{
					data = 0u ;
					if (pOut != NULL)
					{
						data = * pOut++ ;
					}
					* ((__IO uint16_t *) & pSpi->TXDR) = data ;
					txSize -- ;
				}

				// Get received data
				while ((pSpi->SR & SPI_SR_RXP) != 0u)
				{
					data = * ((__IO uint16_t *) & pSpi->RXDR) ;
					if (pIn != NULL)
					{
						* pIn++ = data ;
					}
					rxSize -- ;
				}
			}
		}
		else
		{
			// TX only
			while (txSize != 0u)
			{
				// Write data to send
				if ((pSpi->SR & SPI_SR_TXP) != 0u)
				{
					* ((__IO uint16_t *) & pSpi->TXDR) = * pOut++ ;
					txSize -- ;
				}
			}
		}
	}
	else
	{
		// Send/receive data (16 < data size <= 32)
		uint32_t		data ;
		uint32_t		* pOut = (uint32_t *) pDataOut ;
		uint32_t		* pIn  = (uint32_t *) pDataIn ;

		if ((pSpi->CFG2 & SPI_CFG2_COMM) == 0u)
		{
			ffDepth = pSpiIo->pSpiCap->fifoDepth / 4u ;
			while (txSize != 0u  ||  rxSize != 0u)
			{
				// Write data to send
				if ((txSize > 0u)  &&  ((rxSize - txSize) < ffDepth))
				{
					data = 0u ;
					if (pOut != NULL)
					{
						data = * pOut++ ;
					}
					* ((__IO uint32_t *) & pSpi->TXDR) = data ;
					txSize -- ;
				}

				// Get received data
				while ((pSpi->SR & SPI_SR_RXP) != 0u)
				{
					data = * ((__IO uint32_t *) & pSpi->RXDR) ;
					if (pIn != NULL)
					{
						* pIn++ = data ;
					}
					rxSize -- ;
				}
			}
		}
		else
		{
			// TX only
			while (txSize != 0u)
			{
				// Write data to send
				if ((pSpi->SR & SPI_SR_TXP) != 0u)
				{
					* ((__IO uint32_t *) & pSpi->TXDR) = * pOut++ ;
					txSize -- ;
				}
			}
		}
	}
dbgPin1 (1) ;

	// Wait for end of transmission
	while ((pSpi->SR & SPI_SR_TXC) == 0u)
	{
	}

	// Disable SPI
	pSpi->CR1 |= SPI_CR1_CSUSP ;
	pSpi->CR1 &= ~SPI_CR1_SPE ;
	pSpi->IFCR = SPI_IFCR_TXTFC ;	// Clear End Of Transfer. It is necessary, but why?
dbgPin1 (0) ;

	return SPI_ENONE ;
}

//----------------------------------------------------------------------
//	pDataOut is a pointer to Tx data, can't be NULL
//	pDataIn  is a pointer to Rx data:
//		Full duplex: can be NULL if there is nothing to receive
//		TX only: must be NULL
//	Count is an item count (not a byte count)

// BEWARE: DMA buffers have alignment and size constraints (multiple of cache line size, 32 bytes)
// This is especially important for RX buffers
// This function use cache maintenance, see:
// https://community.st.com/s/article/FAQ-DMA-is-not-working-on-STM32H7-devices

spiStatus_t		spiMasterExchangeDma (spiHandle_t hdl, void * pDataOut, void * pDataIn, uint32_t count)
{
	spiIo_t			* pSpiIo = (spiIo_t *) hdl ;
	dma_t			* pDmaTx ;
	dma_t			* pDmaRx ;
	dmaStream_t		* pStreamTx ;
	dmaStream_t		* pStreamRx = NULL ;	// Avoid GCC false message "may be used uninitialized in this function"
	spiStatus_t		sts ;
	aaError_t		res ;
	uint32_t		size ;

dbgPin1 (1) ;
	// Check if SPI is configured
	if ((pSpiIo->pSpi->CFG2 & SPI_CFG2_AFCNTR) == 0u)
	{
		AA_ASSERT (0) ;
		return SPI_ENOTINIT ;
	}

	// SPI6 and DMA not supported
	if (pSpiIo->pSpi == SPI6)
	{
		AA_ASSERT (0) ;
		return SPI_EARG ;
	}

	// pDataOut can't be NULL
	if (pDataOut == NULL)
	{
		AA_ASSERT (0) ;
		return SPI_EARG ;
	}

	pSpiIo->taskId = aaTaskSelfId () ;
	aaSignalClear (AA_SELFTASKID, AA_SIGNAL_IO) ;

	// Set EOT interrupt enable
	pSpiIo->pSpi->IFCR = 0x0FF8u ;			// Clear all interrupts

	// Set length to transfer
	pSpiIo->pSpi->CR2 = count & SPI_CR2_TSIZE ;

	// Compute buffer size aligned on cache lines
	size = pSpiIo->pSpi->CFG1 & SPI_CFG1_DSIZE ;	// Number of bits per word
	if (size < 8u)
	{
		size = 1u ;
	}
	else if (size < 17u)
	{
		size = 2u ;
	}
	else
	{
		size = 4u ;
	}
	size *= count ;		// Buffer size
	size = (size + 31u) & 0xFFFFFFE0u ;	// Round to upper 32 bytes

	// Start RX DMA
	if (pDataIn != NULL)
	{
		// Invalidate D-cache of RX buffer before reception
		SCB_InvalidateDCache_by_Addr ((uint32_t *) pDataIn, (int32_t) size);

		pDmaRx = DMA_GETDMA (pSpiIo->pSpiDesc->dmaRx.dmaNum) ;
		pStreamRx = & pDmaRx->stream [DMA_GETSTREAM (& pSpiIo->pSpiDesc->dmaRx)]  ;
		pStreamRx->CR	&= ~DMA_SxCR_EN ;
		dmaClearFlagsRaw (pDmaRx, DMA_GETSTREAM (& pSpiIo->pSpiDesc->dmaRx), DMA_FLAG_ALLIF) ;
		pStreamRx->NDTR	= count ;
		pStreamRx->M0AR	= (uint32_t) pDataIn ;
		pStreamRx->CR	|= DMA_SxCR_EN ;

		LL_SPI_EnableDMAReq_RX (pSpiIo->pSpi) ;
	}
	// Clean D-cache for TX buffer
	// Make sure the address is 32-byte aligned and add 32-bytes to length, in case it overlaps cacheline
	SCB_CleanDCache_by_Addr ((uint32_t *) (((uint32_t) pDataOut) & ~0x1Fu), (int32_t) size + 32) ;

	// Configure and start TX DMA
	pDmaTx = DMA_GETDMA (pSpiIo->pSpiDesc->dmaTx.dmaNum) ;
	pStreamTx = & pDmaTx->stream [DMA_GETSTREAM (& pSpiIo->pSpiDesc->dmaTx)]  ;
	pStreamTx->CR	&= ~DMA_SxCR_EN ;
	dmaClearFlagsRaw (pDmaTx, DMA_GETSTREAM (& pSpiIo->pSpiDesc->dmaTx), DMA_FLAG_ALLIF) ;
	pStreamTx->NDTR	= count ;
	pStreamTx->M0AR	= (uint32_t) pDataOut ;
	pStreamTx->CR	|= DMA_SxCR_EN ;

	LL_SPI_EnableDMAReq_TX (pSpiIo->pSpi) ;

	// Enable EOT (End of Transfer Interrupt)
	pSpiIo->pSpi->IER = SPI_IER_EOTIE ;

	// Enable SPI and start transfer
	pSpiIo->pSpi->CR1 |= SPI_CR1_SPE ;
	pSpiIo->pSpi->CR1 |= SPI_CR1_CSTART ;

	// Wait end of DMA
dbgPin1 (0u) ;
	res = aaSignalWait (AA_SIGNAL_IO, NULL, AA_SIGNAL_AND, pSpiIo->timeout) ;
dbgPin1 (0u) ;
	if (res == AA_ENONE)
	{
		sts = SPI_ENONE ;
	}
	else if (res == AA_ETIMEOUT)
	{
		sts = SPI_ETIMEOUT ;
	}
	else
	{
		sts = SPI_EFAIL ;	// Other error
	}
	if (pSpiIo->result == SPI_EDMA)
	{
		sts = SPI_EDMA ;
	}
	pSpiIo->taskId = AA_INVALID_TASK ;

	// Wait for end of transmission
	// No need because we received an EOT from the SPI
dbgPin1 (1) ;

	// Disable DMA
	pStreamTx->CR &= ~DMA_SxCR_EN ;
	if (pDataIn != NULL)
	{
		pStreamRx->CR &= ~DMA_SxCR_EN ;
	}

	// Disable SPI
	pSpiIo->pSpi->CR1 &= ~SPI_CR1_SPE ;

	LL_SPI_DisableDMAReq_TX (pSpiIo->pSpi) ;
	LL_SPI_DisableDMAReq_RX (pSpiIo->pSpi) ;

	// Clear SPI SR flags
	pSpiIo->pSpi->IER &= ~SPI_IER_EOTIE ;
	pSpiIo->pSpi->IFCR = SPI_IFCR_TXTFC ;

dbgPin1 (0u) ;
	return sts ;
}

//----------------------------------------------------------------------
//	SPI interrupt to signal EOT to the waiting task

static	void	spiIrqHandler (spiIo_t * pSpiIo)
{
	if ((pSpiIo->pSpi->SR & SPI_SR_EOT) != 0)
	{
		spiCallback (SPI_EV_EOT, 0, (uintptr_t) pSpiIo) ;
		pSpiIo->pSpi->IFCR = SPI_IFCR_EOTC ;	// Clear EOT flag
	}
}


#if (WITH_SPI_1 == 1)
void SPI1_IRQHandler	(void) ;
void SPI1_IRQHandler	(void)
{
	aaIntEnter () ;
	spiIrqHandler (& spiIo1) ;
	aaIntExit () ;
}
#endif	// WITH_SPI_1

#if (WITH_SPI_2 == 1)
void SPI2_IRQHandler	(void) ;
void SPI2_IRQHandler	(void)
{
	aaIntEnter () ;
	spiIrqHandler (& spiIo2) ;
	aaIntExit () ;
}
#endif	// WITH_SPI_2

#if (WITH_SPI_3 == 1)
void SPI3_IRQHandler	(void) ;
void SPI3_IRQHandler	(void)
{
	aaIntEnter () ;
	spiIrqHandler (& spiIo3) ;
	aaIntExit () ;
}
#endif	// WITH_SPI_3

#if (WITH_SPI_4 == 1)
void SPI4_IRQHandler	(void) ;
void SPI4_IRQHandler	(void)
{
	aaIntEnter () ;
	spiIrqHandler (& spiIo4) ;
	aaIntExit () ;
}
#endif	// WITH_SPI_4

#if (WITH_SPI_5 == 1)
void SPI5_IRQHandler	(void) ;
void SPI5_IRQHandler	(void)
{
	aaIntEnter () ;
	spiIrqHandler (& spiIo5) ;
	aaIntExit () ;
}
#endif	// WITH_SPI_5

#if (WITH_SPI_6 == 1)
void SPI6_IRQHandler	(void) ;
void SPI6_IRQHandler	(void)
{
	aaIntEnter () ;
	spiIrqHandler (& spiIo6) ;
	aaIntExit () ;
}
#endif	// WITH_SPI_6

//----------------------------------------------------------------------
