/*
----------------------------------------------------------------------

	AdAstra - Real Time Kernel

	Alain Chebrou

	testdac.c	DAC basic API tests

	When		Who	What
	05/05/18	ac	Creation
	16/03/20	ac	Transform to interactive test
	26/10/21	ac	Port to G0 (no change)
	12/08/22	ac	Add VBUFREF setting
----------------------------------------------------------------------
*/

#include	"aa.h"
#include	"aaprintf.h"
#include	"aalogmes.h"
#include	"dacbasic.h"
#include	"timbasic.h"
#include	"dmabasic.h"

#if (defined DAC1)

#include	"test.h"

void	dacDmaDoubleCallback	(dacHandle_t hdl, uint32_t event, uint32_t channel, uintptr_t arg) ;
void	dacCallbackOneShot		(dacHandle_t hdl, uint32_t event, uint32_t channel, uintptr_t arg) ;
void	dacDmaCircularCallback	(dacHandle_t hdl, uint32_t event, uint32_t channel, uintptr_t arg) ;


#define		TIMER_DAC	6	// Id of timer descriptor used to trigger DAC
#define		TIMER_FREQ	50	// Timer frequency (Hz)

// BEWARE: If DMA is used for I/O, the library must manage the data cache.
// So align buffers address on multiple of cache line size (32 bytes)
// Buffers size should be a multiple of cache line size (32 bytes). Mandatory for RX buffers.

#define	MAX_DATA	8

ALIGN_32BYTES (const uint32_t	dacDataDual [SIZE_32BYTES (MAX_DATA, uint32_t)]) =
{
	0x0FFF0000,
	0x0DB60249,
	0x0B6D0492,
	0x092406BD,
	0x06DB0924,
	0x04920B6D,
	0x02490DB6,
	0x00000FFF,
} ;

ALIGN_32BYTES (const uint16_t	dacDataDual8 [SIZE_32BYTES (MAX_DATA, uint32_t)]) =
{
	0x00FF,
	0x2480,
	0x4940,
	0x6D00,
	0x9200,
	0xB540,
	0xDA80,
	0xFFFF,
} ;

ALIGN_32BYTES (const uint16_t	dacData1 [SIZE_32BYTES (MAX_DATA, uint16_t)]) =
{
	0x0000,
	0x0249,
	0x0492,
	0x06BD,
	0x0924,
	0x0B6D,
	0x0DB6,
	0x0FFF,
} ;

ALIGN_32BYTES (const uint16_t	dacData2 [SIZE_32BYTES (MAX_DATA, uint16_t)]) =
{
	0x0FFF,
	0x0DB6,
	0x0B6D,
	0x0924,
	0x06BD,
	0x0492,
	0x0249,
	0x0000,
} ;

// For double buffer test: use 3 buffers
ALIGN_32BYTES (const uint16_t	dacData3 [SIZE_32BYTES (MAX_DATA, uint16_t)]) =
{
	0x07FF,
	0x07FF,
	0x07FF,
	0x07FF,
	0x07FF,
	0x07FF,
	0x07FF,
	0x07FF,
} ;

const uint16_t	* dataArray [3] =
{
	dacData1,
	dacData2,
	dacData3,
};

static	uint32_t		dataIx ;	// dataArray index of next buffer for double buffer DMA

//----------------------------------------------------------------------
// DAC and DMA interrupts callback: for double buffer DMA management

// Callback functions from DAC and DMA have the same prototype.
//	DMA: source argument is dmaNum << 16 | streamNum
//	DAC: source argument is DAC_CHAN1 or DAC_CHAN2

// This callback is used for DMA double buffer management

void	dacDmaDoubleCallback (dacHandle_t hdl, uint32_t event, uint32_t channel, uintptr_t arg)
{
	aaLogMes ("Double Callback %08X %u\n", event, channel, 0, 0, 0) ;

	// DAC1 channel 1 interrupt management
	if (channel == DAC_CHAN1)
	{
		// Check if DMA complete
		if (event == DMA_EV_COMPLETE)
		{
			(void) dacDmaBuffer	 (hdl, DAC_CHAN1, (void *) dataArray[dataIx], NULL) ;
			dataIx++ ;
			if (dataIx == 3u)
			{
				dataIx = 0u ;
			}
		}

		// Check if DMA underrun
		if (event == DAC_EV_UNDR)
		{
		}
	}

	// DAC1 channel 2 interrupt management
	if (channel == DAC_CHAN2)
	{
		// Check if DMA complete
		if (event == DMA_EV_COMPLETE)
		{
			(void) dacDmaBuffer	 (hdl, DAC_CHAN2, (void *) dataArray[dataIx], NULL) ;
			dataIx++ ;
			if (dataIx == 3u)
			{
				dataIx = 0u ;
			}
		}

		// Check if DMA underrun
		if (event == DAC_EV_UNDR)
		{
		}
	}

	(void) arg ;
}

//----------------------------------------------------------------------
// DMA Callback used with circular buffer

static	uint32_t	circularCounter ;

void	dacDmaCircularCallback (dacHandle_t hdl, uint32_t event, uint32_t channel, uintptr_t arg)
{
	aaLogMes ("Circular Callback %08X %u %u\n", event, arg, circularCounter, 0, 0) ;
	if (event == DMA_EV_HCOMPLETE)
	{
		// Half DMA complete
		circularCounter++ ;
	}
	(void) hdl ;
	(void) channel ;
	(void) arg ;
}

//----------------------------------------------------------------------
// Callback for one shot DMA on channel 2

static	uint32_t	oneShotCounter ;

void	dacCallbackOneShot (dacHandle_t hdl, uint32_t event, uint32_t channel, uintptr_t arg)
{
	aaLogMes ("One shot callback %08X %u\n", event, oneShotCounter, 0, 0, 0) ;
	if (event == DAC_EV_UNDR)
	{
		// Signal end of DMA to the task
		aaSignalSend ((aaTaskId_t) arg, 0x01u) ;
		oneShotCounter++ ;
	}
	(void) channel ;
	(void) hdl ;
}

//----------------------------------------------------------------------

void	dacTest (void)
{

	dacHandle_t			hdl ;
	dacStatus_t			res ;
	uint32_t			ix ;
	const timDesc_t		* pTimDesc ;
	const dacDesc_t 	* pDacDesc ;
	uint32_t			cc ;

	oneShotCounter  = 0u ;
	circularCounter = 0u ;

	while (1)
	{
		aaPrintf ("DAC test, h for test menu\n") ;
		cc = aaGetChar () ;
		aaPutChar ((char) cc) ;
		aaPutChar ('\n') ;

		if (cc == 'h')
		{
			aaPrintf ("c    Calibration, channel 1 and 2\n") ;
			aaPrintf ("w    Triangle generation, channel 1\n") ;
			aaPrintf ("0    Sample & hold mode, channel 1\n") ;
			aaPrintf ("1    Periodical software trigger, channel 1\n") ;
			aaPrintf ("2    Periodical software trigger, channel 2\n") ;
			aaPrintf ("3    Independent channel 2 with single shot DMA, interrupt, timer trigger\n") ;
#if (defined DMA_SxCR_DBM)	// No double buffer on STM32L4+
			aaPrintf ("4    Independent channel 1 with double buffer DMA, no interrupt, timer trigger\n") ;
			aaPrintf ("5    Independent channel 1 with DMA double buffer, timer trigger\n") ;
#endif
			aaPrintf ("6    Independent channel 1 and 2 with circular DMA buffer, common timer trigger\n") ;
			aaPrintf ("7    Dual channels with circular DMA buffer, timer trigger, 12 bits data\n") ;
			aaPrintf ("8    Dual channels with circular DMA buffer, timer trigger, 8 bits data\n") ;
			aaPrintf ("v    VREFBUF setting\n") ;
			aaPrintf ("+    Inc VREFBUF trim\n") ;
			aaPrintf ("-    Dec VREFBUF trim\n") ;
			aaPrintf ("q    Quit test\n") ;
			continue ;
		}

		if (cc == 'q')
		{
			break ;	// End of test
		}

		switch (cc)
		{
			case 'v':		// VREFBUF setting
				{
					uint32_t	mode  = 0xFFu ;
					uint32_t	scale = 0xFFu ;

					aaPrintf ("Mode o/e/i/h: \n") ;
					cc = aaGetChar () ;
					aaPutChar ((char) cc) ;
					aaPutChar ('\n') ;
					switch (cc)
					{
						case 'o':	mode = VREFBUF_MODE_OFF ;	break ;
						case 'e':	mode = VREFBUF_MODE_EXT ;	break ;
						case 'i':	mode = VREFBUF_MODE_INT ;	break ;
						case 'h':	mode = VREFBUF_MODE_HOLD ;	break ;
						default:
							aaPrintf ("Unknown mode\n") ;
							break ;
					}
					if (mode != 0xFFu)
					{
						aaPrintf ("Scale 0/1/2/3: \n") ;
						cc = aaGetChar () ;
						aaPutChar ((char) cc) ;
						aaPutChar ('\n') ;
						switch (cc)
						{
#if (defined STM32G0 || defined STM32L4)
							case '0':	scale = VREFBUF_SCALE0 ;		break ;
							case '1':	scale = VREFBUF_SCALE1 ;		break ;
#elif (defined STM32G4)
							case '0':	scale = VREFBUF_SCALE_2_048 ;	break ;
							case '1':	scale = VREFBUF_SCALE_2_5 ;		break ;
							case '2':	scale = VREFBUF_SCALE_2_90 ;	break ;
#elif (defined STM32H7)
							case '0':	scale = VREFBUF_SCALE_2_5 ;		break ;
							case '1':	scale = VREFBUF_SCALE_2_048 ;	break ;
							case '2':	scale = VREFBUF_SCALE_1_8 ;		break ;
							case '3':	scale = VREFBUF_SCALE_1_5 ;		break ;
#endif
							default:
								aaPrintf ("Unknown scale\n") ;
								break ;
						}
					}
					if (mode != 0xFFu  && scale != 0xFFu)
					{
						vrefbufSet (mode, scale) ;
					}
				}
				break ;

			case '+':		// Inc VREFBUF trim
				{
					uint32_t	trim = (vrefbufGetTrim () + 1u) & VREFBUF_CCR_TRIM ;
					aaPrintf ("  Trim %X\n", trim) ;
					vrefbufSetTrim (trim) ;
				}
				break ;

			case '-':		// Dec VREFBUF trim
				{
					uint32_t	trim = (vrefbufGetTrim () - 1u) & VREFBUF_CCR_TRIM ;
					aaPrintf ("  Trim %X\n", trim) ;
					vrefbufSetTrim (trim) ;
				}
				break ;

			case 'c':		// Channel calibration
				{
					dacStatus_t	done ;

					//	For calibration the DAC buffer must be enabled
					// Calibrate channel 1
					(void) dacGetDesc			(12, & pDacDesc) ;		// DAC Descriptor without DMA on channel 1
					(void) dacInit				(pDacDesc, DAC_MONO1, & hdl) ;
					(void) dacSetChannel		(hdl, DAC_CHAN1, DAC_MODE_HOLD_BUF_EXT, DAC_TRIGGER_NONE) ;	// Set DAC buffer
					done = dacCalibration		(hdl, DAC_CHAN1, 60u) ;
					if (done ==  DAC_ENONE)
					{
						uint32_t	otrim ;
						otrim = dacGetOtrim	(hdl, DAC_CHAN1, & otrim) ;
						aaPrintf ("Calibration channel 1 OK, otrim = %d\n", otrim) ;
					}
					else
					{
						aaPrintf ("Calibration channel 1 error: %u\n", done) ;
					}

					// Calibrate channel 2
					(void) dacGetDesc			(12, & pDacDesc) ;		// DAC Descriptor without DMA on channel 2
					(void) dacInit				(pDacDesc, DAC_MONO1, & hdl) ;
					(void) dacSetChannel		(hdl, DAC_CHAN2, DAC_MODE_HOLD_BUF_EXT, DAC_TRIGGER_NONE) ;	// Set DAC buffer
					done = dacCalibration		(hdl, DAC_CHAN2, 60u) ;
					if (done ==  DAC_ENONE)
					{
						uint32_t	otrim ;
						otrim = dacGetOtrim	(hdl, DAC_CHAN2, & otrim) ;
						aaPrintf ("Calibration channel 2 OK, otrim = %d\n", otrim) ;
					}
					else
					{
						aaPrintf ("Calibration channel 2 error: %u\n", done) ;
					}
				}
				break ;

			case 'w':		// Waveform generation
				(void) dacGetDesc		(11u, & pDacDesc) ;		// DAC Descriptor without DMA on channel 1
				(void) dacInit			(pDacDesc, DAC_MONO1, & hdl) ;
				(void) dacSetChannel	(hdl, DAC_CHAN1, 0u, LL_DAC_TRIG_EXT_TIM6_TRGO) ;
				(void) dacOuput			(hdl, DAC_CHAN1, 0x800u) ;
				(void) dacSetWave		(hdl, DAC_CHAN1, DAC_GEN_WAVE, 0x0Au) ;

				if (TIM_ENONE != timGetDesc (TIMER_DAC, & pTimDesc))
				{
					aaPrintf ("Timer descriptor  %u not found\n", TIMER_DAC) ;
					break ;
				}
				(void) timTimeBaseInit     (pTimDesc, 0u, 5000u, 0u) ;
				(void) timSetTriggerOutput (pTimDesc, LL_TIM_TRGO_UPDATE) ;		// Set timer TRGO

				(void) dacStart (hdl, DAC_CHAN1) ;
				(void) timStart (pTimDesc) ;

				while (1)
				{
					aaTaskDelay (100) ;
					if (aaCheckGetChar () != 0u)
					{
						break ;		// User break
					}
				}
				(void) timStop (pTimDesc) ;
				(void) dacStop (hdl, DAC_CHAN2) ;
				break ;

			case '0':
				// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
				//	Sample & hold mode, channel 1, no trigger -> software update

				(void) dacGetDesc		(12u, & pDacDesc) ;		// DAC Descriptor without DMA on channel 1
				(void) dacInit			(pDacDesc, DAC_MONO1, & hdl) ;
				(void) dacSetChannel	(hdl, DAC_CHAN1, DAC_MODE_HOLD_BUF_EXT, DAC_TRIGGER_NONE) ;
				(void) dacSetHold		(hdl, DAC_CHAN1, 12u, 62u, 4u) ;

				ix = 0u ;
				dacStart (hdl, DAC_CHAN1) ;

				while (1)
				{
					aaTaskDelay (bspGetTickRate ()) ;	// 1 second delay
					dacOuput (hdl, DAC_CHAN1, dacData1 [ix++]) ;
					if (ix == MAX_DATA)
					{
						ix = 0u ;
					}
					if (aaCheckGetChar () != 0u)
					{
						break ;		// User break
					}
				}
				dacStop (hdl, DAC_CHAN1) ;
				break ;

			case '1':
				// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
				//	Periodical software trigger, channel 1 only

				res = dacGetDesc (12u, & pDacDesc) ;		// DAC Descriptor without DMA on channel 2
				if (res != DAC_ENONE)
				{
					aaPrintf ("dacGetDesc error %u\n", res) ;
					break ;
				}

				res = dacInit (pDacDesc, DAC_MONO1, & hdl) ;
				if (res != DAC_ENONE)
				{
					aaPrintf ("dacInit error %u\n", res) ;
					break ;
				}

				res = dacSetChannel	(hdl, DAC_CHAN1, 0u, LL_DAC_TRIG_SOFTWARE) ;
				if (res != DAC_ENONE)
				{
					aaPrintf ("dacSetChannel error %u\n", res) ;
					break ;
				}

				// Initialize DAC
				ix = 0u ;
				dacStart (hdl, DAC_CHAN1) ;

				while (1)
				{
					aaTaskDelay (5) ;
					dacOuput (hdl, DAC_CHAN1, dacData1 [ix++]) ;
					dacSwTrigger (hdl, DAC_CHAN1) ;
					if (ix == MAX_DATA)
					{
						ix = 0u ;
					}
					if (aaCheckGetChar () != 0u)
					{
						break ;		// User break
					}
				}
				dacStop (hdl, DAC_CHAN1) ;
				break ;

			case '2':
				// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
				//	Periodical Software trigger, channel 2 only, 12 bits

				(void) dacGetDesc		(11u, & pDacDesc) ;		// DAC Descriptor without DMA on channel 2
				(void) dacInit			(pDacDesc, DAC_MONO2, & hdl) ;
				(void) dacSetChannel	(hdl, DAC_CHAN2, 0u, LL_DAC_TRIG_SOFTWARE) ;

				ix = 0u ;
				dacStart (hdl, DAC_CHAN2) ;

				while (1)
				{
					aaTaskDelay (5u) ;
					dacOuput (hdl, DAC_CHAN2, dacData1 [ix++]) ;
					dacSwTrigger (hdl, DAC_CHAN2) ;
					if (ix == MAX_DATA)
					{
						ix = 0u ;
					}
					if (aaCheckGetChar () != 0u)
					{
						break ;		// User break
					}
				}
				dacStop (hdl, DAC_CHAN2) ;
				break ;

			case '3':
				// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
				// Independent channel 2 with single shot DMA, interrupt, timer trigger
				// The task is awaken by the DAC overrun interrupt or DMA TC interrupt when the DMA reach the end of the buffer

				// Configure DAC
				(void) dacGetDesc		(12u, & pDacDesc) ;
				(void) dacInit       	(pDacDesc, DAC_MONO2, & hdl) ;
				(void) dacSetChannel 	(hdl, DAC_CHAN2, 0u, LL_DAC_TRIG_EXT_TIM6_TRGO) ;
				(void) dacDmaInit    	(hdl, DAC_CHAN2, DAC_DMA_SINGLE, MAX_DATA) ;
				(void) dacDmaBuffer	 	(hdl, DAC_CHAN2, (void *) dacData1, NULL) ;
				(void) dacSetCallback	(hdl, DAC_CHAN2, dacCallbackOneShot, aaTaskSelfId ()) ;

				// Configure timer
				if (timGetDesc (TIMER_DAC, & pTimDesc) == TIM_EFAIL)
				{
					aaPrintf ("Timer DAC: Unknown descriptor id %u\n", TIMER_DAC) ;
					break ;
				}
				timTimeBaseInit     (pTimDesc, 0u, TIMER_FREQ, 0u) ;
				timSetTriggerOutput (pTimDesc, LL_TIM_TRGO_UPDATE) ;		// Set timer TRGO

				// WARNING: DAC stop drop the output signal to 0. So we don't use dacStop() here.
				// The task will be awaken by aaSignalSend() from dacCallbackOneShot callback
				while (1)
				{
					dacStart (hdl, DAC_CHAN2) ;
					timStart (pTimDesc) ;
					aaSignalWait (0x01, NULL, AA_SIGNAL_AND, AA_INFINITE) ;
					timStop (pTimDesc) ;
					aaTaskDelay (5u) ;
					if (aaCheckGetChar () != 0u)
					{
						break ;		// User break
					}
				}
				break ;

#if (defined DMA_SxCR_DBM)
			// Some DMAs (like L4 +) do not have a double buffered DMA, others (like H7) have it.
			case '4':
				// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
				// Independent channel 1 with double buffer DMA, no interrupt/callback, timer trigger

				// Configure DAC
				(void) dacGetDesc		(11u, & pDacDesc) ;
				(void) dacInit       	(pDacDesc, DAC_MONO1, & hdl) ;
				(void) dacSetChannel 	(hdl, DAC_CHAN1, 0u, LL_DAC_TRIG_EXT_TIM6_TRGO) ;
				(void) dacDmaInit    	(hdl, DAC_CHAN1, DAC_DMA_DOUBLE | DAC_DMA_NOIT, MAX_DATA) ;
				(void) dacDmaBuffer	 	(hdl, DAC_CHAN1, (void *) dacData1, (void *) dacData2) ;

				// Configure timer
				if (timGetDesc (TIMER_DAC, & pTimDesc) == TIM_EFAIL)
				{
					aaPrintf ("Timer DAC: Unknown descriptor id %u\n", TIMER_DAC) ;
					break ;
				}
				timTimeBaseInit     (pTimDesc, 0u, TIMER_FREQ, 0u) ;
				timSetTriggerOutput (pTimDesc, LL_TIM_TRGO_UPDATE) ;		// Set timer TRGO

				dacStart (hdl, DAC_CHAN1) ;
				timStart (pTimDesc) ;
				while (1)
				{
					aaTaskDelay (5u) ;
					if (aaCheckGetChar () != 0u)
					{
						break ;		// User break
					}
				}
				dacStop (hdl, DAC_CHAN1) ;
				timStop (pTimDesc) ;
				break ;

			case '5':
				// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
				// Independent channel 1 with DMA double buffer, timer trigger
				// Multi buffer management by interrupt callback

				// Configure DAC
				(void) dacGetDesc		(11u, & pDacDesc) ;
				(void) dacInit			(pDacDesc, DAC_MONO1, & hdl) ;
				(void) dacSetChannel	(hdl, DAC_CHAN1, 0u, LL_DAC_TRIG_EXT_TIM6_TRGO) ;
				(void) dacDmaInit		(hdl, DAC_CHAN1, DAC_DMA_DOUBLE, MAX_DATA) ;
				(void) dacDmaBuffer		(hdl, DAC_CHAN1, (void *) dacData1, (void *) dacData2) ;
				(void) dacSetCallback	(hdl, DAC_CHAN1, dacDmaDoubleCallback, DAC_CHAN1) ;

				// Configure timer
				if (timGetDesc (TIMER_DAC, & pTimDesc) == TIM_EFAIL)
				{
					aaPrintf ("Timer DAC: Unknown descriptor id %u\n", TIMER_DAC) ;
					break ;
				}
				timTimeBaseInit     (pTimDesc, 0u, TIMER_FREQ, 0u) ;
				timSetTriggerOutput (pTimDesc, LL_TIM_TRGO_UPDATE) ;		// Set timer TRGO

				dataIx = 2u ;	// Next buffer
				dacStart (hdl, DAC_CHAN1) ;
				timStart (pTimDesc) ;
				while (1)
				{
					aaTaskDelay (5u) ;
					if (aaCheckGetChar () != 0u)
					{
						break ;		// User break
					}
				}
				dacStop (hdl, DAC_CHAN1) ;
				timStop (pTimDesc) ;
				break ;
#endif

			case '6':
				// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
				// Independent channel 1 and 2 with circular DMA buffer, common timer trigger
				// Channel 1 use dacData1, channel 2 use dacData2
				// Interrupt enabled

				// Configure DAC
				(void) dacGetDesc		(10u, & pDacDesc) ;
				(void) dacInit       	(pDacDesc, DAC_MONO1 | DAC_MONO2, & hdl) ;
				(void) dacSetChannel	(hdl, DAC_CHAN1, 0u, LL_DAC_TRIG_EXT_TIM6_TRGO) ;
				(void) dacSetChannel	(hdl, DAC_CHAN2, 0u, LL_DAC_TRIG_EXT_TIM6_TRGO) ;
				(void) dacDmaInit		(hdl, DAC_CHAN1, DAC_DMA_CIRCULAR, MAX_DATA) ;
				(void) dacDmaInit		(hdl, DAC_CHAN2, DAC_DMA_CIRCULAR, MAX_DATA) ;
				(void) dacDmaBuffer		(hdl, DAC_CHAN1, (void *) dacData1, NULL) ;
				(void) dacDmaBuffer		(hdl, DAC_CHAN2, (void *) dacData2, NULL) ;

				(void) dacSetCallback	(hdl, DAC_CHAN1, dacDmaCircularCallback, DAC_CHAN1) ;
				(void) dacSetCallback	(hdl, DAC_CHAN2, dacDmaCircularCallback, DAC_CHAN2) ;

				// Configure the timer
				if (timGetDesc (TIMER_DAC, & pTimDesc) == TIM_EFAIL)
				{
					aaPrintf ("Timer DAC: Unknown descriptor id %u\n", TIMER_DAC) ;
					break ;
				}
				timTimeBaseInit     (pTimDesc, 0u, TIMER_FREQ, 0u) ;
				timSetTriggerOutput (pTimDesc, LL_TIM_TRGO_UPDATE) ;		// Set timer TRGO

				dacStart (hdl, DAC_CHAN1) ;
				dacStart (hdl, DAC_CHAN2) ;
				timStart (pTimDesc) ;
				while (1)
				{
					aaTaskDelay (5u) ;
					if (aaCheckGetChar () != 0u)
					{
						break ;		// User break
					}
				}
				dacStop (hdl, DAC_CHAN1) ;
				dacStop (hdl, DAC_CHAN2) ;
				timStop (pTimDesc) ;
				break ;

			case '7':
				// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
				// Dual channels with circular DMA buffer, timer trigger, 12 bits

				// Configure DAC
				(void) dacGetDesc		(11u, & pDacDesc) ;
				(void) dacInit			(pDacDesc, DAC_DUAL, & hdl) ;
				(void) dacSetChannel	(hdl, DAC_CHAN1, 0u, LL_DAC_TRIG_EXT_TIM6_TRGO) ;
				(void) dacSetChannel	(hdl, DAC_CHAN2, 0u, LL_DAC_TRIG_EXT_TIM6_TRGO) ;
				(void) dacDmaInit		(hdl, DAC_CHAN1, DAC_DMA_CIRCULAR, MAX_DATA) ;
				(void) dacDmaBuffer		(hdl, DAC_CHAN1, (void *) dacDataDual, NULL) ;
				(void) dacSetCallback	(hdl, DAC_CHAN1, dacDmaCircularCallback, DAC_CHAN1) ;

				// Configure timer
				if (timGetDesc (TIMER_DAC, & pTimDesc) == TIM_EFAIL)
				{
					aaPrintf ("Timer DAC: Unknown descriptor id %u\n", TIMER_DAC) ;
					break ;
				}
				timTimeBaseInit     (pTimDesc, 0u, TIMER_FREQ, 0u) ;
				timSetTriggerOutput (pTimDesc, LL_TIM_TRGO_UPDATE) ;		// Set timer TRGO

				dacStart (hdl, DAC_CHAN1) ;
				dacStart (hdl, DAC_CHAN2) ;
				timStart (pTimDesc) ;
				while (1)
				{
					aaTaskDelay (5u) ;
					if (aaCheckGetChar () != 0u)
					{
						break ;		// User break
					}
				}
				dacStop (hdl, DAC_CHAN1) ;
				dacStop (hdl, DAC_CHAN2) ;
				timStop (pTimDesc) ;
				break ;

			case '8':
				// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
				// Dual channels with circular DMA buffer, timer trigger, 8 bits

				// Configure DAC
				(void) dacGetDesc		(11u, & pDacDesc) ;
				(void) dacInit			(pDacDesc, DAC_DUAL, & hdl) ;
				(void) dacSetChannel	(hdl, DAC_CHAN1, DAC_BITS8, LL_DAC_TRIG_EXT_TIM6_TRGO) ;
				(void) dacSetChannel	(hdl, DAC_CHAN2, DAC_BITS8, LL_DAC_TRIG_EXT_TIM6_TRGO) ;
				(void) dacDmaInit		(hdl, DAC_CHAN1, DAC_DMA_CIRCULAR, MAX_DATA) ;
				(void) dacDmaBuffer		(hdl, DAC_CHAN1, (void *) dacDataDual8, NULL) ;
				(void) dacSetCallback	(hdl, DAC_CHAN1, dacDmaCircularCallback, DAC_CHAN1) ;

				// Configure timer
				if (timGetDesc (TIMER_DAC, & pTimDesc) == TIM_EFAIL)
				{
					aaPrintf ("Timer DAC: Unknown descriptor id %u\n", TIMER_DAC) ;
					break ;
				}
				timTimeBaseInit     (pTimDesc, 0u, TIMER_FREQ, 0u) ;
				timSetTriggerOutput (pTimDesc, LL_TIM_TRGO_UPDATE) ;		// Set timer TRGO

				dacStart (hdl, DAC_CHAN1) ;
				dacStart (hdl, DAC_CHAN2) ;
				timStart (pTimDesc) ;
				while (1)
				{
					aaTaskDelay (5u) ;
					if (aaCheckGetChar () != 0u)
					{
						break ;		// User break
					}
				}
				dacStop (hdl, DAC_CHAN1) ;
				dacStop (hdl, DAC_CHAN2) ;
				timStop (pTimDesc) ;
				break ;

			default:
				aaPrintf ("Error: Invalid test number\n") ;
				break ;
		}
		if (aaCheckGetChar () != 0u)
		{
			(void) aaGetChar () ;
		}
	}
}

//----------------------------------------------------------------------
#endif	// DAC1
