/*
----------------------------------------------------------------------

	AdAstra - Real Time Kernel

	Alain Chebrou

	testi2c.c	I2C test

	When		Who	What
	26/01/20	ac	Creation

	BEWARE: to get good aaPrinf traces the console TX buffer must be large (>= 256 bytes)

----------------------------------------------------------------------
*/

#include	"aa.h"
#include	"aalogmes.h"
#include 	"aaprintf.h"

#include	"test.h"

#include 	"i2cbasic.h"

//----------------------------------------------------------------------
/*	Cabling for the test:

    NUCLEO-STM32H743ZI or NUCLEO-STM32L4R5ZI

	SDA ---- 15 ------- PB9 CN7.4 ---- PB11 CN10.34 ---- resistor ---- |3.3V CN10.1
	SCL ---- 14 ------- PB8 CN7.2 ---- PB10 CN10.32 ---- resistor ---- |
             PCF8574A   I2C1(Slave)    I2C2(Master)


	NUCLEO-STM32G431RB / NUCLEO-STM32G491RE

	SDA ---- 15 ------- PA8 CN10.23 ---- PC9 CN10.1 ---- resistor ---- |3.3V CN6.4
	SCL ---- 14 ------- PA9 CN10.21 ---- PC8 CN10.2 ---- resistor ---- |
             PCF8574A   I2C2(Slave)      I2C3(Master)

	Pull up resistors: 2.2k or 3.3k Ohms
*/

//----------------------------------------------------------------------
//	I2C Test

// I2C description Id for slave I2C
#if (defined STM32L4 || defined STM32G0 || defined STM32H7)
	#define		MST_DESCID		2
#endif
#if (defined STM32G4)
	#define		MST_DESCID		3
#endif

#define		PFC_ADDR		0x70		// Address of the PCF8574A
#define		I2CSPEED		100000		// Default I2C baud rate
#define		MAXEDGE			200			// max SDA/SCL edge duration (ns)
#define		TASK_SIGNAL		0x8000		// To signal the task

const	char * i2cErrorText [] =
{
		"I2C_ENONE",		// No error, idle state
		"I2C_EFAIL",		// Some error
		"I2C_ESTATE",		// Not allowed for this I2C state

		"I2C_EOK",			// Ended without error
		"I2C_EBUSY",		// Interface is busy
		"I2C_EAF",			// Acknowledge fail
		"I2C_EARLO",		// Arbitration lost
		"I2C_EBERR",		// Bus error
		"I2C_EOVR",			// I2C Overflow
		"I2C_ESLAVERX",		// Slave addressed as receiver
		"I2C_ESLAVETX",		// Slave addressed as transmitter
} ;

//----------------------------------------------------------------------

static	void	ee24cTest (void) ;

//----------------------------------------------------------------------
// Write 1 blocs of 1 byte to the PCF8574A. With full error handling.
// Toggle bit 0 to toggle the LED
// Test I2C_NOBLOCK

static	uint8_t		pfcData [1] = { 0xFF } ;

static	void	i2cTestWriteXfer (uint32_t speed)
{
	const 		i2cDesc_t		* pI2cDesc ;
				i2cHandle_t		hdl ;
	volatile 	i2cStatus_t		res ;
				uint32_t		size ;
				uint32_t		timing ;

	pI2cDesc =	i2cGetDesc	(MST_DESCID) ;

	if (speed == 0u)
	{
		speed = I2CSPEED ;
	}

	pfcData [0] ^= 1u ;	// Toggle the LED

	// ---------------------------- Configure I2C
	timing = i2cTiming (pI2cDesc, speed, MAXEDGE) ;
	res = i2cConfigure (pI2cDesc, timing, TASK_SIGNAL, & hdl) ;
	if (res != I2C_ENONE)
	{
		aaPrintf ("i2cConfigure error %d %s\n", res, i2cErrorText[res]) ;
		AA_ASSERT (0) ;
		return ;
	}

	// ---------------------------- Send Start and slave address
	res = i2cMasterStart (hdl, PFC_ADDR, I2C_NOBLOCK | I2C_TX) ;
	if (res != I2C_ENONE)
	{
		aaPrintf ("i2cMasterStart error %d %s\n", res, i2cErrorText[res]) ;
		AA_ASSERT (0) ;
		i2cDeactivate (hdl) ;
		return ;
	}

	// Wait for end of start
	aaSignalWait (TASK_SIGNAL, NULL, AA_SIGNAL_OR, AA_INFINITE) ;
	res = i2cResult (hdl, & size) ;

	if (res == I2C_EOK)
	{
		// Slave responding
		aaPrintf ("%s Slave ACK\n", i2cErrorText[res]) ;
	}
	else
	{
		// Slave not responding or error
		aaPrintf ("%s Error\n", i2cErrorText[res]) ;
		i2cDeactivate (hdl) ;
		return ;
	}

	// ---------------------------- Send data bytes
	res = i2cMasterXfer (hdl, pfcData, 1u, I2C_NOBLOCK | I2C_LAST) ;

	// Wait for end of xfer
	aaSignalWait (TASK_SIGNAL, NULL, AA_SIGNAL_OR, AA_INFINITE) ;
	res = i2cResult (hdl, & size) ;

	if (res == I2C_EOK)
	{
		aaPrintf ("%s Transmitted %d bytes\n", i2cErrorText[res], size) ;
	}
	else
	{
		aaPrintf ("%s Error %s\n", i2cErrorText[res]) ;
		i2cDeactivate (hdl) ;
		return ;
	}

	// ---------------------------- Send Stop
	(void) i2cMasterStop (hdl) ;
}

//----------------------------------------------------------------------
// Read 1 blocs of 1 byte from the PCF8574A. With full error handling.
// Displays the read data, to show the LED state.
// Test I2C_NOBLOCK

static	void	i2cTestReadXfer (uint32_t speed)
{
	const 		i2cDesc_t		* pI2cDesc ;
				i2cHandle_t		hdl ;
	volatile 	i2cStatus_t		res ;
				uint8_t			data [1] = { 0u } ;
				uint32_t		size ;
				uint32_t		timing ;

	pI2cDesc =	i2cGetDesc	(MST_DESCID) ;

	if (speed == 0u)
	{
		speed = I2CSPEED ;
	}

	// ---------------------------- Configure I2C
	timing = i2cTiming (pI2cDesc, speed, MAXEDGE) ;
	res = i2cConfigure (pI2cDesc, timing, TASK_SIGNAL, & hdl) ;
	if (res != I2C_ENONE)
	{
		aaPrintf ("i2cConfigure error %d %s\n", res, i2cErrorText[res]) ;
		AA_ASSERT (0) ;
		return ;
	}

	// ---------------------------- Send Start and slave address
	res = i2cMasterStart (hdl, PFC_ADDR, I2C_NOBLOCK | I2C_RX | I2C_1LAST) ;
	if (res != I2C_ENONE)
	{
		aaPrintf ("i2cMasterStart error %d %s\n", res, i2cErrorText[res]) ;
		AA_ASSERT (0) ;
		i2cDeactivate (hdl) ;
		return ;
	}

	// Wait for end of start
	aaSignalWait (TASK_SIGNAL, NULL, AA_SIGNAL_OR, AA_INFINITE) ;
	res = i2cResult (hdl, & size) ;

	if (res == I2C_EOK)
	{
		// Slave responding
		aaPrintf ("%s Slave ACK\n", i2cErrorText[res]) ;
	}
	else
	{
		// Slave not responding or error
		aaPrintf ("%s Error\n", i2cErrorText[res]) ;
		i2cDeactivate (hdl) ;
		return ;
	}

	// ---------------------------- Receive data bytes
	res = i2cMasterXfer (hdl, data, 1u, I2C_NOBLOCK | I2C_LAST) ;

	// Wait for end of xfer
	aaSignalWait (TASK_SIGNAL, NULL, AA_SIGNAL_OR, AA_INFINITE) ;
	res = i2cResult (hdl, & size) ;
	if (res == I2C_EOK)
	{
		aaPrintf ("%s Received %d bytes\n", i2cErrorText[res], size) ;
	}
	else
	{
		aaPrintf ("%s Error\n", i2cErrorText[res]) ;
		i2cDeactivate (hdl) ;
		return ;
	}

	// ---------------------------- Send Stop
	(void) i2cMasterStop (hdl) ;
	aaPrintf ("Data: %02X \n", data[0]) ;
}

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//	I2C Slave test
//----------------------------------------------------------------------

// I2C description Id for slave I2C
#if (defined STM32L4 || defined STM32G0 || defined STM32H7)
	#define		SLV_DESCID		1u
#endif
#if defined (STM32G4)
	#define		SLV_DESCID		2u
#endif

#define		MAX_MSG			10u		// Max size of message (buffer size)
#define		SLV_ADDRS		0x50u	// I2c Slave address (left shifted)
#define		SLV_STACKSIZE	128u	// Slave management stack size
#define		SLV_PRIO		8u		// Slave management task priority

static	uint8_t		slvBufferRx [MAX_MSG] ;
static	uint8_t		slvBufferTx [MAX_MSG] = { 0x88u, 0xC8u, 0xE8u, 0xF8u } ;

static	uint32_t	slvBlock		= 1u ;		// 1 or 2 blocks of slvMsgSize bytes
static	uint32_t	slvReady		= 1u ;		// Ready to receive or transmit
static	uint32_t	slvMsgSize		= 2u ; 		// Slave message size, to send or receive

static	uint32_t	masterBlock		= 1u ; 		// 1 or 2 blocks of msgSize bytes
static	uint32_t	masterRestart	= 0u ;		// With or without restart

//----------------------------------------------------------------------
// The task to manage slave transfer
// use function blocking mode

static	void	slaveTask (uintptr_t arg)
{
	const 		i2cDesc_t		* pI2cDesc ;
				i2cHandle_t		hdl ;
	volatile 	i2cStatus_t		res ;		// To avoid optimize out
				uint32_t		timing ;
				uint32_t		size ;

	(void) arg ;

	aaPrintf ("I2C slave task created\n") ;

	pI2cDesc =	i2cGetDesc	(SLV_DESCID) ;

	// ---------------------------- Configure I2C
	timing = i2cTiming (pI2cDesc, I2CSPEED, MAXEDGE) ;
	aaPrintf ("Slave timing %08X\n", timing) ;
	res = i2cConfigure (pI2cDesc, timing, TASK_SIGNAL, & hdl) ;
	if (res != I2C_ENONE)
	{
		aaPrintf ("i2cConfigure error %d %s\n", res, i2cErrorText[res]) ;
		AA_ASSERT (0) ;
		return ;
	}

	i2cSlaveInit (hdl, SLV_ADDRS) ;
	aaPrintf ("I2C slave initialized\n") ;

	while (1)
	{
		// ---------------------------- Set slave in addressable state
		res = i2cSlaveStart (hdl, 0u) ;	// Flag 0 => blocking mode

		if (slvReady == 0u)
		{
			// The slave have nothing to do => NACK
			aaPrintf ("Slave addressed as %s, nothing to transmit or receive\n",
					(res == I2C_ESLAVETX) ? "TX" : "RX") ;
			i2cReset (hdl) ;
			continue ;
		}

		if (res == I2C_ESLAVETX)
		{
			// The master requests the slave to send data

			if (slvBlock == 2)
			{
				// For the test to succeed, the size of the master message must be twice as large as the slave message
				(void) i2cSlaveXfer (hdl, slvBufferTx, slvMsgSize, I2C_NEXT) ;
				res = i2cResult (hdl, & size) ;
				aaPrintf ("  %s Slave TX, size %d / %d\n", i2cErrorText[res], size, slvMsgSize) ;
				aaPrintf ("  Data: %02X %02X %02X %02X\n", slvBufferTx[0], slvBufferTx[1], slvBufferTx[2], slvBufferTx[3]) ;
			}

			(void) i2cSlaveXfer (hdl, slvBufferTx, slvMsgSize, I2C_LAST) ;
			res = i2cResult (hdl, & size) ;
			aaPrintf ("  %s Slave TX, size %d / %d\n", i2cErrorText[res], size, slvMsgSize) ;
			aaPrintf ("  Data: %02X %02X %02X %02X\n", slvBufferTx[0], slvBufferTx[1], slvBufferTx[2], slvBufferTx[3]) ;
		}
		else if (res == I2C_ESLAVERX)
		{
			// The master requests the slave to receive data

			if (slvBlock == 2)
			{
				// For the test to succeed, the size of the master message must be twice as large as the slave message
				aaMEMSET (slvBufferRx, 0, sizeof (slvBufferRx)) ;
				(void) i2cSlaveXfer (hdl, slvBufferRx, slvMsgSize, I2C_NEXT) ;
				res = i2cResult (hdl, & size) ;
				aaPrintf ("  %s Slave RX, size %d / %d\n", i2cErrorText[res], size, slvMsgSize) ;
				aaPrintf ("  Data: %02X %02X %02X %02X\n", slvBufferRx[0], slvBufferRx[1], slvBufferRx[2], slvBufferRx[3]) ;
			}

			aaMEMSET (slvBufferRx, 0, sizeof (slvBufferRx)) ;
			(void) i2cSlaveXfer (hdl, slvBufferRx, slvMsgSize, I2C_LAST) ;
			res = i2cResult (hdl, & size) ;
			aaPrintf ("  %s Slave RX, size %d / %d\n", i2cErrorText[res], size, slvMsgSize) ;
			aaPrintf ("  Data: %02X %02X %02X %02X\n", slvBufferRx[0], slvBufferRx[1], slvBufferRx[2], slvBufferRx[3]) ;
		}
		else
		{
			aaPrintf ("  Invalid result: %d %s, size %d\n", res, i2cErrorText[res], size) ;
		}
	}
}

//----------------------------------------------------------------------
// Master read blocs of msgSize bytes.
// If masterBlock is 2 :         startRX - xfer - xfer - stop
// Else if masterRestart is 1 :  startRX - xfer - startRX - xfer - stop
// Else :                        startRX - xfer - stop

static	void	i2cMasterRxSlaveTx (uint32_t msgSize)
{
	const 		i2cDesc_t		* pI2cDesc ;
				i2cHandle_t		hdl ;
	volatile 	i2cStatus_t		res ;
				uint8_t			data [MAX_MSG] = { 0u } ;
				uint32_t		sentSize ;
				uint32_t		timing ;
				uint32_t		temp ;

	pI2cDesc =	i2cGetDesc	(MST_DESCID) ;

	// ---------------------------- Configure I2C
	timing = i2cTiming (pI2cDesc, I2CSPEED, MAXEDGE) ;
	aaPrintf ("Master timing %08X\n", timing) ;
	res = i2cConfigure (pI2cDesc, timing, TASK_SIGNAL, & hdl) ;
	if (res != I2C_ENONE)
	{
		aaPrintf ("i2cConfigure error %d %s\n", res, i2cErrorText[res]) ;
		AA_ASSERT (0) ;
		return ;
	}

	// ---------------------------- Send Start and slave address
	// If there is only 1 block of 1 byte to receive => special case, need I2C_1LAST
	temp = I2C_RX | ((msgSize == 1u  &&  masterBlock == 1u) ? (uint32_t) I2C_1LAST : 0u) ;
	res = i2cMasterStart (hdl, SLV_ADDRS, temp) ;
	if (res == I2C_EOK)
	{
		// Slave responding
		aaPrintf ("%s Slave ACK\n", i2cErrorText[res]) ;
	}
	else
	{
		// Slave not responding or error
		aaPrintf ("%s Error, deactivate I2C\n", i2cErrorText[res]) ;
		i2cDeactivate (hdl) ;
		return ;
	}

	if (masterBlock == 2)
	{
		//	Receive 1st block
		// ---------------------------- Receive data bytes
		aaMEMSET (data, 0, sizeof (data)) ;
		(void) i2cMasterXfer (hdl, data, msgSize, I2C_NEXT) ;
		res = i2cResult (hdl, & sentSize) ;
		aaPrintf ("%s, Master Received %d / %d bytes\n", i2cErrorText[res], sentSize, msgSize) ;
		aaPrintf ("Data: %02X %02X %02X %02X\n", data[0], data[1], data[2], data[3]) ;
	}
	else
	{
		if (masterRestart == 1u)
		{
			//	Receive 1st block and ReStart
			// ---------------------------- Receive data bytes
			aaMEMSET (data, 0, sizeof (data)) ;
			(void) i2cMasterXfer (hdl, data, msgSize, I2C_LAST) ;
			res = i2cResult (hdl, & sentSize) ;
			aaPrintf ("%s, Master Received %d / %d bytes\n", i2cErrorText[res], sentSize, msgSize) ;
			aaPrintf ("Data: %02X %02X %02X %02X\n", data[0], data[1], data[2], data[3]) ;

			// ---------------------------- Send Start and slave address
			// If there is only 1 byte to receive => special case, need I2C_1LAST
			temp = I2C_RX | ((msgSize == 1u) ? (uint32_t) I2C_1LAST : 0u) ;
			res = i2cMasterStart (hdl, SLV_ADDRS, temp) ;
			aaPrintf ("%s Master ReSTART\n", i2cErrorText[res]) ;

			if (res == I2C_EOK)
			{
				// Slave responding
				aaPrintf ("%s Slave ACK\n", i2cErrorText[res]) ;
			}
			else
			{
				// Slave not responding or error
				aaPrintf ("%s Error, deactivate I2C\n", i2cErrorText[res]) ;
				i2cDeactivate (hdl) ;
				return ;
			}
		}
	}

	// Receive last block
	// ---------------------------- Receive data bytes
	aaMEMSET (data, 0, sizeof (data)) ;
	(void) i2cMasterXfer (hdl, data, msgSize, I2C_LAST) ;
	res = i2cResult (hdl, & sentSize) ;

	// ---------------------------- Send Stop
	(void) i2cMasterStop (hdl) ;
	aaPrintf ("%s, Master Received %d / %d bytes\n", i2cErrorText[res], sentSize, msgSize) ;
	aaPrintf ("Data: %02X %02X %02X %02X\n", data[0], data[1], data[2], data[3]) ;
}

//----------------------------------------------------------------------
// Master write blocs of msgSize bytes

// If masterBlock is 2 :         startTX - xfer - xfer - stop
// Else if masterRestart is 1 :  startTX - xfer - startTX - xfer - stop
// Else (masterBlock is 1):      startTX - xfer - stop

static	void	i2cMasterTxSlaveRx (uint32_t msgSize)
{
	const 		i2cDesc_t		* pI2cDesc ;
				i2cHandle_t		hdl ;
	volatile 	i2cStatus_t		res ;
				uint8_t			data [MAX_MSG] = { 'M', 'a', 's', 't' } ;
				uint32_t		sentSize ;
				uint32_t		timing ;

	pI2cDesc =	i2cGetDesc	(MST_DESCID) ;

	// ---------------------------- Configure I2C
	timing = i2cTiming (pI2cDesc, I2CSPEED, MAXEDGE) ;
	aaPrintf ("Master timing %08X\n", timing) ;
	res = i2cConfigure (pI2cDesc, timing, TASK_SIGNAL, & hdl) ;
	if (res != I2C_ENONE)
	{
		aaPrintf ("i2cConfigure error %d %s\n", res, i2cErrorText[res]) ;
		AA_ASSERT (0) ;
		return ;
	}

	// ---------------------------- Send Start and slave address
	res = i2cMasterStart (hdl, SLV_ADDRS, I2C_TX) ;

	if (res == I2C_EOK)
	{
		// Slave responding
		aaPrintf ("%s Slave ACK\n", i2cErrorText[res]) ;
	}
	else
	{
		// Slave not responding or error
		aaPrintf ("%s Error, deactivate I2C\n", i2cErrorText[res]) ;
		i2cDeactivate (hdl) ;
		return ;
	}

	if (masterBlock == 2)
	{
		(void) i2cMasterXfer (hdl, data, msgSize, I2C_NEXT) ;
		res = i2cResult (hdl, & sentSize) ;
		aaPrintf ("%s, Master Sent %d / %d bytes\n", i2cErrorText[res], sentSize, msgSize) ;
	}
	else
	{
		if (masterRestart == 1u)
		{
			//	Transmit 1st block and ReStart
			// ---------------------------- Transmit data bytes
			(void) i2cMasterXfer (hdl, data, msgSize, I2C_LAST) ;
			res = i2cResult (hdl, & sentSize) ;
			aaPrintf ("%s, Master Sent %d / %d bytes\n", i2cErrorText[res], sentSize, msgSize) ;

			// ---------------------------- Send ReStart and slave address
			res = i2cMasterStart (hdl, SLV_ADDRS, I2C_TX) ;
			aaPrintf ("%s Master ReSTART\n", i2cErrorText[res]) ;

			if (res == I2C_EOK)
			{
				// Slave responding
				aaPrintf ("%s Slave ACK\n", i2cErrorText[res]) ;
			}
			else
			{
				// Slave not responding or error
				aaPrintf ("%s Error, deactivate I2C\n", i2cErrorText[res]) ;
				i2cDeactivate (hdl) ;
				return ;
			}
		}
	}

	// ---------------------------- Transmit data bytes
	(void) i2cMasterXfer (hdl, data, msgSize, I2C_LAST) ;
	res = i2cResult (hdl, & sentSize) ;

	// ---------------------------- Send Stop
	(void) i2cMasterStop (hdl) ;
	aaPrintf ("%s, Master Sent %d / %d bytes\n", i2cErrorText[res], sentSize, msgSize) ;
	aaPrintf ("Data: %02X %02X %02X %02X\n", data[0], data[1], data[2], data[3]) ;
}

//----------------------------------------------------------------------
//----------------------------------------------------------------------
// test of i2cMasterProbeDevice()

static	void	i2cTestProbe (uint32_t bExist)
{
	const i2cDesc_t			* pI2cDesc ;
	i2cHandle_t				hdl ;
	volatile i2cStatus_t	res ;
	uint32_t				timing, addrs ;

	pI2cDesc =	i2cGetDesc	(MST_DESCID) ;

	// Configure I2C
	timing = i2cTiming (pI2cDesc, I2CSPEED, MAXEDGE) ;
	res = i2cConfigure (pI2cDesc, timing, TASK_SIGNAL, & hdl) ;
	if (res != I2C_ENONE)
	{
		aaPrintf ("i2cConfigure error %d %s\n", res, i2cErrorText[res]) ;
		AA_ASSERT (0) ;
		return ;
	}

	addrs = (bExist == 1u) ? SLV_ADDRS : SLV_ADDRS+2u ;	// On board slave
//	addrs = (bExist == 1u) ? PFC_ADDR : PFC_ADDR+2u ;	// PCF8574A slave
	res = i2cMasterProbeDevice (hdl, addrs) ;
	aaPrintf ("i2cMasterCheckDevice %02X : %d %s\n", addrs, res, i2cErrorText[res]) ;
}

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//	I2C interactive tests
//	Use one master I2C and one slave I2C on the same board
//	The master have the descriptor ID MST_DESCID, and the slave have the descriptor ID SLV_DESCID

void	i2cTest ()
{
	char		cc ;
	aaTaskId_t	taskId = AA_INVALID_TASK ;
	uint32_t	msgSize = 4u ;

	aaPrintf ("h for test menu\n") ;
	aaPrintf ("t to start the slave task\n") ;
	while (1)
	{
		cc = aaGetChar () ;
		aaPutChar (cc) ;
		aaPutChar ('\n') ;

		if (cc == 'q')
		{
			break ;	// End of test
		}

		if (cc == ' ')		// Line space on message log terminal
		{
			aaLogMes ("\n", 0, 0, 0, 0, 0) ;
		}

		if (cc == 'h')
		{
			aaPrintf ("?    Print test configuration\n") ;
			aaPrintf ("r    Master read  PCF8574A\n") ;
			aaPrintf ("w    Master Write PCF8574A\n") ;
			aaPrintf ("D    Device Probe OK\n") ;
			aaPrintf ("d    Device Probe fail\n") ;
			aaPrintf ("i    Test timing\n") ;
			aaPrintf ("p    (put) Master TX\n") ;
			aaPrintf ("g    (get) Master RX\n") ;

			aaPrintf ("t    Initialize slave task\n") ;
			aaPrintf ("+-   Inc / Dec master message size\n") ;
			aaPrintf ("s v  Slave message size (1..9)\n") ;
			aaPrintf ("B    Toggle master message block (1..2)\n") ;
			aaPrintf ("b    Toggle slave  message block (1..2)\n") ;
			aaPrintf ("y    Toggle slave ready\n") ;
			aaPrintf ("e    Toggle master restart on/off\n") ;

			aaPrintf ("m    EEPROM 24Cxx test\n") ;
			aaPrintf ("q    Quit test\n") ;
		}

		if (cc == 'r')
		{
			i2cTestReadXfer (0u) ;	// Master read  PCF8574A
		}

		if (cc == 'w')
		{
			i2cTestWriteXfer (0u) ;	// Master Write PCF8574A
		}

		if (cc == 'd')
		{
			i2cTestProbe (0u) ;		// Non existing device
		}

		if (cc == 'D')
		{
			i2cTestProbe (1u) ;		// Existing device
		}

		if (cc == 'i')
		{
			// Generate some timing register values and test
			const i2cDesc_t		* pI2cDesc ;
			uint32_t			timing, speed ;

			pI2cDesc =	i2cGetDesc	(MST_DESCID) ;

			speed = 5000 ;
			while (speed <= 100000)
			{
				timing = i2cTiming (pI2cDesc, speed, MAXEDGE) ;
				aaPrintf ("\ntiming %6d %08X\n", speed, timing) ;

				// Test with this timing
//				i2cTestReadXfer (speed) ;
				speed += 5000 ;
			}
		}

		if (cc == 't')				// Start I2C slave task
		{
			if (taskId == AA_INVALID_TASK)
			{
				aaTaskCreate (	SLV_PRIO,		// Task priority
								"tI2cSlv",		// Task name
								slaveTask,		// Entry point,
								0u,				// Entry point parameter
								NULL,			// Stack pointer
								SLV_STACKSIZE,	// Stack size
								0u,				// Flags
								& taskId) ;		// Created task id
			}
			else
			{
				aaPrintf ("Already created\n") ;
			}
		}

		if (cc == '+')
		{
			if (msgSize < MAX_MSG)
			{
				msgSize++ ;
			}
			aaPrintf ("Master message size: %d\n", msgSize) ;
		}
		if (cc == '-')
		{
			if (msgSize > 1u)
			{
				msgSize-- ;
			}
			aaPrintf ("Master message size: %d\n", msgSize) ;
		}

		if (cc == 's')	// Slave message size
		{
			aaPrintf ("Slave message size (1..9): ") ;
			cc = aaGetChar () ;
			aaPutChar (cc) ;
			if (cc >= '1'  &&  cc <= '9')
			{
				slvMsgSize = (uint32_t) (cc - '0') ;
			}
			aaPrintf ("\nSlave message size: %d\n", slvMsgSize) ;
		}

		if (cc == 'B')	// Count of Master blocks
		{
			masterBlock = (masterBlock == 1u) ? 2u : 1u ;
			aaPrintf ("Master block blocks: %d\n", masterBlock) ;
		}

		if (cc == 'b')	// Count of Slave blocks
		{
			slvBlock = (slvBlock == 1u) ? 2u : 1u ;
			aaPrintf ("Slave blocks: %d\n", slvBlock) ;
		}

		if (cc == 'e')	// Master ReSTART on/off
		{
			masterRestart ^= 1u ;
			aaPrintf ("Master ReSTART: %d\n", masterRestart) ;
		}

		if (cc == 'y')	// Slave ready on/off
		{
			slvReady ^= 1u ;
			aaPrintf ("Slave ready:  %d\n", slvReady) ;
		}

		if (cc == 'g')
		{
			// Get: Master RX, slave TX
			i2cMasterRxSlaveTx (msgSize) ;
		}

		if (cc == 'p')
		{
			// Put: Master TX, slave RX
			i2cMasterTxSlaveRx (msgSize) ;
		}

		if (cc == '?')	// Print test configuration
		{
			aaPrintf ("\nMaster message size: %d\n", msgSize) ;
			aaPrintf ("Master ReSTART:      %d\n", masterRestart) ;
			aaPrintf ("Master blocks:       %d\n", masterBlock) ;
			aaPrintf ("Slave message size:  %d\n", slvMsgSize) ;
			aaPrintf ("Slave blocks:        %d\n", slvBlock) ;
			aaPrintf ("Slave ready:         %d\n", slvReady) ;
		}

		if (cc == 'm')
		{
			ee24cTest () ;	// EEPROM 24Cxx test
		}
	}
}

//----------------------------------------------------------------------
//----------------------------------------------------------------------
//	EE24CXX test

#include	"aautils.h"
#include	"ee24Cxx.h"

/*	How to connect the EEPROM 24Cxx for the test

    H743 / L4R5:
                      -------
	     GND A0 ------|1   8|------ VCC CN10-1
	     GND A1 ------|2   7|------ WP  NC
	     GND A2 ------|3   6|------ SCL PB10 CN10-32
	 CN10-5 GND ------|4   5|------ SDA PB11 CN10-34
	                  -------

    G431 / G491:
                      -------
	     GND A0 ------|1   8|------ VCC CN7-5
	     GND A1 ------|2   7|------ WP  NC
	     GND A2 ------|3   6|------ SCL PC9 CN10-1
	  CN7-8 GND ------|4   5|------ SDA PC8 CN10-2
	                  -------
*/

#define	EE24CADDDR	0xA0		// I2C address of the EEPROM (left shifted)

// Data buffer to write / read
static	uint8_t	eeBuffer [256] ;

// The I2C descriptor to use with the EEPROM
#if (defined STM32L4 || defined STM32H7)
static	const i2cDesc_t		i2cEeDesc =
{
	2,						// Descriptor identifier
	{	AA_GPIO_PORTB,		// port Sda
		11,					// pin  Sda
		AA_GPIO_AF_4,		// af   Sda
		AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH | AA_GPIO_OPEN_DRAIN | AA_GPIO_PULL_NO,
	},
	{	AA_GPIO_PORTB,		// port Scl
		10,					// pin  Scl
		AA_GPIO_AF_4,		// af   Scl
		AA_GPIO_MODE_ALTERNATE | AA_GPIO_SPEED_HIGH | AA_GPIO_OPEN_DRAIN | AA_GPIO_PULL_NO,
	},
	I2C2,					// I2C_TypeDef ;
	BSP_IRQPRIOMAX_MINUS(2),
} ;
#endif	// L4 H7

#if (defined STM32G4)
static	const i2cDesc_t		i2cEeDesc =
{
	2,						// Descriptor identifier
	{	AA_GPIO_PORTC,		// port Sda
		9,					// pin  Sda
		AA_GPIO_AF_8,		// af   Sda
		AA_GPIO_MODE_ALT_OD | AA_GPIO_SPEED_HIGH | AA_GPIO_PULL_NO,
	},
	{	AA_GPIO_PORTC,		// port Scl
		8,					// pin  Scl
		AA_GPIO_AF_8,		// af   Scl
		AA_GPIO_MODE_ALT_OD | AA_GPIO_SPEED_HIGH | AA_GPIO_PULL_NO,
	},
	I2C3,					// I2C_TypeDef ;
	BSP_IRQPRIOMAX_MINUS(2),
} ;
#endif	// G4

#if (defined STM32G0)
static	const i2cDesc_t		i2cEeDesc =
{
	2,						// Descriptor identifier
	{	AA_GPIO_PORTB,		// port Sda
		14,					// pin  Sda
		AA_GPIO_AF_6,		// af   Sda
		AA_GPIO_MODE_ALT_OD | AA_GPIO_SPEED_HIGH | AA_GPIO_PULL_NO,
	},
	{	AA_GPIO_PORTB,		// port Scl
		13,					// pin  Scl
		AA_GPIO_AF_6,		// af   Scl
		AA_GPIO_MODE_ALT_OD | AA_GPIO_SPEED_HIGH | AA_GPIO_PULL_NO,
	},
	I2C2,					// I2C_TypeDef ;
	BSP_IRQPRIOMAX_MINUS(2),
} ;
#endif	// G0

//----------------------------------------------------------------------
//	Check the content of a read buffer

static	i2cStatus_t	eeCheck (uint8_t * pBuffer, uint32_t size)
{
	uint32_t	ii ;

	for (ii = 0u ; ii < size ; ii++)
	{
		if (pBuffer [ii] != ii)
		{
			return I2C_EFAIL ;
		}
	}
	return (I2C_EOK) ;
}

//----------------------------------------------------------------------

static	void	ee24cTest (void)
{
	ee24cDesc_t		eeDesc ;
	i2cStatus_t		sts ;
	uint32_t		size, offset, ii, timing ;

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// Initialize the I2C interface used by the EEPROM
	timing = i2cTiming (& i2cEeDesc, I2CSPEED, MAXEDGE) ;
	sts = i2cConfigure (& i2cEeDesc, timing, TASK_SIGNAL, & eeDesc.i2cHdl) ;
	if (sts != I2C_ENONE)
	{
		aaPrintf ("i2cConfigure error %d %s\n", sts, i2cErrorText[sts]) ;
		AA_ASSERT (0) ;
		return ;
	}

	// Initialize the EEPROM descriptor
	sts = ee24cInit (& eeDesc, EE24C02, EE24CADDDR) ;
	if (sts != I2C_EOK)
	{
		aaPrintf ("ee24cInit error: %d\n") ;
		return ;
	}

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// Write 4 bytes at the beginning of page 0
	aaMEMSET (eeBuffer, 0, sizeof (eeBuffer)) ;
	for (ii = 0u ; ii < 4u ; ii++)
	{
		eeBuffer [ii] = (uint8_t) ii ;
	}

	size   = 4u ;
	offset = 0u ;
	sts = ee24cWrite (& eeDesc, offset, eeBuffer, size) ;
	if (sts == I2C_EOK)
	{
		aaPrintf ("Write page 0, %u bytes OK sts = %d\n", size, sts) ;
		aaDump (eeBuffer, size) ;
	}
	else
	{
		aaPrintf ("Write page 0 error sts = %d\n", sts) ;
		return ;
	}

	// Read page 0
	aaMEMSET (eeBuffer, 0, sizeof (eeBuffer)) ;
	size   = EE24C_PAGESIZE (eeDesc.eeType) ;
	offset = 0u ;
	sts = ee24cRead (& eeDesc, offset, eeBuffer, size) ;
	if (sts == I2C_EOK)
	{
		aaPrintf ("Read page 0, %u bytes OK sts = %d\n", size, sts) ;
		aaDump (eeBuffer, size) ;
		if (I2C_EFAIL == eeCheck (eeBuffer, 4))
		{
			aaPrintf ("Read content error\n") ;
			return ;
		}
	}
	else
	{
		aaPrintf ("Read page 0 error sts = %d\n", sts) ;
		return ;
	}

	// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	// Write more than 1 page, from middle of page 4
	aaMEMSET (eeBuffer, 0, sizeof (eeBuffer)) ;
	size   = 2 * EE24C_PAGESIZE (eeDesc.eeType) ;
	offset = (EE24C_PAGESIZE (eeDesc.eeType) * 4) + (EE24C_PAGESIZE (eeDesc.eeType) / 2) ;
	for (ii = 0u ; ii < 2u * EE24C_PAGESIZE (eeDesc.eeType) ; ii++)
	{
		eeBuffer [ii] = (uint8_t) ii ;
	}
	sts = ee24cWrite (& eeDesc, offset, eeBuffer, size) ;
	if (sts == I2C_EOK)
	{
		aaPrintf ("\nWrite page 4, %u bytes OK sts = %d\n", size, sts) ;
		aaDump (eeBuffer, size) ;
	}
	else
	{
		aaPrintf ("\nWrite page 4 error sts = %d\n", sts) ;
		return ;
	}

	// Read data from page 4
	aaMEMSET (eeBuffer, 0, sizeof (eeBuffer)) ;
	sts = ee24cRead (& eeDesc, offset, eeBuffer, size) ;
	if (sts == I2C_EOK)
	{
		aaPrintf ("Read page 4, %u bytes OK sts = %d\n", size, sts) ;
		aaDump (eeBuffer, size) ;
		if (I2C_EFAIL == eeCheck (eeBuffer, size))
		{
			aaPrintf ("Read content error\n") ;
			return ;
		}
	}
	else
	{
		aaPrintf ("Read page 4 error sts = %d\n", sts) ;
		return ;
	}

	aaPrintf ("EEPROM test OK\n") ;
}

//----------------------------------------------------------------------

