/*
----------------------------------------------------------------------

	AdAstra - Real Time Kernel

	Alain Chebrou

	testtim.c	timer basic API tests

	When		Who	What
	08/03/18	ac	Creation
	12/07/21	ac	Add input capture test

----------------------------------------------------------------------
*/

//	Timer basic : API tests

#include	"aa.h"
#include	"aaprintf.h"
#include	"aalogmes.h"

#include	"timbasic.h"
#include	"test.h"

//------------------------------------------------------------------------------
// This test can test PWM or timebase.
// Timebase use interrupt and callback, but no output pin.
// PWM output on PA0

static	void	tim2Callback	(const timDesc_t * pTimDesc, uintptr_t arg, uint32_t cause) ;

#define	PWM_ID			200u		// Timer descriptor id for TIM2.
#define	PWM_FREQ		19000u		// PWM frequency (PWM test)
#define	PWM_CHANNEL		1u			// 1..4.  PWM is output on PA0

#define	TIMEBASE		100u		// Time base frequency (time base test)
#define	TIM_PRIORITY	BSP_IRQPRIOMAX_MINUS(2)		// (time base test)

// The time base test use an interrupt callback
// The only cause is UPDATE
// cause argument is the pTimDesc->pTim->SR register: Do not use SR register in the callback
// User timGetRegPtr() to get a pointer to the timer registers
static	void tim2Callback (const timDesc_t * pTimDesc, uintptr_t arg, uint32_t cause)
{
	aaLogMes ("%c", arg, 0, 0, 0, 0) ;
	(void) pTimDesc ;
	(void) arg ;
	(void) cause ;
	// UIF flag is cleared by interrupt handler
}

//	bPwm = 0 : Frequency gene
//	bPwm = 1 : PWM gene

void	timPwmTest (uint32_t bPwm)
{
	const timDesc_t	* pTimDesc ;
	uint32_t		reload ;

	if (bPwm == 1u)
	{
		aaPrintf ("Timer PWM out test\n") ;
	}
	else
	{
		aaPrintf ("Timer time base test\n") ;
	}

	if (timGetDesc (PWM_ID, & pTimDesc) == TIM_EFAIL)
	{
		aaPrintf ("Unknown timer descriptor id: %u\n", PWM_ID) ;
		return ;
	}

	if (bPwm == 1u)
	{
		// Test PWM
		timTimeBaseInit (pTimDesc, 0u, PWM_FREQ, TIM_INIT_ARR_PRELOADEN) ;
		reload = timGetReload (pTimDesc) + 1u ;		// To allows to compute PWM ratio
		aaPrintf ("Reload value: %u\n", reload) ;

		timOutputChannelInit (pTimDesc, PWM_CHANNEL, TIM_OCMODE_PWM1 | TIM_OCMODE_PRELOADEN) ;
		timChannelEnable     (pTimDesc, 1u << PWM_CHANNEL) ;

		timStart (pTimDesc) ;

		while (1)
		{
			aaTaskDelay (bspGetTickRate ()) ;
			timOutputChannelSet (pTimDesc, PWM_CHANNEL, (reload * 2u) / 10u) ;	// 20%

			aaTaskDelay (bspGetTickRate ()) ;
			timOutputChannelSet (pTimDesc, PWM_CHANNEL, (reload * 8u) / 10u) ;	// 80%

			if (aaCheckGetChar () != 0u)
			{
				timStop (pTimDesc) ;
				break ;		// Test cancelled by the user
			}
		}
	}
	else
	{
		// Test time base with UPDATE interrupt (UIF)
		// No GPIO output
		timTimeBaseInit (pTimDesc, 0u, TIMEBASE, TIM_INIT_ARR_PRELOADEN) ;
		timConfigureIntr (pTimDesc, TIM_PRIORITY, tim2Callback, '/', TIM_INTR_UPDATE) ;
		timStart (pTimDesc) ;
		while (1)
		{
			aaTaskDelay (1000u / bspGetTickRate ()) ;

			if (aaCheckGetChar () != 0u)
			{
				timStop (pTimDesc) ;
				break ;		// Test cancelled by the user
			}
		}
	}
	timReset (pTimDesc) ;
}

//------------------------------------------------------------------------------
//	Interactive frequency generator
//	Use the timer used by timPwmTest()

#include "stdlib.h"		// for strtoul
#include "string.h"		// for strtok_r

void	timFrequencyGeneratorTest ()
{
	const timDesc_t	* pTimDesc ;
	char			cmd [32] ;
	char			* pStr, * pCmd ;
	uint32_t		fr ;

	aaPrintf ("Frequency generator, q for quit\n") ;

	if (timGetDesc (PWM_ID, & pTimDesc) == TIM_EFAIL)
	{
		aaPrintf ("Unknown timer descriptor id: %u\n", PWM_ID) ;
		return ;
	}

	timTimeBaseInit (pTimDesc, 0u, PWM_FREQ, TIM_INIT_ARR_PRELOADEN) ;
	timOutputChannelInit (pTimDesc, PWM_CHANNEL, TIM_OCMODE_PWM1 | TIM_OCMODE_PRELOADEN) ;
	timChannelEnable     (pTimDesc, 1u << PWM_CHANNEL) ;

	timStart (pTimDesc) ;

	while (1)
	{
		aaGets (cmd, sizeof (cmd)) ;
		pCmd = strtok_r (cmd, " \t", & pStr);
		if (pCmd == NULL)
		{
			continue ;
		}
		if (* pCmd == 'q')
		{
			timStop (pTimDesc) ;
			break ;
		}

		fr = strtoul (pCmd, NULL, 0) ;
		if (fr != 0)
		{
			timTimeBaseInit (pTimDesc, 0u, fr, TIM_INIT_ARR_PRELOADEN) ;
			timOutputChannelInit (pTimDesc, PWM_CHANNEL, TIM_OCMODE_PWM1 | TIM_OCMODE_PRELOADEN) ;
			timChannelEnable     (pTimDesc, 1u << PWM_CHANNEL) ;
			timStart (pTimDesc) ;
		}
	}
}

//------------------------------------------------------------------------------
//	Test one shot pulse
//	Positive pulse from 1 to 10 us, 1 us resolution

void	timOneShotTest (void)
{
	const timDesc_t	* pTimDesc ;
	uint32_t		reload ;
	TIM_TypeDef 	* pTim ;

	aaPrintf ("Timer single pulse test\n") ;

	// Configure the timer
	if (timGetDesc (PWM_ID, & pTimDesc) == TIM_EFAIL)
	{
		aaPrintf ("Unknown timer descriptor id: %u\n", PWM_ID) ;
		return ;
	}

	// Configure the time base for 1MHz timer clock and 11 us period (period is ARR+1)
	timGetRegPtr (pTimDesc, & pTim) ;	// Get native pointer to TIM registers
	aaPrintf ("Timer clock frequency: %u Hz\n", rccGetTimerClockFrequency (pTim)) ;

	timTimeBaseInit (pTimDesc, (rccGetTimerClockFrequency (pTim)) / 1000000u, 10u, TIM_INIT_ONEPULSE_MODE | TIM_INIT_FORCE_ARR) ;
	reload = timGetReload (pTimDesc) ;
	aaPrintf ("Reload value: %u\n", reload) ;

	// If PWM1: add TIM_OCMODE_POLLOW to get a positive pulse
	timOutputChannelInit	(pTimDesc, PWM_CHANNEL, TIM_OCMODE_PWM1 | TIM_OCMODE_POLLOW) ;
	timOutputChannelSet		(pTimDesc, PWM_CHANNEL, 1u) ;	// => pulse of: ARR+1 - 1  = 10 us
//	timOutputChannelSet		(pTimDesc, PWM_CHANNEL, 9u) ;	// => pulse of: ARR+1 - 9  = 2 us
//	timOutputChannelSet		(pTimDesc, PWM_CHANNEL, 10u) ;	// => pulse of: ARR+1 - 10 = 1 us
	timChannelEnable		(pTimDesc, 1u << PWM_CHANNEL) ;

	while (1)
	{
		timStart (pTimDesc) ;
		aaTaskDelay (bspGetTickRate () / 2u) ;
		aaPrintf (".") ;

		if (aaCheckGetChar () != 0u)
		{
			break ;		// Test cancelled by the user
		}
	}
}

//------------------------------------------------------------------------------
//	Radio Controller signal generator

static	void	rcCallback	(const timDesc_t * pTimDesc, uintptr_t arg, uint32_t cause) ;


#define	RC_MIN			1000u		// Minimum pulse width (us)
#define	RC_GAP			300u			// Gap between pulses
#define	RC_COUNT		6u			// Count of pulses

#define	RC_ID			200u		// Timer descriptor id for TIM2.
#define	RC_CHANNEL		1u			// 1..4.  PWM is output on PA0

// Pulses duration
uint32_t	rcValues [RC_COUNT] =
{
	   0u,
	1000u,
	   0u,
	1000u,
	 500u,
	 300u
} ;

static	uint32_t	rcIndex ;		// Index of next value in rcValues

// The RC test use an interrupt callback
// cause argument is the pTimDesc->pTim->SR register: Do not use SR register in the callback
// The only cause is UPDATE
static	void rcCallback (const timDesc_t * pTimDesc, uintptr_t arg, uint32_t cause)
{
	uint32_t	eot = 0u ;

	if (rcIndex == (RC_COUNT - 1u))
	{
		eot = 8000u ;	// Large gap between frames
	}

	timSetReload (pTimDesc, RC_MIN + RC_GAP + rcValues [rcIndex] + eot) ;
	timOutputChannelSet (pTimDesc, RC_CHANNEL, RC_MIN + rcValues [rcIndex]) ;

	rcIndex ++ ;
	if (rcIndex == RC_COUNT)
	{
		rcIndex = 0u ;
	}
	(void) arg ;
	(void) cause ;
	// UIF flag is cleared by interrupt handler
}

void	timRcTest (void)
{
	const timDesc_t	* pTimDesc ;
	TIM_TypeDef 	* pTim ;


	aaPrintf ("Timer RC generator test, output on PA0\n") ;

	// Get timer descriptor address
	if (timGetDesc (RC_ID, & pTimDesc) == TIM_EFAIL)
	{
		aaPrintf ("Unknown timer descriptor id: %u\n", RC_ID) ;
		return ;
	}

	// Configure the time base for 1MHz timer clock and min RC period
	timGetRegPtr (pTimDesc, & pTim) ;	// Get native pointer to TIM registers
	aaPrintf ("Timer clock frequency: %u Hz\n", rccGetTimerClockFrequency (pTim)) ;

	timTimeBaseInit  (pTimDesc, rccGetTimerClockFrequency (pTim) / 1000000u, RC_MIN + RC_GAP - 1u, TIM_INIT_FORCE_ARR) ;
	timConfigureIntr (pTimDesc, TIM_PRIORITY, rcCallback, 0u, TIM_INTR_UPDATE) ;

	// Configure channel. Use TIM_OCMODE_PWM1 to get positive pulses
	timOutputChannelInit (pTimDesc, RC_CHANNEL, TIM_OCMODE_PWM1) ;
	timOutputChannelSet  (pTimDesc, RC_CHANNEL, RC_MIN + rcValues [0]) ;
	timChannelEnable     (pTimDesc, 1u << RC_CHANNEL) ;

	rcIndex = 1u ;
	timStart (pTimDesc) ;

	while (1)
	{
		aaTaskDelay (bspGetTickRate ()) ;
		aaPrintf (".") ;
		rcValues [RC_COUNT - 1u] += 100u ;

		if (rcValues [RC_COUNT - 1u] > 1000u)
		{
			rcValues [RC_COUNT - 1u] = 0u ;
		}


		if (aaCheckGetChar () != 0u)
		{
			timStop (pTimDesc) ;
			break ;		// Test cancelled by the user
		}
	}
}

//------------------------------------------------------------------------------
//	Time base and external clock test
//	One timer generate a frequency, the second timer use external clock.
//	The generator output pin should be linked to the selected clock input of the second timer

#define	FREQ_GENID		3u		// Identifier of descriptor of time base frequency of generator timer
#define	FREQ_FREQID		2u		// Identifier of descriptor of timer using external clock
#define	FREQ_FREQ		1000u	// Time base frequency of generator timer

// TIM3 output frequency:	PC6 (channel 1)
// TIM2 input clock:		PA0	(channel 1, or ETR)

// Choose input to test:
#define	FREQ_TEST_TI1FP1
//#define	FREQ_TEST_TI1F_ED
//#define	FREQ_TEST_ETRN

void	timFreqTest (void)
{
	const timDesc_t		* pTimGen ;
	const timDesc_t		* pTimFreq ;
	uint32_t			timClock ;
	uint32_t			cnt ;

	aaPrintf ("Timer: Time base and external clock test\n") ;

	timClock = rccGetTimerClockFrequency (TIM1) ;
	aaPrintf ("TIM1 clock %10d Hz\n", timClock) ;

#if (defined TIM2)
	timClock = rccGetTimerClockFrequency (TIM2) ;
	aaPrintf ("TIM2 clock %10d Hz\n", timClock) ;
#endif

	// Generate the frequency with FREQ_GENID timer
	if (timGetDesc (FREQ_GENID, & pTimGen) == TIM_EFAIL)
	{
		aaPrintf ("Timer gen: Unknown descriptor id %u\n", FREQ_GENID) ;
		return ;
	}
	timTimeBaseInit (pTimGen, 0u, FREQ_FREQ, 0u) ;
	timOutputChannelInit (pTimGen, 1u, TIM_OCMODE_PWM1) ;
	timChannelEnable     (pTimGen, 1u << 1u) ;
	timStart (pTimGen) ;

	// Configure the input timer FREQ_FREQID
	if (timGetDesc (FREQ_FREQID, & pTimFreq) == TIM_EFAIL)
	{
		aaPrintf ("Timer freq: Unknown descriptor id %u\n", FREQ_FREQID) ;
		return ;
	}
	#if defined FREQ_TEST_TI1FP1
		if (TIM_EFAIL == timExtClockInit (pTimFreq, LL_TIM_TS_TI1FP1, 1u, TIM_ECLK_EDGE_RISING))
	#elif defined FREQ_TEST_ETRN
		if (TIM_EFAIL == timExtClockInit (pTimFreq, LL_TIM_TS_ETRF, 1u, 0u) )
	#elif defined FREQ_TEST_TI1F_ED
		if (TIM_EFAIL == timExtClockInit (pTimFreq, LL_TIM_TS_TI1F_ED, 1u, 0u))
	#endif
	{
		aaPrintf ("timExtClockInit fail\n") ;
		return ;
	}
	timStart (pTimFreq) ;

	while (1)
	{
		aaTaskDelay (bspGetTickRate ()) ;
		cnt = timGetCounter (pTimFreq) ;
		aaPrintf ("CNT = %u\n", cnt) ;

		if (aaCheckGetChar () != 0u)
		{
			timStop (pTimGen) ;
			timStop (pTimFreq) ;
			break ;		// Test cancelled by the user
		}

	}
}

//------------------------------------------------------------------------------
//	Encoder timer mode test

const timDesc_t	encoderTimerDesc =
{
	30u,							// Descriptor identifier
	3u,								// Timer number 1..n
#if (defined STM32F1 | defined STM32F4)
	RCC_APB1RSTR_TIM3RST_Pos,		// RCC_APBxRSTR_TIMnRST_Pos
	RCC_APB1ENR_TIM3EN_Pos,			// RCC_APBxRSTR_TIMnEN_Pos
#endif
#if (defined STM32H7)
	RCC_APB1LRSTR_TIM3RST_Pos,		// RCC_APBxLRSTR_TIMnRST_Pos
	RCC_APB1LENR_TIM3EN_Pos,		// RCC_APBxLRSTR_TIMnEN_Pos
#endif
	TIM_CAP_TYPE_GPT1 | TIM_CAP_CHAN_1_4 | TIM_CAP_UP_DOWN | TIM_CAP_DMA | TIM_CAP_ETR, 	// Capabilities flags
	TIM3,							// Pointer to TIM registers
	#if (defined STM32G0)
		#if (defined TIM4)
			TIM3_TIM4_IRQn,
		#else
			TIM3_IRQn,
		#endif
	#else
		TIM3_IRQn,
	#endif
	{
		{	'B',	4u,	AA_GPIO_AF_2,	AA_GPIO_MODE_ALT_PP | AA_GPIO_SPEED_MEDIUM | AA_GPIO_PULL_UP},		// Ch1
		{	'B',	5u,	AA_GPIO_AF_2,	AA_GPIO_MODE_ALT_PP | AA_GPIO_SPEED_MEDIUM | AA_GPIO_PULL_UP},		// Ch2
//		{	'A',	6u,	AA_GPIO_AF_2,	AA_GPIO_MODE_ALT_PP | AA_GPIO_SPEED_MEDIUM },		// Ch1
//		{	'C',	7u,	AA_GPIO_AF_2,	AA_GPIO_MODE_ALT_PP | AA_GPIO_SPEED_MEDIUM },		// Ch2
		{	0u,		8u,	AA_GPIO_AF_2,	AA_GPIO_MODE_ALT_PP | AA_GPIO_SPEED_MEDIUM },		// Ch3
		{	0u,		9u,	AA_GPIO_AF_2,	AA_GPIO_MODE_ALT_PP | AA_GPIO_SPEED_MEDIUM },		// Ch4
		{	0u,		0u,	AA_GPIO_AF_2,	AA_GPIO_MODE_ALT_PP },		// Ch1N
		{	0u,		0u,	AA_GPIO_AF_2,	AA_GPIO_MODE_ALT_PP },		// Ch2N
		{	0u,		0u,	AA_GPIO_AF_2,	AA_GPIO_MODE_ALT_PP },		// Ch3N
		{	0u,		2u,	AA_GPIO_AF_2,	AA_GPIO_MODE_ALT_PP | AA_GPIO_SPEED_HIGH },		// ETR
		{	0u,		0u,	AA_GPIO_AF_2,	AA_GPIO_MODE_ALT_PP },		// BKIN
	},
} ;

// Convert unsigned 16 bit word to signed 32 bits word
static inline int32_t u16ToI32 (uint32_t op1)
{
	int32_t result;

	__ASM volatile ("sxth %0, %1" : "=r" (result) : "r" (op1));
	return (result) ;
}


// If the timer has a 16 bits counter use u16ToI32() to get 32 bits signed values
// If the timer has a 32 bit counter, simply cast the value to int32_t
// Delta is the variation of the signed timer counter since the last poll.
// Delta accumulated in totalCount

void	timEncTest (void)
{
	const timDesc_t		* pTimEnc ;
	int32_t				timCount, prevTimCount ;
	int32_t				delta ;
	int32_t				totalCount ;

	aaPrintf ("Timer: Encoder test\n") ;

	// Get the timer descriptor
	pTimEnc = & encoderTimerDesc ;

	// Configure timer
	timEncoderInit (pTimEnc, TIM_ENC_X1) ;

	// Start encoder
	timStart (pTimEnc) ;

//	timSetCounter (pTimEnc, 0xFFFC) ;	// To test timer counter overflow
	timSetCounter (pTimEnc, 0u) ;
	prevTimCount = u16ToI32 (timGetCounter (pTimEnc)) ;
	totalCount = 0u ;
	while (1)
	{
		if (aaCheckGetChar () != 0u)
		{
			timStop (pTimEnc) ;
			break ;		// Test cancelled by the user
		}

		aaTaskDelay (bspGetTickRate () / 5) ;	// Poll 5 Hz
		timCount = u16ToI32 (timGetCounter (pTimEnc)) ;
		if (timCount == prevTimCount)
		{
			// No counter change
			continue ;
		}

		delta = timCount - prevTimCount ;
		totalCount += delta ;
		prevTimCount = timCount ;
		aaPrintf ("Enc = count %d  delta %d  tim %d\n", totalCount, delta, timCount) ;
	}
}

//------------------------------------------------------------------------------
//	Input capture test

#if (defined TIM2)		// Some devices like STM30G070 doesn't have TIM2

#if (defined STM32G0)
	#define	TIM2_AF		AA_GPIO_AF_2
#else
	#define	TIM2_AF		AA_GPIO_AF_1
#endif

const timDesc_t	inputTimerDesc =
{
	21u,							// Descriptor identifier
	2u,								// Timer number 1..n
#if (defined STM32F1 | defined STM32F4)
	RCC_APB1RSTR_TIM2RST_Pos,		// RCC_APBxRSTR_TIMnRST_Pos
	RCC_APB1ENR_TIM2EN_Pos,			// RCC_APBxRSTR_TIMnEN_Pos
#endif
#if (defined STM32H7)
	RCC_APB1LRSTR_TIM2RST_Pos,		// RCC_APBxRSTR_TIMnRST_Pos
	RCC_APB1LENR_TIM2EN_Pos,		// RCC_APBxRSTR_TIMnEN_Pos
#endif
	TIM_CAP_TYPE_GPT1 | TIM_CAP_CHAN_1_4 | TIM_CAP_UP_DOWN | TIM_CAP_DMA | TIM_CAP_ETR, 	// Capabilities flags
	TIM2,							// Pointer to TIM registers
	TIM2_IRQn,
	{
		{	'A',	15,	TIM2_AF,		AA_GPIO_MODE_ALT_PP_UP	| AA_GPIO_SPEED_MEDIUM },	// Ch1
		{	'A',	1,	TIM2_AF,		AA_GPIO_MODE_ALT_PP		| AA_GPIO_SPEED_MEDIUM },	// Ch2
		{	0,		2,	AA_GPIO_AF_0,	AA_GPIO_MODE_ALT_PP		| AA_GPIO_SPEED_MEDIUM },	// Ch3
		{	0,		3,	AA_GPIO_AF_0,	AA_GPIO_MODE_ALT_PP		| AA_GPIO_SPEED_MEDIUM },	// Ch4
		{	0,		0,	AA_GPIO_AF_0,	0 },	// Ch1N
		{	0,		0,	AA_GPIO_AF_0,	0 },	// Ch2N
		{	0,		0,	AA_GPIO_AF_0,	0 },	// Ch3N
		{	0,		5,	AA_GPIO_AF_0,	AA_GPIO_MODE_ALT_PP		| AA_GPIO_SPEED_MEDIUM },	// ETR
		{	0,		0,	AA_GPIO_AF_0,	0 },	// BKIN
	},
} ;

// This callback is used only for CCR1 interrupt
// cause argument is the pTimDesc->pTim->SR register: Do not use RS register in the callback

static	void inputCaptureCallback (const timDesc_t * pTimDesc, uintptr_t arg, uint32_t cause)
{
	TIM_TypeDef		* pTim ;
	uint32_t		delay ;

	timGetRegPtr (pTimDesc, & pTim) ;

	// If CCR2 no present ignore
	if ((cause & TIM_SR_CC2IF) == 0u)
	{
		// CCIF1 without CC2IF : ignore
		// The CC1IF flag is cleared by reading CCR1
		delay = pTim->CCR1 ;
		(void) delay ;		// To avoid compiler unused variable warning
		return ;
	}

	// The CC1IF flag is cleared by reading CCR1
	delay = pTim->CCR1 - pTim->CCR2 ;
	aaLogMes ("1:%08X 2:%08X 1-2=%8u (%8u us)\n", pTim->CCR1, pTim->CCR2, delay, delay / (rccGetTimerClockFrequency (pTim)/1000000u), 0) ;

	(void) arg ;
	// UIF flag is cleared by the interrupt handler
}

// Use TIM2_CH1 and TIM2_CH2 to capture rising and falling edges of a negative pulse
// => allows to compute the pulse duration

void	timInputCaptureTest (void)
{
	timInputInit			(& inputTimerDesc, 1u) ;
	// Using TIM_ICMODE_TISWAP on channel 2, we use only one pin: the channel 1 pin
	timInputChannelInit		(& inputTimerDesc, 1u, TIM_ICMODE_TISTD  | TIM_ICMODE_RISING) ;
	timInputChannelInit		(& inputTimerDesc, 2u, TIM_ICMODE_TISWAP | TIM_ICMODE_FALLING) ;
	timChannelEnable		(& inputTimerDesc, (1u << 1u) | (1u << 2u)) ;
	timConfigureIntr		(& inputTimerDesc, TIM_PRIORITY, inputCaptureCallback, '/', LL_TIM_DIER_CC1IE) ;

	timStart				(& inputTimerDesc) ;

	while (1)
	{
		if (aaCheckGetChar () != 0u)
		{
			break ;		// Test cancelled by the user
		}

		aaTaskDelay (2u) ;
	}
	timStop (& inputTimerDesc) ;
	timChannelDisable (& inputTimerDesc, 1) ;
	timChannelDisable (& inputTimerDesc, 2) ;
}

#endif	// TIM2

//------------------------------------------------------------------------------
//	Interactive timer test

void	timTest ()
{
	uint32_t			cc ;

	cc = rccGetTimerClockFrequency (TIM1) ;
	aaPrintf ("TIM1  %3u.%06u MHz\n", cc / 1000000, cc % 1000000) ;

#if (defined TIM2)
	cc = rccGetTimerClockFrequency (TIM2) ;
	aaPrintf ("TIM2  %3u.%06u MHz\n", cc / 1000000, cc % 1000000) ;
#endif

#if (defined TIM3)
	cc = rccGetTimerClockFrequency (TIM3) ;
	aaPrintf ("TIM3  %3u.%06u MHz\n", cc / 1000000, cc % 1000000) ;
#endif
#if (defined TIM4)
	cc = rccGetTimerClockFrequency (TIM4) ;
	aaPrintf ("TIM4  %3u.%06u MHz\n", cc / 1000000, cc % 1000000) ;
#endif
#if (defined TIM6)
	cc = rccGetTimerClockFrequency (TIM6) ;
	aaPrintf ("TIM6  %3u.%06u MHz\n", cc / 1000000, cc % 1000000) ;
#endif
#if (defined TIM8)
	cc = rccGetTimerClockFrequency (TIM8) ;
	aaPrintf ("TIM8  %3u.%06u MHz\n", cc / 1000000, cc % 1000000) ;
#endif
#if (defined TIM14)
	cc = rccGetTimerClockFrequency (TIM14) ;
	aaPrintf ("TIM14 %3u.%06u MHz\n", cc / 1000000, cc % 1000000) ;
#endif

	while (1)
	{
		aaPrintf ("TIMER test, h for menu help\n") ;
		cc = aaGetChar () ;
		aaPutChar ((char) cc) ;
		aaPutChar ('\n') ;

		if (cc == 'h')
		{
			aaPrintf ("1    Time base generator\n") ;
			aaPrintf ("2    PWM output\n") ;
			aaPrintf ("3    One shot pulse\n") ;
			aaPrintf ("4    Radio Controller signal generator\n") ;
			aaPrintf ("5    External clock\n") ;
			aaPrintf ("6    Quadratic encoder test\n") ;
#if (defined TIM2)		// Some devices like STM30G070 doesn't have TIM2
			aaPrintf ("7    Input capture test\n") ;
#endif	// TIM2
			aaPrintf ("8    Frequency generator\n") ;
			aaPrintf ("q    Quit test\n") ;
		}

		if (cc == 'q')
		{
			break ;	// End of test
		}

		switch (cc)
		{
			case '1':
				timPwmTest (0u) ;		// Time base
				break ;

			case '2':
				timPwmTest (1u) ;		// PWM output
				break ;

			case '3':
				timOneShotTest () ;
				break ;

			case '4':
				timRcTest () ;
				break ;

			case '5':
				timFreqTest () ;
				break ;

			case '6':
				timEncTest () ;
				break ;

#if (defined TIM2)		// Some devices like STM30G070 doesn't have TIM2
			case '7':
				timInputCaptureTest () ;
				break ;
#endif	// TIM2

			case '8':
				timFrequencyGeneratorTest () ;
				break ;

			default:
				aaPrintf ("Error: Invalid test number\n") ;
				break ;
		}
		if (aaCheckGetChar () != 0u)
		{
			(void) aaGetChar () ;
		}
	}
}

//------------------------------------------------------------------------------
